/**
 * \addtogroup osSWinc
 * @{
 * \addtogroup uipApp
 * @{
 * \defgroup dhcpclient DHCP client Dynamic Host Configuration Protocol
 * @{
 * \details
 * The Dynamic Host Configuration Protocol (DHCP) is used for getting IP
 * configuration data from a server.
 *
 * This is Adam Dunkels' DHCP client implementation on top of his
 * \ref uip "uIP" TCP/IP stack &mdash; Albrecht Weinert's adaption to
 * \ref intro_sec "weAutSys" (and \ref intro_sec weAut_01):<ul>
 * <li> Doxygen documentation </li>
 * <li> publishing more symbols and functions (.c/.h re-organising) </li>
 * <li> performance improvements for sake of ATMega 8bit C and ENC28J60 </li>
 * <li> re-do of call-backs lest to loose valuable DHCP information needed
 *      by customer's applications </li>
 * <li> requesting (and using) more DHCP options </li>
 * <li> re-doing DHCP automatically after
 *      \ref DHCP_OPTION_RENEW_TIME "renew time" set by server answer
 *      (using a  \ref sec_dur_TIMER_TYPE "seconds duration" timer) </li></ul>
 * Besides the impact on dhcp.c and dhcp.h those improvements implied
 * some changes on other \ref uip "uIP" files, too.
 */

/**
 * \file dhcpc.h
 *
 * Definitions for the DHCP client
 *
 * The Dynamic Host Configuration Protocol's (DHCP) purpose is getting IP
 * configuration data from a server.
 *
 *  Adam Dunkels' original copyright notice <br /> &nbsp;
 *  Copyright (c) 2005, Swedish Institute of Computer Science.
 *  All rights reserved. <br />
 *  (contained in this file) still holds. For modifications:
 *
 *  <br /> &nbsp;
 * \ref copyLicNote " Copyright &copy; " 2011 Albrecht Weinert, Bochum
 * \author Adam Dunkels <adam@dunkels.com>
 * \author  Albrecht Weinert  &nbsp; <<a href="http://a-weinert.de">a-weinert.de</a>>
 * 2
 * 25.01.2017
 */
//  SVN last change by A. Weinert


/* Copyright (c) 2005, Swedish Institute of Computer Science
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * This file is part of the uIP TCP/IP stack.
 *
 * Modified (adapted) by: Albrecht Weinert
 * For the modification: Copyright 2011 Albrecht Weinert <a-weinert.de>
 *
 * This file is also part of the \ref intro_sec "weAutSys" runtime system
 * by weinert - automation <weinert-automation.de>
 */
#ifndef __DHCPC_H__
#define __DHCPC_H__

#include "we-aut_sys/timing.h"
#include "pt/pt.h"

#define DHCPC_SERVER_PORT  67 //!< Well known DHCP server port \showinitializer
#define DHCPC_CLIENT_PORT  68 //!< Used DHCP client port \showinitializer

#define DHCPDISCOVER  1 //!< DHCP message type \showinitializer
#define DHCPOFFER     2 //!< DHCP message type \showinitializer
#define DHCPREQUEST   3 //!< DHCP message type \showinitializer
#define DHCPDECLINE   4 //!< DHCP message type \showinitializer
#define DHCPACK       5 //!< DHCP message type \showinitializer
#define DHCPNAK       6 //!< DHCP message type \showinitializer
#define DHCPRELEASE   7 //!< DHCP message type \showinitializer
#define DHCPINFORM    8 //!< DHCP message type \showinitializer

#define DHCP_REQUEST  1 //!< DHCP message operation type \showinitializer
#define DHCP_REPLY    2 //!< DHCP message operation type \showinitializer

/** The DHCP message structure */
struct dhcpMsg_t {
  uint8_t op, htype, hlen, hops;
  uint8_t xid[4]; //!< To match incoming DHCP messages with pending requests
  uint16_t secs, flags;
  uint8_t ciaddr[4];
  uint8_t yiaddr[4];
  uint8_t siaddr[4];
  uint8_t giaddr[4];
  uint8_t chaddr[16];
#if ! defined(UIP_CONF_DHCP_LIGHT)
  uint8_t sname[64];
  uint8_t file[128];
#endif
  uint8_t options[312];
};

#define BOOTP_BROADCAST 0x8000 //!< DHCP message flag value \showinitializer

#define DHCP_HTYPE_ETHERNET 1 //!< DHCP message type value \showinitializer
#define DHCP_HLEN_ETHERNET  6 //!< DHCP message hardware address length \showinitializer


/** DHCP option field code: message type
 *
 *  This is usually (if not always) the first option in a DHCP package. It
 *  redundantly gives the message type in form of an option.<br />
 *  The rationale behind is to inform an option parser / option handler in
 *  an uniform way about the nature of the package in question.
 *
 *  Especially \ref  DHCPACK would mean that all other options are valid
 *  and meant to be used by the client.
 *
 *  For the possible options see also RFC2132.
 *  \showinitializer
 *  \internal comment by A. Weinert
 */
#define DHCP_OPTION_MSG_TYPE     53

#define DHCP_OPTION_SUBNET_MASK   1 //!< DHCP option field code \showinitializer

/** DHCP option field code: time offset
 *
 *  This is the time offset of the client's subnet (location) in seconds
 *  relative to UTC. This value can be used to calculate local time from
 *  NTP server responses.
 *  \showinitializer
 *  \internal addition by A. Weinert
 */
#define DHCP_OPTION_TIME_OFFSET  2

/** DHCP option field code: time server(s)
 *
 *  This is a list of 1..n time server addresses with decreasing preference.
 *  By DHCP definition those servers use time protocol (RFC 868, port 37) not
 *  NTP (RFC 1305, port 123).
 *  \showinitializer
 *
 *  <b> &nbsp; See also: &nbsp; </b> \ref  DHCP_OPTION_NTP_SERVERS
 *  \internal addition by A. Weinert
 */
#define DHCP_OPTION_TIME_SERVERS 4


#define DHCP_OPTION_ROUTER        3 //!< DHCP option field code \showinitializer

/** DHCP option field code: name server(s)
 *
 *  This is a list of 1..n name server addresses with decreasing preference.
 *  \showinitializer
 *  \internal addition by A. Weinert
 */
#define DHCP_OPTION_NAME_SERVERS 5


#define DHCP_OPTION_DNS_SERVER    6 //!< DHCP option field code \showinitializer

/** DHCP option field code: host name
 *
 *  This is to give the client a name by the central organisation.
 *  \showinitializer
 *  \internal addition by A. Weinert
 */
#define DHCP_OPTION_CLIENT_NAME 12

/** DHCP option field code: domain name
 *
 *  This is to give the domain name (sometimes called DNS suffix) that client
 *  should use when resolving hostnames via the Domain Name System
 *  (\ref resolv "DNS"). This option can deliver just one sufix, not a list
 *  of.
 *  \showinitializer
 *  \internal addition by A. Weinert
 */
#define DHCP_OPTION_DOMAIN_NAME 15

/** DHCP option field code: default IP TTL
 *
 *  This is the (default) time to live for outgoing connections.
 *  \showinitializer
 *  \internal addition by A. Weinert
 */
#define DHCP_OPTION_DEFAULT_TTL 23

/** DHCP option field code: broadcast address
 *
 *  \showinitializer
 *  \internal addition by A. Weinert
 */
#define DHCP_OPTION_BROADCAST_ADDR 28

/** DHCP option field code: time server(s)
 *
 *  This is a list of 1..n NTP server addresses with decreasing preference.
 *  Those servers use NTP (RFC 1305, port 123).
 *
 *  <b> &nbsp; See also: &nbsp; </b> \ref  DHCP_OPTION_TIME_SERVERS
 *  \internal addition by A. Weinert
 *  \showinitializer
 */
#define DHCP_OPTION_NTP_SERVERS 42


#define DHCP_OPTION_REQ_IPADDR 50 //!< DHCP option field code \showinitializer

/** The address lease time
 *
 *  This option is the requested respectively granted lease time for the
 *  address given by the DHCP server in units of seconds (unsigned 32 bit,
 *  wrong endian).
 *
 *  Time to renew (T1) is 50% of the lease time by default and may be sent
 *  by the server as \ref  DHCP_OPTION_RENEW_TIME option too.<br />
 *  Time to rebind (T2) is 87.5% (7/8) of the lease time by default and may
 *  be sent by the server as \ref  DHCP_OPTION_REBND_TIME option too.<br />
 * \showinitializer
 */
#define DHCP_OPTION_LEASE_TIME 51
#define DHCP_OPTION_RENEW_TIME 58 //!< DHCP option field code \showinitializer
#define DHCP_OPTION_REBND_TIME 59 //!< DHCP option field code \showinitializer

/** DHCP option field code: DHCH server IP address
 *
 *  This option allows to distinguish several DHCP servers in requests and
 *  offers.
 *  \showinitializer
 */
#define DHCP_OPTION_SERVER_ID  54 //!< DHCP option field code \showinitializer

/** DHCP request option field code: Parameter Request List
 *
 *  \showinitializer
 */
#define DHCP_OPTION_REQ_LIST     55

/** DHCP option field code: SMTP server(s)
 *
 *  This is a list of 1..n server addresses for the
 *  Simple Mail Transport Protocol (SMTP) with decreasing preference.
 *  \showinitializer
 *  \internal addition by A. Weinert
 */
#define DHCP_OPTION_SMTP_SERVERS 69


/**  DHCP option field code: end of all options
 *
 *   This is always the last option in a DHCP package. It is empty and just
 *   signals the end of the list.
 */
#define DHCP_OPTION_END 255 //!< DHCP option field code \showinitializer


#define STATE_INITIAL         0 //!< DHCP state machine state \showinitializer
#define STATE_SENDING         1 //!< DHCP state machine state \showinitializer
#define STATE_OFFER_RECEIVED  2 //!< DHCP state machine state \showinitializer
#define STATE_CONFIG_RECEIVED 3 //!< DHCP state machine state \showinitializer

/** Structure for DHCP application state */
struct dhcpState_t {
  struct pt pt; //!< the protothreads (raw) structure
  char state;   //!< DHCP state machine state
  struct uip_udp_conn *conn;  //!< pointer to the connection (structure)
  struct timer_t timer; //!< timer (in ms while in protocol; in seconds renew time)
  const void *mac_addr; //!< pointer to device's MAC address

  uint8_t serverid[4]; //!< the DHCP server's IP address

  uint16_t ipaddr[2]; //!< the client's wanted / granted IP address
};

extern struct dhcpState_t dhcpState; //!< DHCP application state

/** Initialise the DHCP (client)
 *
 *  This function initialises the DHCP structures and state machine
 *  (= protothread) like does \ref dhcpReset. Then it is tried to establish
 *  connections to DHCP \ref DHCPC_SERVER_PORT "server" and
 *  \ref DHCPC_CLIENT_PORT "client" port(s).
 *
 *  The protocol is not started by this function; the work is done by
 *  \ref dhcpc_appcall.
 *  \param mac_addr pointer to the device's MAC address
 */
void dhcpInit(void const * mac_addr);

/** Reset the DHCP (client)
 *
 *  This function resets all DHCP state to initial. In contrast to
 *  \ref dhcpInit no trial to connect to DHCP ports is made and the protocol
 *  is not (not even indirectly) started.
 */
void dhcpReset(void);


/** Handle DHCP server events
 *
 *  This function handles DHCP events. And it is a (as a protothread) the
 *  DHCP client state machine. It has to be called in the \ref udp_appcall
 *  in a manner like \code
   const uint16_t remPort = convert16endian(uip_udp_conn->rport);

   if(remPort == DHCPC_SERVER_PORT) {  // DHCP protocol
      dhcpc_appcall();
      return;
   } // DHCP protocol
 *  \endcode
 */
ptfnct_t dhcpc_appcall(void);


/** The structure of a DHCP message option field */
struct dhcpOption_t {
   uint8_t code;   //!< the option code
   uint8_t length; //!< the length of value (0 implied if  code == \ref DHCP_OPTION_END)
   char value[] ; //!< the value(s); an array of length bytes
}  ;

/** DHCP option received
 *
 *  This (callback) signals an option send by the DHCP server. Its up to the
 *  implementing application or system software to utilise (or ignore) the
 *  respective values.
 *  \param dhcpOption the option (one of the options) received
 *
 */
void dhcpcGotOption(const struct dhcpOption_t *dhcpOption);

/** DHCP success
 *
 *  This (callback) signals a successful walk through the DHCP protocol
 *  steps ended by acknowledge.
 *  \param respType at present always \ref DHCPACK; (no callback in other
 *                  cases yet)
 *  \param ipAddr   the assigned / leased IP address (in case of success)
 */
void dhcpc_configured(uint8_t respType, const uint16_t ipAddr[]);

#endif /* __DHCPC_H__ */

/** @} */
/** @} */
/** @} */ // osSWinc
