/**
 * \file
 * Common definitions for the uIP TCP/IP stack
 *
 * \author Adam Dunkels <adam@dunkels.com>
 *
 * The uIP TCP/IP stack header file contains definitions for a number
 * of C macros that are used by uIP programs as well as internal uIP
 * structures, TCP/IP header structures and function declarations.
 */


/*
 * Copyright (c) 2001-2003, Adam Dunkels.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * This file is part of the uIP TCP/IP stack.
 *
 * Modified (adapted) by: Albrecht Weinert
 * For the modification: Copyright 2011 Albrecht Weinert <a-weinert.de>
 *
 * This file is also part of the \ref intro_sec "weAutSys" runtime system
 * by weinert - automation <weinert-automation.de>
 *
 * $Id: uip.h 2 2017-01-25 16:49:03Z albrecht $
 */

#ifndef __UIP_H__
#define __UIP_H__

#include "uip/uipopt.h"
#include "pt/pt.h"

/**
 * \addtogroup osSWinc
 * @{
 * \addtogroup uip
 * @{
 */

#if UIP_STATISTICS
/** Statistics datatype
 *
 *  This type is used for uIP's statistic numbers.
 * \hideinitializer
 */
typedef uint16_t uip_stats_t;
#endif

/** Representation of an IP V4 address
 *
 *  For IP V4 (used as of this writing in small modules like weAut_01) this
 *  is \ref uip "uIP's" type \ref uip_ipaddr_t being just an array
 *  uint16_t [2] in the end.
 *
 *  \note An Internet address of e.g. 1.2.3.4 would reside in this array in
 *  the (wrong i.e. network i.e. big) endianess: { 0x0201 , 0x0403 } .
 */
typedef uint16_t uip_ip4addr_t[2]; //!< Representation of an IP (V4) address
typedef uint16_t uip_ip6addr_t[8]; //!< Representation of an IP (V6) address

#if UIP_CONF_IPV6
typedef uip_ip6addr_t uip_ipaddr_t; //!< IP address  is of type IP V6
#else /* UIP_CONF_IPV6 */
typedef uip_ip4addr_t uip_ipaddr_t; //!< IP address  is of type IP V4
#endif /* UIP_CONF_IPV6 */

/**
 * \defgroup uipconffunc uIP configuration functions
 * @{
 * \details
 * The uIP configuration functions are used for setting run-time
 * parameters in uIP such as IP addresses.
 */

/** Set the IP address of this host
 *
 *  The IP address is represented as a 4-byte array where the first octet
 *  of the IP address is put in the first member of the 4-byte array.
 *
 *  Example:
 \code

 uip_ipaddr_t addr;

 uip_ipaddr(&addr, 192,168,1,2);
 uip_sethostaddr(&addr);
 
 \endcode
 * \param addr A pointer to an IP address of type uip_ipaddr_t
 *
 *  <b> &nbsp; See also: &nbsp; </b> \ref uip_ipaddr()
 * \hideinitializer
 */
#define uip_sethostaddr(addr) uip_ipaddr_copy(uip_hostaddr, (addr))

/** Get the IP address of this host
 *
 *  The IP address is represented as a 4-byte array where the first octet
 *  of the IP address is put in the first member of the 4-byte array.
 *
 * Example:
 \code
 uip_ipaddr_t hostaddr;

 uip_gethostaddr(&hostaddr);
 \endcode
 * \param addr A pointer to a uip_ipaddr_t variable that will be
 * filled in with the currently configured IP address.
 *
 * \hideinitializer
 */
#define uip_gethostaddr(addr) uip_ipaddr_copy((addr), uip_hostaddr)

/** Set the default router's IP address
 *
 * \param addr  A pointer to a uip_ipaddr_t variable containing the IP
 *             address of the default router.
 * \sa uip_ipaddr()
 * \hideinitializer
 */
#define uip_setdraddr(addr) uip_ipaddr_copy(uip_draddr, (addr))

/** Set the netmask
 *
 * \param addr A pointer to a uip_ipaddr_t variable containing the IP
 * address of the netmask.
 *
 *  <b> &nbsp; See also: &nbsp; </b> \ref uip_ipaddr()
 * \hideinitializer
 */
#define uip_setnetmask(addr) uip_ipaddr_copy(uip_netmask, (addr))


/** Get the default router's IP address
 *
 * \param addr A pointer to a uip_ipaddr_t variable that will be
 *             filled in with the IP address of the default router.
 *
 * \hideinitializer
 */
#define uip_getdraddr(addr) uip_ipaddr_copy((addr), uip_draddr)

/** Get the netmask
 *
 * \param addr A pointer to a uip_ipaddr_t variable that will be
 *             filled in with the value of the netmask.
 *
 * \hideinitializer
 */
#define uip_getnetmask(addr) uip_ipaddr_copy((addr), uip_netmask)

/** @} */

/**
 * \defgroup uipinit uIP initialisation functions
 * @{
 * \details
 * The uIP initialisation functions are used for booting uIP.
 */

/** uIP initialisation function
 *
 * This function should be called at boot up to initialize the uIP
 * TCP/IP stack.
 */
void uip_init(void);

/** uIP initialisation function
 *
 * This function may be used at boot time to set the initial ip_id which is
 * an increasing number that is used for the IP ID field.
 */
void uip_setipid(uint16_t id);

/** @} */

/**
 * \defgroup uipdevfunc uIP device driver functions
 * @{
 * \details
 * These functions are used by a network device driver for interacting
 * with uIP.
 */

/** Process an incoming packet
 *
 *  This function should be called when the device driver has received
 *  a packet from the network. The packet from the device driver must
 *  be present in the uip_buf buffer, and the length of the packet
 *  should be placed in the uip_len variable.
 *
 *  When the function returns, there may be an outbound packet placed
 *  in the uip_buf packet buffer. If so, the uip_len variable is set to
 *  the length of the packet. If no packet is to be sent out, the
 *  uip_len variable is set to 0.
 *
 *  The usual way of calling the function is presented by the source
 *  code below.
 \code
  uip_len = devicedriver_poll();
  if(uip_len > 0) {
    uip_input();
    if(uip_len > 0) {
      devicedriver_send();
    }
  }
 \endcode
 *
 * \note If you are writing a uIP device driver that needs ARP
 * (Address Resolution Protocol), e.g., when running uIP over
 * Ethernet, you will need to call the uIP ARP code before calling
 * this function:
 \code
  #define BUF ((struct uip_eth_hdr *)&uip_buf[0])
  uip_len = ethernet_devicedrver_poll();
  if(uip_len > 0) {
    if(BUF->type == HTONS(UIP_ETHTYPE_IP)) {
      uip_arp_ipin();
      uip_input();
      if(uip_len > 0) {
        uip_arp_out();
	ethernet_devicedriver_send();
      }
    } else if(BUF->type == HTONS(UIP_ETHTYPE_ARP)) {
      uip_arp_arpin();
      if(uip_len > 0) {
	ethernet_devicedriver_send();
      }
    }
 \endcode
 *
 * \hideinitializer
 */
#define uip_input() uip_process(UIP_DATA)

/** Periodic processing for a connection identified by its number
 *
 * This function does the necessary periodic processing (timers, polling)
 * for a uIP TCP connection, and should be called when the periodic uIP timer
 * goes off. It should be called for every connection, regardless of whether
 * they are open or closed.
 *
 * When the function returns, it may have an outbound packet waiting for
 * service in the uIP packet buffer, and if so the \ref uip_len variable
 * is set to a value larger than zero. The device driver should hence be
 * called to send out the packet.
 *
 * The usual way of calling the function is through a for() loop like
 * this:
 \code
  for(i = 0; i < UIP_CONNS; ++i) {
    uip_periodic(i);
    if(uip_len > 0) {
      devicedriver_send();
    }
  }
 \endcode
 *
 * \note If you are writing a uIP device driver that needs ARP
 * (Address Resolution Protocol), e.g., when running uIP over
 * Ethernet, you will need to call the uip_arp_out() function before
 * calling the device driver:
 \code
  for(i = 0; i < UIP_CONNS; ++i) {
    uip_periodic(i);
    if(uip_len > 0) {
      uip_arp_out();
      ethernet_devicedriver_send();
    }
  }
 \endcode
 *
 * \param conn number of the connection which is to be periodically polled
 *
 * \hideinitializer
 * \sa uip_conn_active
 */
#define uip_periodic(conn) do { uip_conn = &uip_conns[conn]; \
                                uip_process(UIP_TIMER); } while (0)

/** Check if a connection identified by its number is active
 */
#define uip_conn_active(conn) (uip_conns[conn].tcpstateflags != UIP_CLOSED)


/** Perform periodic processing for a connection identified by pointer
 *
 * Same as \ref uip_periodic() but takes a pointer to the actual uip_conn
 * struct instead of its number. This function can be used to force periodic
 * processing of a specific connection.
 *
 *  <b> &nbsp; See also: &nbsp; </b> \ref uip_periodic
 *
 * \param conn A pointer to the uip_conn struct of the connection
 * \hideinitializer
 */
#define uip_periodic_conn(conn) do { uip_conn = conn; \
                                     uip_process(UIP_TIMER); } while (0)

/** Request that a particular connection should be polled
 *
 * Similar to uip_periodic_conn() but does not perform any timer
 * processing. The application is polled for new data.
 *
 * \param conn A pointer to the uip_conn struct for the connection to
 * be processed.
 *
 * \hideinitializer
 */
#define uip_poll_conn(conn) do { uip_conn = conn; \
                                 uip_process(UIP_POLL_REQUEST); } while (0)


#if UIP_UDP

/** Periodic processing for a UDP connection identified by its number
 *
 * This function is essentially the same as \ref uip_periodic(), but for
 * UDP connections. It is called in a similar fashion as the
 * uip_periodic() function:
 \code
  for(i = 0; i < UIP_UDP_CONNS; i++) {
    uip_udp_periodic(i);
    if(uip_len > 0) {
      devicedriver_send();
    }
  }
 \endcode
 *
 * \note As for the uip_periodic() function, special care has to be
 * taken when using uIP together with ARP and Ethernet:
 \code
  for(i = 0; i < UIP_UDP_CONNS; i++) {
    uip_udp_periodic(i);
    if(uip_len > 0) {
      uip_arp_out();
      ethernet_devicedriver_send();
    }
  }
 \endcode
 *
 * \param conn The number of the UDP connection to be processed.
 *
 * \hideinitializer
 */
#define uip_udp_periodic(conn) do { uip_udp_conn = &uip_udp_conns[conn]; \
                                uip_process(UIP_UDP_TIMER); } while (0)

/** Periodic processing for a UDP connection  pointed to
 *
 * Same as uip_udp_periodic() but takes a pointer to the actual uip_conn
 * struct instead of an integer as its argument. This function can be used
 * to force periodic processing of a specific connection.
 *
 * \param conn A pointer to the uip_udp_conn struct for the connection
 * to be processed.
 *
 * \hideinitializer
 */
#define uip_udp_periodic_conn(conn) do { uip_udp_conn = conn; \
                    uip_process(UIP_UDP_TIMER); } while (0)

#endif /* UIP_UDP */


/** The uIP packet buffer
 *
 *  The uip_buf array is used to hold incoming and outgoing packets.
 *  The device driver should place incoming data into this buffer.
 *  When sending data, the device driver should read the link level headers
 *  and the TCP/IP headers from this buffer. The size of the link level
 *  headers is configured by the UIP_LLH_LEN define.
 *
 *  \note The application data need not be placed in this buffer, so
 *  the device driver must read it from the place pointed to by the
 *  uip_appdata pointer as illustrated by the following example:
 \code
 void
 devicedriver_send(void)
 {
    hwsend(&uip_buf[0], UIP_LLH_LEN);
    if(uip_len <= UIP_LLH_LEN + UIP_TCPIP_HLEN) {
      hwsend(&uip_buf[UIP_LLH_LEN], uip_len - UIP_LLH_LEN);
    } else {
      hwsend(&uip_buf[UIP_LLH_LEN], UIP_TCPIP_HLEN);
      hwsend(uip_appdata, uip_len - UIP_TCPIP_HLEN - UIP_LLH_LEN);
    }
 }
 \endcode
 */
extern uint8_t uip_buf[];
extern char * uip_buf_alias; //!< alias to \ref  uip_buf


/** @} */

//--------------------------------------------------------------------------


/**
 * \defgroup uipappfunc uIP application functions
 * @{
 * \details
 * Functions used by an application running of top of uIP.
 *
 * These functions do the opening and closing of connections,
 * sending and receiving data, etc.
 */

/** Start listening on the specified port
 *
 * \note Since this function expects the port number in network byte
 * order, a conversion using HTONS() or htons() is necessary.
 *
 \code
 uip_listen(HTONS(80));
 \endcode
 * \param port A 16-bit port number in network byte order.
 */
void uip_listen(uint16_t port);

/** Stop listening on the specified port
 *
 *  \note Since this function expects the port number in network byte
 *  order, a conversion using HTONS() or htons() is necessary.
 *
 \code
 uip_unlisten(HTONS(80));
 \endcode
 *  \param port A 16-bit port number in network byte order.
 */
void uip_unlisten(uint16_t port);

/** Connect to a remote host using TCP
 *
 *  This function is used to start a new connection to the specified port on
 *  the specified host. It allocates a new connection identifier, sets the
 *  connection to the SYN_SENT state and sets the retransmission timer to 0.
 *  This will cause a TCP SYN segment to be sent out the next time this
 *  connection is periodically processed, which usually is done within 0.5
 *  seconds after the call to uip_connect().
 *
 *  \note This function is available only if support for active open
 *  has been configured by defining UIP_ACTIVE_OPEN to 1 in uipopt.h.
 *
 *  \note Since this function requires the port number to be in network
 *  byte order, a conversion using HTONS() or htons() is necessary.
 *
 \code
 uip_ipaddr_t ipaddr;

 uip_ipaddr(&ipaddr, 192,168,1,2);
 uip_connect(&ipaddr, HTONS(80));
 \endcode
 *  \param ripaddr the IP address of the remote hot.
 *  \param port    a 16-bit port number in network byte order.
 *  \return        a pointer to the uIP connection identifier for the new
 *                 connection, or NULL if no connection could be allocated.
 */
struct uipConn_t *uip_connect(uip_ipaddr_t *ripaddr, uint16_t port);


/**
 * \internal
 *
 * Check if a connection has outstanding (i.e., unacknowledged) data.
 *
 * \param conn A pointer to the uip_conn structure for the connection.
 * \hideinitializer
 */
#define uip_outstanding(conn) ((conn)->len)

/** Send data on the current connection
 *
 *  This function is used to send out a single segment of TCP data. Only
 *  applications that have been invoked by uIP for event processing can
 *  send data.
 *
 *  The amount of data that actually is sent out after a call to this
 *  function is determined by the maximum amount of data TCP allows. uIP will
 *  automatically crop the data so that only the appropriate amount of data is
 *  sent. The function uip_mss() can be used to query uIP for the amount of
 *  data that actually will be sent.
 *
 *  \note This function does not guarantee that the sent data will arrive at
 *  the destination. If the data is lost in the network, the application will
 *  be invoked with the uip_rexmit() event being set. The application will
 *  then have to re-send the data using this function.
 *
 * \param data A pointer to the data which is to be sent <br />
 *  Using the buffer pointed to by \ref uip_appdata to prepare the send data
 *  (and then use as this parameter) is strongly recommended.
 * \param len The maximum amount of data bytes to be sent. <br />
 *  Due to internal buffer limits this must not be more than
 *  \ref UIP_APPDATA_SIZE
 * \hideinitializer
 */
void uip_send(const void *data, int len);

/** The length of any incoming data that is currently available (if available)
 *  in the uip_appdata buffer
 *
 * This returns the length of any incoming data that is currently
 * available in the uip_appdata buffer.<br />
 * If works only if there is any data available at all: the test function
 * \ref uip_newdata() must first be used to check this condition.
 *
 * \hideinitializer
 */
#define uip_datalen()   uip_len

#if UIP_URGDATA > 0
/** The length of any out-of-band data (urgent data) that has arrived
 *  on the connection
 *
 * \note The configuration parameter UIP_URGDATA must be set for this
 * function to be enabled.
 *
 * \hideinitializer
 */
#define uip_urgdatalen()    uip_urglen
#endif

/** Close the current connection
 *
 *  This function will close the current connection in a nice way.
 *
 * \hideinitializer
 */
#define uip_close()         (uip_flags = UIP_CLOSE)

/** Abort the current connection
 *
 * This function will abort (reset) the current connection, and is
 * usually used when an error has occurred that prevents using the
 * uip_close() function.
 *
 * \hideinitializer
 */
#define uip_abort()         (uip_flags = UIP_ABORT)

/** Tell the sending host to stop sending data
 *
 * This function will close our receiver's window so that we stop
 * receiving data for the current connection.
 *
 * \hideinitializer
 */
#define uip_stop()          (uip_conn->tcpstateflags |= UIP_STOPPED)

/** Find out if the current connection has been previously stopped with
 * uip_stop()
 *
 * \hideinitializer
 */
#define uip_stopped(conn)   ((conn)->tcpstateflags & UIP_STOPPED)

/** Restart the current connection, if is has previously been stopped
 *  with uip_stop()
 *
 * This function will open the receiver's window again so that we
 * start receiving data for the current connection.
 *
 * \hideinitializer
 */
#define uip_restart()do { uip_flags |= UIP_NEWDATA; \
   uip_conn->tcpstateflags &= ~UIP_STOPPED; \
} while(0)


/* uIP tests that can be made to determine in what state the current
   connection is, and what the application function should do. */

/** Is the current connection a UDP connection?
 *
 * This function checks whether the current connection is a UDP connection.
 *
 * \hideinitializer
 */
#define uip_udpconnection() (uip_conn == NULL)

/** Is new incoming data available?
 *
 *  Will return non-zero resp. true if there is new data for the application
 *  present at the uip_appdata pointer. The size of the data is
 *  available through the uip_len variable.
 *
 * \hideinitializer
 */
#define uip_newdata()   (uip_flags & UIP_NEWDATA)

/** Has previously sent data been acknowledged?
 *
 *  Will return non-zero if the previously sent data has been
 *  acknowledged by the remote host. This means that the application
 *  can send new data.
 *
 * \hideinitializer
 */
#define uip_acked()   (uip_flags & UIP_ACKDATA)

/** Has the connection just been connected?
 *
 *  Will return non-zero if the current connection has been connected to
 *  a remote host. This will happen both if the connection has been
 *  actively opened (with uip_connect()) or passively opened (with
 *  uip_listen()).
 *
 * \hideinitializer
 */
#define uip_connected() (uip_flags & UIP_CONNECTED)

/** Has the connection been closed by the other end?
 *
 * Is non-zero if the connection has been closed by the remote
 * host. The application may then do the necessary clean-ups.
 *
 * \hideinitializer
 */
#define uip_closed()    (uip_flags & UIP_CLOSE)

/** Has the connection been aborted by the other end
 *
 *  Non-zero if the current connection has been aborted (reset) by the
 *  remote host.
 *
 * \hideinitializer
 */
#define uip_aborted()    (uip_flags & UIP_ABORT)

/** Has the connection timed out?
 *
 *  Non-zero if the current connection has been aborted due to too many
 *  retransmissions.
 *  \hideinitializer
 */
#define uip_timedout()    (uip_flags & UIP_TIMEDOUT)

/** Do we need to retransmit previously data?
 *
 *  Returns non-zero if the previously sent data has been lost in the network,
 *  and the application should retransmit it. The application should send
 *  exactly the same data as it did the last time, using the \ref uip_send()
 *  function.
 *  \hideinitializer
 */
#define uip_rexmit()     (uip_flags & UIP_REXMIT)

/** Is the connection being polled by uIP?
 *
 * Is non-zero if the reason the application is invoked is that the
 *  current connection has been idle for a while and should be polled.
 *
 *  The polling event can be used for sending data without having to
 *  wait for the remote host to send data.
 *  \hideinitializer
 */
#define uip_poll()       (uip_flags & UIP_POLL)

/** Get the initial maximum segment size (MSS) of the current connection
 *
 * \hideinitializer
 */
#define uip_initialmss()             (uip_conn->initialmss)

/** Current maximum segment size that can be sent on the current connection
 *
 *  The current maximum segment size that can be sent on the
 *  connection is computed from the receiver's window and the MSS of
 *  the connection (which also is available by calling
 *  uip_initialmss()).
 *  \hideinitializer
 */
#define uip_mss()             (uip_conn->mss)

/** Set up a new UDP connection
 *
 *  This function sets up a new UDP connection. The function will
 *  automatically allocate an unused local port for the new connection.
 *  However, another port can be chosen by using the uip_udp_bind() call,
 *  after the uip_udp_new() function has been called.
 *
 * Example:
 \code
 uip_ipaddr_t addr;
 struct uip_udp_conn *c;
 
 uip_ipaddr(&addr, 192,168,2,1);
 c = uip_udp_new(&addr, HTONS(12345));
 if(c != NULL) {
   uip_udp_bind(c, HTONS(12344));
 }
 \endcode
 * \param ripaddr the IP address of the remote host
 * \param rport   the remote port number in network byte order
 * \return        the uip_udp_conn structure for the new connection
 *                or NULL if no connection could be allocated.
 */
struct uip_udp_conn * uip_udp_new(uip_ipaddr_t *ripaddr, uint16_t rport);

/** Remove / give up an UDP connection
 *
 *  \param conn pointer to the uip_udp_conn structure for the connection
 *  \hideinitializer
 */
#define uip_udp_remove(conn) (conn)->lport = 0

/** Bind a UDP connection to a local port
 *
 * \param conn pointer to the uip_udp_conn structure for the connection
 * \param port The local port number, in network byte order
 *
 * \hideinitializer
 */
#define uip_udp_bind(conn, port) (conn)->lport = port

/** Send a UDP datagram of length len on the current connection
 *
 *  This function can only be called in response to a UDP event (poll
 *  or newdata). The data must be present in the uip_buf buffer, at the
 *  place pointed to by the uip_appdata pointer.
 *
 * \param len The length of the data in the uip_buf buffer
 * \hideinitializer
 */
#define uip_udp_send(len) uip_send((char *)uip_appdata, len)

/** @} */

/**
 * \defgroup uipconvfunc uIP conversion functions
 * @{
 * \details
 * These functions can be used for converting between different data
 * formats used by uIP.
 */
 
/** Construct an IP address from four bytes
 *
 *  This function constructs an IP address of the type that uIP handles
 *  internally from four bytes. The function is handy for specifying IP
 *  addresses to use with e.g. the uip_connect() function.
 *
 * Example:
 \code
 uip_ipaddr_t ipaddr;
 struct uip_conn *c;
 
 uip_ipaddr(&ipaddr, 192,168,1,2);
 c = uip_connect(&ipaddr, HTONS(80));
 \endcode
 *
 * \param addr A pointer to a uip_ipaddr_t variable that will be
 * filled in with the IP address.
 *
 * \param addr0 The first octet of the IP address.
 * \param addr1 The second octet of the IP address.
 * \param addr2 The third octet of the IP address.
 * \param addr3 The forth octet of the IP address.
 * \hideinitializer
 */
#define uip_ipaddr(addr, addr0,addr1,addr2,addr3) do { \
   ((uint16_t *)(addr))[0] = HTONS(((addr0) << 8) | (addr1)); \
   ((uint16_t *)(addr))[1] = HTONS(((addr2) << 8) | (addr3)); \
} while(0)

/** Construct an IPv6 address from eight 16-bit words
 *
 *  This function constructs an IPv6 address.
 *  \hideinitializer
 */
#define uip_ip6addr(addr, addr0,addr1,addr2,addr3,addr4,addr5,addr6,addr7) do { \
           ((uint16_t *)(addr))[0] = HTONS((addr0)); \
           ((uint16_t *)(addr))[1] = HTONS((addr1)); \
           ((uint16_t *)(addr))[2] = HTONS((addr2)); \
           ((uint16_t *)(addr))[3] = HTONS((addr3)); \
           ((uint16_t *)(addr))[4] = HTONS((addr4)); \
           ((uint16_t *)(addr))[5] = HTONS((addr5)); \
           ((uint16_t *)(addr))[6] = HTONS((addr6)); \
           ((uint16_t *)(addr))[7] = HTONS((addr7)); \
} while(0)

/** Copy an IP address to another IP address
 *
 *  This is just a copy of 4 Bytes in the (default) case of IP V4 and 8 bytes
 *  in the case of IP V6 from \c src to \c dest.
 *
 *  No endianess is changed. The content of a type \ref uip_ipaddr_t variable
 *  is considered to be in network byte order.
 *
 * Example:
 \code
 uip_ipaddr_t ipaddr1, ipaddr2;

 uip_ipaddr(&ipaddr1, 192,16,1,2);
 uip_ipaddr_copy(&ipaddr2, &ipaddr1);
 \endcode
 *
 * \param dest pointer to the destination to copy to
 * \param src  pointer to the source where to copy from
 *
 * \hideinitializer
 */
#if !UIP_CONF_IPV6
#define uip_ipaddr_copy(dest, src) memcpy(dest, src, 4)
#else /* !UIP_CONF_IPV6 */
#define uip_ipaddr_copy(dest, src) memcpy(dest, src, sizeof(uip_ip6addr_t))
#endif /* !UIP_CONF_IPV6 */

/** Compare two IP addresses
 *
 * \param addr1 The first IP address.
 * \param addr2 The second IP address.
 * \hideinitializer
 * \internal warning: This (original) macro might fail with simple C pointer
 * expressions.
 */
#if !UIP_CONF_IPV6
#define uip_ipaddr_cmp(addr1, addr2) (((uint16_t*)addr1)[0] == ((uint16_t*)addr2)[0]\
    &&  ((uint16_t*)addr1)[1] == ((uint16_t*)addr2)[1])
#else /* !UIP_CONF_IPV6 */
#define uip_ipaddr_cmp(addr1, addr2) (memcmp(addr1, addr2, sizeof(uip_ip6addr_t)) == 0)
#endif /* !UIP_CONF_IPV6 */

/** Compare two IP addresses with netmasks
 *
 *  The masks are used to mask out the bits that are to be compared.
 *  \param addr1 The first IP address.
 *  \param addr2 The second IP address.
 *  \param mask  The netmask.
 *  \hideinitializer
 *  \internal warning: This (original) macro might fail with simple C pointer
 *                     expressions. (A.W.)
 */
#define uip_ipaddr_maskcmp(addr1, addr2, mask) \
                          (((((uint16_t *)addr1)[0] & ((uint16_t *)mask)[0]) == \
                            (((uint16_t *)addr2)[0] & ((uint16_t *)mask)[0])) && \
                           ((((uint16_t *)addr1)[1] & ((uint16_t *)mask)[1]) == \
                            (((uint16_t *)addr2)[1] & ((uint16_t *)mask)[1])))


/** Mask out the network part of an IP address
 *
 * Example:
 \code
 uip_ipaddr_t ipaddr1, ipaddr2, netmask;

 uip_ipaddr(&ipaddr1, 192,16,1,2);
 uip_ipaddr(&netmask, 255,255,255,0);
 uip_ipaddr_mask(&ipaddr2, &ipaddr1, &netmask);
 \endcode
 *
 * In the example above, the variable "ipaddr2" will contain the IP
 * address 192.168.1.0.
 *
 * \param dest Where the result is to be placed.
 * \param src  The IP address.
 * \param mask The netmask.
 *
 * \hideinitializer
 */
#define uip_ipaddr_mask(dest, src, mask) do { \
                     ((uint16_t *)dest)[0] = ((uint16_t *)src)[0] & ((uint16_t *)mask)[0]; \
                     ((uint16_t *)dest)[1] = ((uint16_t *)src)[1] & ((uint16_t *)mask)[1]; \
                  } while(0)

/** Pick the first octet of an IP address
 *
 * Example:
 \code
 uip_ipaddr_t ipaddr;
 uint8_t octet;

 uip_ipaddr(&ipaddr, 1,2,3,4);
 octet = uip_ipaddr1(&ipaddr);
 \endcode
 *
 * In the example above, the variable "octet" will contain the value 1.
 *
 * \hideinitializer
 */
#define uip_ipaddr1(addr) (htons(((uint16_t *)(addr))[0]) >> 8)

/** Pick the second octet of an IP address
 *
 * Example:
 \code
 uip_ipaddr_t ipaddr;
 uint8_t octet;

 uip_ipaddr(&ipaddr, 1,2,3,4);
 octet = uip_ipaddr2(&ipaddr);
 \endcode
 *
 * In the example above, the variable "octet" will contain the value 2.
 *
 * \hideinitializer
 */
#define uip_ipaddr2(addr) (htons(((uint16_t *)(addr))[0]) & 0xff)

/** Pick the third octet of an IP address
 *
 *  Example:
 \code
 uip_ipaddr_t ipaddr;
 uint8_t octet;

 uip_ipaddr(&ipaddr, 1,2,3,4);
 octet = uip_ipaddr3(&ipaddr);
 \endcode
 *
 * In the example above, the variable "octet" will contain the value 3.
 *
 * \hideinitializer
 */
#define uip_ipaddr3(addr) (htons(((uint16_t *)(addr))[1]) >> 8)

/** Pick the fourth octet of an IP address
 *
 * Example:
 \code
 uip_ipaddr_t ipaddr;
 uint8_t octet;

 uip_ipaddr(&ipaddr, 1,2,3,4);
 octet = uip_ipaddr4(&ipaddr);
 \endcode
 *
 * In the example above, the variable "octet" will contain the value 4.
 * \hideinitializer
 */
#define uip_ipaddr4(addr) (htons(((uint16_t *)(addr))[1]) & 0xff)

/** Convert 16-bit quantity from host byte order to network byte order
 *
 *  This macro is primarily used for converting constants from host
 *  byte order to network byte order. For converting variables to
 *  network byte order, use the htons() function instead.
 *  \hideinitializer
 */
#ifndef HTONS
#   if UIP_BYTE_ORDER == UIP_BIG_ENDIAN
#      define HTONS(n) (n)
#      define htons(n) (n)
#      define ntohs(n) (n)
#   else /* UIP_BYTE_ORDER == UIP_BIG_ENDIAN */
#      define HTONS(n) (uint16_t)((((uint16_t) (n)) << 8) | (((uint16_t) (n)) >> 8))
    uint16_t convert16endian(uint16_t value); //!< see we-aut_sys/utils.h
#      define htons(n) convert16endian(n) //!< swap endianess
#      define ntohs(n) convert16endian(n) //!< swap endianess
#   endif /* UIP_BYTE_ORDER == UIP_BIG_ENDIAN */
#else
#error "HTONS already defined!"
#endif /* HTONS */


/** @} */

/** Pointer to the application data in the packet buffer
 *
 *  This pointer points to the application data when the application is
 *  called. If the application wishes to send data, the application may
 *  use this space to write the data into before calling \ref uip_send().
 */
extern void *uip_appdata;

#if UIP_URGDATA > 0
/* uint8_t *uip_urgdata:
 *
 * This pointer points to any urgent data that has been received. Only
 * present if compiled with support for urgent data (UIP_URGDATA).
 */
extern void *uip_urgdata;
#endif /* UIP_URGDATA > 0 */


/**
 * \defgroup uipdrivervars  Variables for uIP device drivers
 * @{
 * \details
 * uIP has a few global variables that are used in device drivers for uIP.
 */

/** The length of the packet in the uip_buf buffer
 *
 * The global variable uip_len holds the length of the packet in the
 * uip_buf buffer.
 *
 * When the network device driver calls the uIP input function,  uip_len
 * should be set to the length of the packet in the uip_buf buffer.
 *
 * When sending packets, the device driver should use the contents of
 * the uip_len variable to determine the length of the outgoing packet.
 */
extern uint16_t uip_len;

#if UIP_URGDATA > 0
extern uint16_t uip_urglen, uip_surglen;
#endif /* UIP_URGDATA > 0 */


/** Representation of a uIP TCP connection
 *
 * The uip_conn structure is used for identifying a connection. All
 * but one field in the structure are to be considered read-only by
 * application software. The only exception is the field \c appstate whose
 * purpose is to let the application store application-specific state  for
 * the connection in a structure of appropriate type (e.g. \ref cliThr_data_t
 * or \ref ntpState_t) not longer than the specified
 *  \ref SIZE_OF_BIGGEST_APPSTATE "specified maximum" size.
 *
 * \internal   Adam Dunkels' programming style of having variables and type
 * the same name lead to many (tooling) troubles besides confusing human
 * beings. So this type was renamed from \c uip_conn to more clear
 * \c uipConn_t by A.W.
 */
struct uipConn_t {
  uip_ipaddr_t ripaddr; //!< The IP address of the remote host
  
  uint16_t lport;  //!< The local TCP port, in network byte order
  uint16_t rport;  //!< The remote TCP port, in network byte order
  
  uint8_t rcv_nxt[4]; //!< The sequence number that we expect to  receive next
  uint8_t snd_nxt[4]; //!< The sequence number that was last sent by us.
  uint16_t len;       //!< Length of the data that was previously sent.
  uint16_t mss;       //!< Current maximum segment size for the connection.
  uint16_t initialmss; //!< Initial maximum segment size for the connection
  uint8_t sa;         //!< Retransmission time-out calculation state variable
  uint8_t sv;         //!< Retransmission time-out calculation state variable
  uint8_t rto;        //!< Retransmission time-out
  uint8_t tcpstateflags; //!< TCP state and flags
  uint8_t timer;         //!< The retransmission timer
  uint8_t nrtx; //!< The number of retransmissions for the last segment sent

  uint16_t readIndex;   //!< for  the LAN (stream) input functions
  uint16_t writeIndex;  //!< for  the LAN (stream) output functions

  uip_tcp_appstate_t appstate;  //!< The application state
};


/** Pointer to the current TCP connection
 *
 * The uip_conn pointer can be used to access the current TCP connection
 *
 */
extern struct uipConn_t *uip_conn;

/** The array containing all uIP connections
 *
 *  <b> &nbsp; See also: &nbsp; </b> \ref  UIP_CONNS
 */
extern struct uipConn_t uip_conns[];

/** @} */

/**
 * \addtogroup uiparch
 * @{
 */

/** Calculate the Internet checksum over a buffer
 *
 *  This function adds the sum of all all 16-bit words in the buffer \c data
 *  to the parameter \c sum and returns the result.
 *
 *  \c sum and the returned result are in architecture byte order, being
 *  little endian (intel, ATmega, \ref intro_sec weAut_01).
 *
 *  The 16 bit words in the buffer on the other hand are big endian. If \c len
 *  is odd the last byte in the buffer data is hence the high byte of word
 *  with an imaginary low byte 0x00.
 *
 *  Another peculiarity of this type of unsigned checksum is the feeding
 *  back of a carry out of any 16 bit addition step.
 *
 *  This characteristics are adversarial to higher programming languages
 *  even to C. For \ref intro_sec "weAutSys" this function is significantly
 *  improved by inline assembler.
 *
 *  \param sum  the checksum's start value
 *  \param data a pointer to the buffer over which the checksum is to be
 *         computed
 *  \param len the length of the buffer
 *  \return the checksum
 */
uint16_t chksum(uint16_t sum, const uint8_t *data, uint16_t len);

/** @} */


#if UIP_UDP
/** Representation of a uIP UDP connection */
struct uip_udp_conn {
  uip_ipaddr_t ripaddr;   /**< The IP address of the remote peer. */
  uint16_t lport;        /**< The local port number in network byte order. */
  uint16_t rport;        /**< The remote port number in network byte order. */
  uint8_t  ttl;          /**< Default time-to-live. */

  /** The application state. */
  uip_udp_appstate_t appstate;
};

/** The current UDP connection */
extern struct uip_udp_conn *uip_udp_conn;

/** The list of all UDP connections
 *
 *  The number of possible UPP connections is fixed (\ref UIP_UDP_CONNS);
 *  the are held in this list (array). An entry with the local UDP port
 *  (\ref uip_udp_conn .lport)  0 is considered unused.
 */
extern struct uip_udp_conn uip_udp_conns[UIP_UDP_CONNS];
#endif /* UIP_UDP */


#if UIP_STATISTICS
/** The structure holding the TCP/IP statistics
 *
 *  Statistics are gathered if \ref UIP_STATISTICS is set to 1.
 */
struct uip_stats {
/** IP statistics */
  struct {
    uip_stats_t drop; /**< Number of dropped packets at the IP layer. */
    uip_stats_t recv;     /**< Number of received packets at the IP
			     layer. */
    uip_stats_t sent;     /**< Number of sent packets at the IP layer. */
    uip_stats_t vhlerr;   /**< Number of packets dropped due to wrong
			     IP version or header length. */
    uip_stats_t hblenerr; /**< Number of packets dropped due to wrong
			     IP length, high byte. */
    uip_stats_t lblenerr; /**< Number of packets dropped due to wrong
			     IP length, low byte. */
    uip_stats_t fragerr;  /**< Number of packets dropped since they
			     were IP fragments. */
    uip_stats_t chkerr;   /**< Number of packets dropped due to IP
			     checksum errors. */
    uip_stats_t protoerr; /**< Number of packets dropped since they
			     were neither ICMP, UDP nor TCP. */
  } ip;

/** ICMP statistics */
  struct {
    uip_stats_t drop;     /**< Number of dropped ICMP packets. */
    uip_stats_t recv;     /**< Number of received ICMP packets. */
    uip_stats_t sent;     /**< Number of sent ICMP packets. */
    uip_stats_t typeerr;  /**< Number of ICMP packets with a wrong  type. */
  } icmp;

/** TCP statistics  */
  struct {
    uip_stats_t drop;     /**< Number of dropped TCP segments. */
    uip_stats_t recv;     /**< Number of received TCP segments. */
    uip_stats_t sent;     /**< Number of sent TCP segments. */
    uip_stats_t chkerr;   /**< Number of TCP segments with a bad
			     checksum. */
    uip_stats_t ackerr;   /**< Number of TCP segments with a bad ACK
			     number. */
    uip_stats_t rst;      /**< Number of received TCP RST (reset) segments. */
    uip_stats_t rexmit;   /**< Number of retransmitted TCP segments. */
    uip_stats_t syndrop;  /**< Number of dropped SYNs due to too few
			     connections was available. */
    uip_stats_t synrst;   /**< Number of SYNs for closed ports,
			     triggering a RST. */
  } tcp;



#if UIP_UDP
/** UDP statistics */
  struct {
    uip_stats_t drop;     /**< Number of dropped UDP segments. */
    uip_stats_t recv;     /**< Number of received UDP segments. */
    uip_stats_t sent;     /**< Number of sent UDP segments. */
    uip_stats_t chkerr;   /**< Number of UDP segments with a bad checksum. */
  } udp;
#endif /* UIP_UDP */
};

/** The uIP TCP/IP statistics
 *
 *  This is the variable in which the uIP TCP/IP statistics are gathered.
 */
extern struct uip_stats uip_stat;

#endif

//--------------------------------------------------------------------------
//   All the stuff below this point is internal to uIP and should
//   not be used directly by an application or by a device driver.
//--------------------------------------------------------------------------

/*  Reason to call application function
 *
 *  When the application is called, uip_flags will contain the flags
 *  that are defined in this file.
 */
extern uint8_t uip_flags;

/* The following flags may be set in the global variable uip_flags
   before calling the application callback. The UIP_ACKDATA,
   UIP_NEWDATA, and UIP_CLOSE flags may both be set at the same time,
   whereas the others are mutually exclusive. Note that these flags
   should *NOT* be accessed directly, but only through the uIP
   functions/macros. */

#define UIP_ACKDATA   1     /* Signifies that the outstanding data was
			       acked and the application should send
			       out new data instead of retransmitting
			       the last data. */
#define UIP_NEWDATA   2     /* Flags the fact that the peer has sent
			       us new data. */
#define UIP_REXMIT    4     /* Tells the application to retransmit the
			       data that was last sent. */
#define UIP_POLL      8     /* Used for polling the application, to
			       check if the application has data that
			       it wants to send. */
#define UIP_CLOSE     16    /* The remote host has closed the
			       connection, thus the connection has
			       gone away. Or the application signals
			       that it wants to close the
			       connection. */
#define UIP_ABORT     32    /* The remote host has aborted the
			       connection, thus the connection has
			       gone away. Or the application signals
			       that it wants to abort the
			       connection. */
#define UIP_CONNECTED 64    /* We have got a connection from a remote
                               host and have set up a new connection
                               for it, or an active connection has
                               been successfully established. */

#define UIP_TIMEDOUT  128   /* The connection has been aborted due to
			       too many retransmissions. */

/** The actual uIP function which does all the work.
 *
 *  \param flag says on what reason this function is called
 */
void uip_process(uint8_t flag);

/* The following flags are passed as an argument to the uip_process()
   function. They are used to distinguish between the two cases where
   uip_process() is called. It can be called either because we have
   incoming data that should be processed, or because the periodic
   timer has fired. These values are never used directly, but only in
   the macros defined in this file. */
 
#define UIP_DATA 1 //!< flag to \ref uip_process(): incoming data in \ref uip_buf
#define UIP_TIMER 2 //!< flag to \ref uip_process(): periodic timer has fired
#define UIP_POLL_REQUEST 3 //!< flag to \ref uip_process(): poll a connection
#define UIP_UDP_SEND_CONN 4 //!< flag to \ref uip_process(): construct UDP datagram in \ref uip_buf
#if UIP_UDP
#define UIP_UDP_TIMER 5 //!< flag to \ref uip_process(): periodic timer has fired
#endif /* UIP_UDP */

/* The TCP states used in the uip_conn->tcpstateflags. */
#define UIP_CLOSED      0 //!< state value in uip_conn->tcpstateflags
#define UIP_SYN_RCVD    1 //!< state value in uip_conn->tcpstateflags
#define UIP_SYN_SENT    2 //!< state value in uip_conn->tcpstateflags
#define UIP_ESTABLISHED 3 //!< state value in uip_conn->tcpstateflags
#define UIP_FIN_WAIT_1  4 //!< state value in uip_conn->tcpstateflags
#define UIP_FIN_WAIT_2  5 //!< state value in uip_conn->tcpstateflags
#define UIP_CLOSING     6 //!< state value in uip_conn->tcpstateflags
#define UIP_TIME_WAIT   7 //!< state value in uip_conn->tcpstateflags
#define UIP_LAST_ACK    8 //!< state value in uip_conn->tcpstateflags
#define UIP_TS_MASK    15 //!< mask for state values in uip_conn->tcpstateflags
  
#define UIP_STOPPED    16 //!< extra state bit in uip_conn->tcpstateflags

/** The TCP and IP headers */
struct uip_tcpip_hdr {
#if UIP_CONF_IPV6
  /* IPv6 header. */
  uint8_t vtc,
    tcflow;
  uint16_t flow;
  uint8_t len[2];
  uint8_t proto, ttl;
  uip_ip6addr_t srcipaddr, destipaddr;
#else /* UIP_CONF_IPV6 */
  /* IPv4 header. */
  uint8_t vhl,
    tos,
    len[2],
    ipid[2],
    ipoffset[2],
    ttl,
    proto;
  uint16_t ipchksum;
  uint16_t srcipaddr[2],
    destipaddr[2];
#endif /* UIP_CONF_IPV6 */
  
  /* TCP header. */
  uint16_t srcport,
    destport;
  uint8_t seqno[4],
    ackno[4],
    tcpoffset,
    flags,
    wnd[2];
  uint16_t tcpchksum;
  uint8_t urgp[2];
  uint8_t optdata[4];
};

/** The ICMP and IP headers */
struct uip_icmpip_hdr {
#if UIP_CONF_IPV6
  /* IPv6 header. */
  uint8_t vtc,
    tcf;
  uint16_t flow;
  uint8_t len[2];
  uint8_t proto, ttl;
  uip_ip6addr_t srcipaddr, destipaddr;
#else /* UIP_CONF_IPV6 */
  /* IPv4 header. */
  uint8_t vhl,
    tos,
    len[2],
    ipid[2],
    ipoffset[2],
    ttl,
    proto;
  uint16_t ipchksum;
  uint16_t srcipaddr[2],
    destipaddr[2];
#endif /* UIP_CONF_IPV6 */
  
  /* ICMP (echo) header. */
  uint8_t type, icode;
  uint16_t icmpchksum;
#if !UIP_CONF_IPV6
  uint16_t id, seqno;
#else /* !UIP_CONF_IPV6 */
  uint8_t flags, reserved1, reserved2, reserved3;
  uint8_t icmp6data[16];
  uint8_t options[1];
#endif /* !UIP_CONF_IPV6 */
};


/** The UDP and IP headers */
struct uip_udpip_hdr {
#if UIP_CONF_IPV6
  /* IPv6 header. */
  uint8_t vtc,
    tcf;
  uint16_t flow;
  uint8_t len[2];
  uint8_t proto, ttl;
  uip_ip6addr_t srcipaddr, destipaddr;
#else /* UIP_CONF_IPV6 */
  /* IP header. */
  uint8_t vhl,
    tos,
    len[2],
    ipid[2],
    ipoffset[2],
    ttl,
    proto;
  uint16_t ipchksum;
  uint16_t srcipaddr[2],
    destipaddr[2];
#endif /* UIP_CONF_IPV6 */
  
  /* UDP header. */
  uint16_t srcport,
    destport;
  uint16_t udplen;
  uint16_t udpchksum;
};


/** The buffer size available for user data in the buffer \ref uip_buf
 *
 * This macro holds the available size for user data in \ref uip_buf .
 * It is intended to be used for checking bounds of available user data.
 * \hideinitializer
 */
#define UIP_APPDATA_SIZE (UIP_BUFSIZE - UIP_LLH_LEN - UIP_TCPIP_HLEN)
         //     566            620         14             40

#define UIP_PROTO_ICMP  1
#define UIP_PROTO_TCP   6
#define UIP_PROTO_UDP   17
#define UIP_PROTO_ICMP6 58

/* Header sizes. */
#if UIP_CONF_IPV6
#define UIP_IPH_LEN    40
#else /* UIP_CONF_IPV6 */
#define UIP_IPH_LEN    20    /* Size of IP header */
#endif /* UIP_CONF_IPV6 */
#define UIP_UDPH_LEN    8    /* Size of UDP header */
#define UIP_TCPH_LEN   20    /* Size of TCP header */
#define UIP_IPUDPH_LEN (UIP_UDPH_LEN + UIP_IPH_LEN) // Size of IP + UDP header
#define UIP_IPTCPH_LEN (UIP_TCPH_LEN + UIP_IPH_LEN) // Size of IP + TCP header
#define UIP_TCPIP_HLEN UIP_IPTCPH_LEN


#if UIP_FIXEDADDR
extern const uip_ipaddr_t uip_hostaddr, uip_netmask, uip_draddr;
#else /* UIP_FIXEDADDR */
extern uip_ipaddr_t uip_hostaddr, uip_netmask, uip_draddr;
#endif /* UIP_FIXEDADDR */



/** Representation of a 48-bit Ethernet address / MAC address  */
struct uip_eth_addr {
  uint8_t addr[6];
};

/** The (actual) MAC address
 *
 *  It can be changed on initialisation and has to be held consistent with
 *  the ENC28J60's MAC setting.
 *
 *  This is the (one) MAC address used for the Ethernet link. It is to be
 *  known by the Ethernet stack uIP and by the driver (of ENC28J60).
 *
 *  \internal This variable was originally named uip_ethaddr in uIP.
 */
extern struct uip_eth_addr actMACadd;


/** Calculate the Internet checksum over a buffer
 *
 * The Internet checksum is the one's complement of the one's complement sum
 * of all 16-bit words in the buffer.
 *
 *  See RFC1071.<br />
 *  See also the ENC28J60's data sheet. This Ethernet driver chip can
 *  calculate just this type of checksum as an extra service described under
 *  the the (miss-leading) header of "DMA functions". As a pity, uIP's
 *  internal proceedings calculate this checksum before the data will be put
 *  to the Ethernet chip respectively after the are copied out by the
 *  driver interface functions. So it would be a hard job with deep impact on
 *  uIP to have the chip calculate the IP checksum too en lieu of this
 *  function uip_chksum which is quite expensive on micro-controllers.
 *
 * \param buf A pointer to the buffer over which the checksum is to be
 * computed.
 * \param len The length of the buffer over which the checksum is to
 * be computed.
 *
 * \return The Internet checksum of the buffer.
 */
uint16_t uip_chksum(uint16_t *buf, uint16_t len);

/** Calculate the IP header checksum of the packet header in uip_buf
 *
 * The IP header checksum is the Internet checksum of the 20 bytes of
 * the IP header.
 *
 * \return The IP header checksum of the IP header in the uip_buf buffer.
 */
uint16_t uip_ipchksum(void);

/** Calculate the TCP checksum of the packet in uip_buf and uip_appdata
 *
 * The TCP checksum is the Internet checksum of data contents of the
 * TCP segment, and a pseudo-header as defined in RFC793.
 *
 * \return The TCP checksum of the TCP segment in uip_buf and pointed
 * to by uip_appdata.
 */
uint16_t uip_tcpchksum(void);

/** Calculate the UDP checksum of the packet in uip_buf and uip_appdata
 *
 * The UDP checksum is the Internet checksum of data contents of the
 * UDP segment, and a pseudo-header as defined in RFC768.
 *
 *  \note The uip_appdata pointer that points to the packet data may
 *    point anywhere in memory, so it is not possible to simply calculate
 *    the Internet checksum of the contents of the uip_buf buffer.
 *
 * \return The UDP checksum of the UDP segment in uip_buf and pointed
 *          to by uip_appdata.
 */
uint16_t uip_udpchksum(void);

/** One second tick for the uiP stack
 *
 *  This function has to be called by the runtime every second.
 *  (\ref intro_sec "weAutSys" does this in the 1s system thread.)
 *
 *  The implementation efficiently handles the ARP and DHCP timers.
 */
void uipSecTick(void);

/** @} */
/** @} */ // osSWinc
#endif /* __UIP_H__ */



