/**
 * \addtogroup osSWinc
 * @{
 * \addtogroup uipApp
 * @{
 * \defgroup uiparp uIP Address Resolution Protocol
 * @{
 * \details
 * The Address Resolution Protocol ARP is used for mapping between IP
 * addresses and link level addresses such as the Ethernet MAC addresses.
 * ARP uses broadcast queries to ask for the link level address of a known
 * IP address and the host which is configured with the IP address for which
 * the query was meant, will respond with its link level address.
 *
 * \note uIP's ARP implementation only supports Ethernet.
 *
 * The ARP code needs to know the MAC address of the Ethernet card in
 * order to be able to respond to ARP queries and to generate working
 * Ethernet headers. So \ref actMACadd has to be \ref setMACadd "set" to
 * the correct value before the ARP implementation can work.
 */

/**
 * \file uip_arp.h
 * Definitions for the ARP module
 *
 *  Adam Dunkels' original copyright notice <br /> &nbsp;
 *  Copyright (c) 2001-2003, Adam Dunkels.
 *  All rights reserved. <br />
 *  (contained in this file) still holds. For modifications:
 *
 *  <br /> &nbsp;
 * \ref copyLicNote " Copyright &copy; " 2011 Albrecht Weinert, Bochum
 * \author Adam Dunkels <adam@dunkels.com>
 * \author  Albrecht Weinert  &nbsp; <<a href="http://a-weinert.de">a-weinert.de</a>>
 * $Revision: 2 $
 * $Date: 2017-01-25 16:49:03 +0000 (Wed, 25 Jan 2017) $
 */
  

/* Copyright (c) 2001-2003, Adam Dunkels.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * This file is part of the uIP TCP/IP stack.
 *
 * Modified (adapted) by: Albrecht Weinert
 * For the modification: Copyright 2011 Albrecht Weinert <a-weinert.de>
 *
 * This file is also part of the \ref intro_sec "weAutSys" runtime system
 * by weinert - automation <weinert-automation.de>
 */

/** @} */
/** @} */
/** @} */ // osSWinc

#ifndef __UIP_ARP_H__
#define __UIP_ARP_H__

#include "uip.h"

/**
 * \addtogroup osSWinc
 * @{
 * \addtogroup uipApp
 * @{
 * \addtogroup uiparp
 * @{
 */

extern const struct uip_eth_addr broadcast_ethaddr;   //!< The broadcast MAC
extern const uint16_t broadcast_ipaddr[]; //!< The broadcast IP address


/** An ARP entry
 *
 *  This type describes the result of one address resolution. It is a
 *  relation of a MAC and an IP address plus a time stamp of when this
 *  relation was entered first or updated.
 */
struct arp_entry {
  uint16_t ipaddr[2]; //!< the IP (V4) address
  struct uip_eth_addr ethaddr; //!< the 48 bit MAC address

/** Time stamp and valid flag
 *
 *  The time stamp of this MAC/IP address relation is in terms of a ARP tick.
 *  The tick time (hence the unit) is usually 10 s. A longer value may
 *  be chosen to get longer ARP timeouts by an 8 bit tick and this 8 bit
 *  time stamp.
 *
 *  The value 0 is neither used as tick nor (hence) as time stamp value.
 *  \c time == 0 just means this entry is invalid.
 */
  uint8_t time;
};  // struct arp_entry

/** The ARP table
 *
 *  <b> &nbsp; See also: &nbsp; </b> \ref   UIP_ARPTAB_SIZE
 */
extern struct arp_entry arp_table[];
// static uint16_t ipaddr[2];

/** The Ethernet header */
struct uip_eth_hdr {

  struct uip_eth_addr dest; //!< source 48 bit MAC address
  struct uip_eth_addr src; //!< destination 48 bit MAC address
  uint16_t type; //!< The (big endian) type if > 1500
}; // struct uip_eth_hdr

#define UIP_ETHTYPE_ARP 0x0806  //!< Type ARP (in little endian)
#define UIP_ETHTYPE_IP  0x0800  //!< Type IP (V.4) (in little endian)
#define UIP_ETHTYPE_IP6 0x86dd  //!< Type IPV6 (in little endian)

#define ARP_REQUEST 1 //!< packet type request "which MAC has IP...?"
#define ARP_REPLY   2 //!< packet type answer "I have IP.. and MAC.."

#define ARP_HWTYPE_ETH 1 //!< ARP packet hardware Ethernet (the only supported one)


/** The ARP header */
struct arp_hdr {
  struct uip_eth_hdr ethhdr; //!< the Ethernet header
  uint16_t hwtype; //!< hardware type (only Ethernet supported in uIP)
  uint16_t protocol;
  uint8_t hwlen;
  uint8_t protolen;
  uint16_t opcode;
  struct uip_eth_addr shwaddr;
  uint16_t sipaddr[2];
  struct uip_eth_addr dhwaddr;
  uint16_t dipaddr[2];
};  // struct arp_hdr

/** The IP header */
struct ethip_hdr {
  struct uip_eth_hdr ethhdr; //!< the Ethernet header
  /* IP header. */
  uint8_t vhl,
    tos,
    len[2],
    ipid[2],
    ipoffset[2],
    ttl,
    proto;
  uint16_t ipchksum;
  uint16_t srcipaddr[2],
    destipaddr[2];
}; // struct ethip_hdr


/** Initialise the ARP module
 *
 *  This functions just clears all entries in the \ref arp_table "ARP table"
 *  and the ARP time tick.<br />
 *  It should be called before calling any other ARP functions.
 */
void uip_arp_init(void);

/** ARP processing for incoming IP packets
 *
 * This function should be called by the device driver when an IP packet
 * has been received. The function will check if the address is in the ARP
 * cache, and if so the ARP cache entry will be refreshed. If no ARP cache
 * entry was found, a new one is created.
 *
 * This function expects an IP packet with a prepended Ethernet header
 * in the uip_buf[] buffer, and the length of the packet in the global
 * variable \ref uip_len.
 */
void uip_arp_ipin(void);


/** ARP processing for incoming ARP packets
 *
 *  This function should be called by the device driver when an ARP packet has
 *  been received. The function will act differently depending on the ARP
 *  packet type: if it is a reply for a request that we previously sent out,
 *  the ARP cache will be filled in with the values from the ARP reply. If the
 *  incoming ARP packet is an ARP request for our IP address, an ARP reply
 *  packet is created and put into the uip_buf[] buffer.
 *
 *  When the function returns, the value of the global variable \ref uip_len
 *  indicates whether the device driver should send out a packet. If uip_len
 *  is zero, no packet should be sent. If uip_len is non-zero, it contains
 *  the length of the outbound packet that is present in the uip_buf[] buffer.
 *
 *  This function expects an ARP packet with a prepended Ethernet header in
 *  the uip_buf[] buffer, and the length of the packet in the global variable
 *  \ref uip_len.
 */
void uip_arp_arpin(void);

/** The ARP output prepare function
 *
 * It should be called before sending out an IP packet. The
 * function checks the destination IP address of the IP packet to see
 * what Ethernet MAC address that should be used as a destination MAC
 * address on the Ethernet.
 *
 * If the destination IP address is in the local network (determined
 * by logical ANDing of netmask and our IP address), the function
 * checks the ARP cache to see if an entry for the destination IP
 * address is found. If so, an Ethernet header is prepended and the
 * function returns. If no ARP cache entry is found for the
 * destination IP address, the packet in the uip_buf[] is replaced by
 * an ARP request packet for the IP address. The IP packet is dropped
 * and it is assumed that the higher level protocols (e.g., TCP)
 * eventually will retransmit the dropped packet.
 *
 * If the destination IP address is not on the local network, the IP
 * address of the default router is used instead.
 *
 * When the function returns, a packet is present in the uip_buf[]
 * buffer, and the length of the packet is in the global variable
 * \ref uip_len.
 */
void uip_arp_out(void);

/** Periodic ARP processing function
 *
 * This function performs periodic timer processing in the ARP module
 * and should be called at regular intervals. It will clear old (out-dated)
 * entries in the ARP table.
 *
 * The recommended frequency is every 10 seconds.
 *
 *  <b> &nbsp; See also: &nbsp; </b> \ref  UIP_ARP_MAXAGE
 */
void uip_arp_timer(void);

/** @} */
/** @} */
/** @} */ // osSWinc

#endif /* __UIP_ARP_H__ */
