// Copyright (c) 2011 Albrecht Weinert, Bochum *
// enc28j60.h 2 2017-01-25 16:49:03Z albrecht

/**
 * \addtogroup weAutSys
 * @{
 * \addtogroup llSySer
 * @{
 * \addtogroup ethCom
 * @{
 *
 * \file enc28j60.h
 * \ref intro_sec "weAutSys"' (weAut_01's)  28J60 Ethernet driver
 *
 * This file contains the definitions of (low level) system calls and services
 * for the  \ref intro_sec "weAutSys"' (weAut_01's)  28J60 Ethernet driver.
 *
 *  This is system software and must not be modified for user or
 *  application programs.
 *
 * This file is part of \ref intro_sec "weAutSys" &nbsp;
 *          <<a href="http://weinert-automation.de">weinert-automation.de</a>>
 *
 *  <br /> &nbsp;
 * \ref copyLicNote " Copyright &copy; " 2011 Albrecht Weinert, Bochum
 * \author  Albrecht Weinert  &nbsp;
 *                           <<a href="http://a-weinert.de">a-weinert.de</a>>
 * 2
 * 25.01.2017
 */
//  SVN last change by A. Weinert

/** @} */ // Doxygen addtogroup ethcom

#ifndef ENC_H
#define ENC_H

#include "we-aut_sys/spi.h"
#include "we-aut_sys/network.h"

/** \addtogroup ethCom
 * @{
 */

//  SPI Opcodes for ENC28J60
#define RCR_CMD    0x00 //!< Read control register command
#define RBM_CMD    0x3A //!< Read buffer memory command
#define WCR_CMD    0x40 //!< Write control register command
#define WBM_CMD    0x7A //!< Write buffer memory command
#define BFS_CMD    0x80 //!< Bit field set command
#define BFC_CMD    0xA0 //!< Bit field clear command
#define SYSRST_CMD 0xFF //!< Reset ENC28J60 command  \showinitializer

// Definition for ENC28J60's physical Registers (PHY Register 16 bit)
//      R,B    Register, bit number in register
#define PHCON1   0x00  //!< PHY control register 1
#define PHSTAT1  0x01  //!< PHY status register 1
#define   PFDPX 12 //!< full duplex capable (bit 12 is always set)
#define   PHDPX 11 //!< half duplex capable (bit 11 is always set)
#define   LLSTAT 2 //!< link was up continuously since last PHSTAT1 read
#define   JBSTAT 1 //!< transmission met jabber criteria since last PHSTAT1 read
#define PHID1   0x02   //!< PHY identification register 1
#define PHID2   0x03   //!< PHY identification register 2
#define PHCON2  0x10   //!< PHY control register 2
#define   FRCLNK 14 //!< make link up even when no partner station detected (Bit number)
#define   TXDIS  13 //!< disable twisted-pair transmitter hardware driver (Bit number)
#define   JABBER 10 //!< turn jabber correction off (Bit number)
#define   HDLDIS 8  //!< half duplex loopback disable (Bit number)
#define PHSTAT2   0x11  //!< PHY status register 2
#define   TXSTAT   13   //!<   PHY is currently transmitting (Bit number)
#define   RXSTAT   12   //!<   PHY is currently receiving (Bit number)
#define   COLSTAT  11   //!<   collision is occurring (Bit number)
#define   LSTAT    10   //!<   link is currently up (Bit number)
#define   DPXSTAT   9   //!<   configured for full-duplex (Bit number)
#define   PLRITY    5   //!<   polarity of TPIN is reversed
#define PHIE   0x12   //!< PHY interrupt enable register
#define PHIR   0x13   //!< PHY interrupt request register

/** LED Configuration Register
 *
 *  Eleven bits (1..11) of this register control the operation of the two
 *  LEDs that may be driven by the ENC28J60. They are usually, as in the
 *  weAut_01 board, integrated in the RJ45 jack.
 *
 *  Bits 15 and 14 must be written as 1; bits 13, 12 and 0 must be written
 *  as 0.
 *
 *  Both LEDs, referred to as LEDA and LEDB, have the same set of options
 *  coded to other bit positions. Available bit field options by macros
 *  are named accordingly beginning with LA or LB:
 *
 * Lx_DSC: &nbsp; Duplex status and collision activity (always stretched)<br />
 * Lx_LSTXRX: Link status transmission and receive activity (always stretched)<br />
 * Lx_LSRX:   Link status receive activity (always stretched)<br />
 * Lx_BSLOW:  Blink slow<br />
 * Lx_BFAST:  Blink fast<br />
 * Lx_LEDOFF:     LED off<br />
 * Lx_LEDON:      LED on<br />
 * Lx_RXTX:       Receive and Transmission activity<br />
 * Lx_DSTAT:      Duplex status<br />
 * Lx_LSTAT:      Link status<br />
 * Lx_COL: &nbsp; Collision activity (stretchable)<br />
 * Lx_RX: &nbsp;  Receive activity (stretchable)<br />
 * Lx_TX: &nbsp;  Transmit activity (stretchable)<br />
 *
 *  The configuration for LED on's stretching for stretchable events is
 *  common for both LEDs.
 *  \hideinitializer
 */
#define PHLCON 0x14

/*! \cond PRIVATE */
// LEDA Configuration bits
#define     LACFG3   11
#define     LACFG2   10
#define     LACFG1   9
#define     LACFG0   8

// LEDB Configuration bits
#define     LBCFG3   7
#define     LBCFG2   6
#define     LBCFG1   5
#define     LBCFG0   4
/*! \endcond */


// LEDA available options
#define LA_DSC      (1 << LACFG3) | (1 << LACFG2) | (1 << LACFG1) //!< see \ref PHLCON
#define LA_LSTXRX   (1 << LACFG3) | (1 << LACFG2) | (1 << LACFG0) //!< see \ref PHLCON
#define LA_LSRX     (1 << LACFG3) | (1 << LACFG2) //!< see \ref PHLCON
#define LA_BSLOW (1 << LACFG3) | (1 << LACFG1) | (1 << LACFG0) //!< see \ref PHLCON
#define LA_BFAST (1 << LACFG3) | (1 << LACFG1) //!< see \ref PHLCON
#define LA_LEDOFF   (1 << LACFG3) | (1 << LACFG0) //!< see \ref PHLCON
#define LA_LEDON (1 << LACFG3) //!< see \ref PHLCON
#define LA_RXTX     (1 << LACFG2) | (1 << LACFG1) | (1 << LACFG0) //!< see \ref PHLCON
#define LA_DSTAT (1 << LACFG2) | (1 << LACFG0) //!< see \ref PHLCON
#define LA_LSTAT (1 << LACFG2) //!< see \ref PHLCON
#define LA_COL      (1 << LACFG1) | (1 << LACFG0) //!< see \ref PHLCON
#define LA_RX    (1 << LACFG1) //!< see \ref PHLCON
#define LA_TX    (1 << LACFG0) //!< see \ref PHLCON


// LEDB available options
#define LB_DSC      (1 << LBCFG3) | (1 << LBCFG2) | (1 << LBCFG1) //!< see \ref PHLCON
#define LB_LSTXRX   (1 << LBCFG3) | (1 << LBCFG2) | (1 << LBCFG0) //!< see \ref PHLCON
#define LB_LSRX     (1 << LBCFG3) | (1 << LBCFG2) //!< see \ref PHLCON
#define LB_BSLOW (1 << LBCFG3) | (1 << LBCFG1) | (1 << LBCFG0) //!< see \ref PHLCON
#define LB_BFAST (1 << LBCFG3) | (1 << LBCFG1) //!< see \ref PHLCON
#define LB_LEDOFF   (1 << LBCFG3) | (1 << LBCFG0) //!< see \ref PHLCON
#define LB_LEDON (1 << LBCFG3) //!< see \ref PHLCON
#define LB_RXTX     (1 << LBCFG2) | (1 << LBCFG1) | (1 << LBCFG0) //!< see \ref PHLCON
#define LB_DSTAT (1 << LBCFG2) | (1 << LBCFG0) //!< see \ref PHLCON
#define LB_LSTAT (1 << LBCFG2) //!< see \ref PHLCON
#define LB_COL      (1 << LBCFG1) | (1 << LBCFG0) //!< see \ref PHLCON
#define LB_RX    (1 << LBCFG1) //!< see \ref PHLCON
#define LB_TX    (1 << LBCFG0) //!< see \ref PHLCON

// LED pulse stretch options
#define  TNSTRCH  2   //!<   LED pulse stretch, normal  40ms  (see \ref PHLCON)
#define  TMSTRCH  6   //!<   LED pulse stretch, medium  70ms (see \ref PHLCON)
#define  TLSTRCH  0x0A  //!<   LED pulse stretch, long 140ms (see \ref PHLCON)
#define  NOSTRCH  0     //!<   do not stretch LED events (see \ref PHLCON)

/** \name  Control registers
 * @{
 */

#define ESTAT 0x1D  //!< ETHERNET status register
#define     CLKRDY 0  //!< Clock (oscillator) is ready (\ref ESTAT bit number)
#define      TXABRT 1  //!< The transmit request was aborted (\ref ESTAT bit number)

#define ECON2 0x1E  //!< ETHERNET control register 2
#define  AUTOINC  7 //!< automatic increment and wrap of RAM buffer addresses (\ref ECON2 bit number)
#define  PKTDEC   6 //!< decrement (received) packets count (\ref ECON2 bit number)
#define     PWRSV 5 //!< switch the PHY interface off (save power when not needed; \ref ECON2 bit number)
#define     VRPS  3 //!< only if \ref PWRSV set voltage regulator to low current (\ref ECON2 bit number)

#define ECON1 0x1F  //!< ETHERNET control register 1
#define  TXRST 7  //!< Transmit only reset (\ref ECON1 bit number)
#define  RXRST 6  //!< Receive only reset (\ref ECON1 bit number)
#define  DMAST 5  //!< DMA operation (within internal RAM) start and busy bit (\ref ECON1 bit number)
#define  CSUMEN   4  //!< DMA hardware calculates checksums (\ref ECON1 bit number)
#define  TXRTS 3  //!< The transmit logic is attempting to transmit a packet (\ref ECON1 bit number)

/** Receive enable
 *
 *  Packets complying to the filter criteria set will be written into the
 *  receive buffer (\ref ECON1 bit number).
 */
#define  RXEN  2

#define EIE 0x1B  //!< Ethernet interrupt enable register
#define  INTIE  7 //!< \ref EIE bit number
#define  PKTIE  6 //!< \ref EIE bit number
#define  DMAIE  5 //!< \ref EIE bit number
#define  LINKIE 4 //!< \ref EIE bit number
#define  TXIE   3 //!< \ref EIE bit number
#define  TXERIE 1 //!< \ref EIE bit number
#define  RXERIE 0 //!< \ref EIE bit number

#define EIR 0x1C //!< Ethernet interrupt request register
#define  PKTIF  6 //!< \ref EIR bit number
#define  DMAIF  5 //!< \ref EIR bit number
#define  LINKIF 4 //!< \ref EIR bit number
#define  TXIF   3 //!< \ref EIR bit number
#define  TXERIF 1 //!< \ref EIR bit number
#define  RXERIF 0 //!< \ref EIR bit number

/**  @}  */ // name  Chip control registers


/** \name Symbolic names for direct register access
 *  @{
 *
 *  The lower 5 bits (0..31) give the register number in each bank.
 *  The bits 5 and 5 (mask 0x60) are used for the bank number (0..3 * 32).
 */

/**  The currently set bank of registers
 *
 *   Holds only the bank select bits, BSEL1:BSEL0 (mask 0x03) as in
 *   ECON1 register shifted 5 bits to left resp. * 32 (mask 0x60) as in
 *   the symbolic names.
 *
 *   Not to be modified by user software.
 */
extern uint8_t currentBank;

/* Bank 0 Registers
 *
 *  \hideinitializer
 */

/** The buffer read address register
 *
 *  The ENC28J60's 8K dual port RAM can be read using the (indirect)
 *  (13 bit) address in  \ref ERDPTH \ref ERDPTL .
 *
 *  If the \ref AUTOINC bit is set in \ref ECON2 this register is incremented
 *  after each read automatically wrapping around within the receive buffer
 *  range  \ref ERXSTL "ERXST .. ERXND".

 */
#define ERDPTL    (0x00)
#define ERDPTH    (0x01) //!< the buffer read address register  \sa ERDPTL

/** The buffer write address register
 *
 *  The ENC28J60's 8K dual port RAM can be written using the (indirect)
 *  (13 bit) address in  \ref EWRPTH : \ref EWRPTL  .
 *
 *  If the \ref AUTOINC bit is set in \ref ECON2 this address is incremented
 *  after each write automatically wrapping around within the RAM address
 *  range 0x0000 .. 0x1FFF .
 */
#define EWRPTL    (0x02)
#define EWRPTH    (0x03)  //!< the buffer write address register \sa EWRPTL


/** The write / transmit buffer range
 *
 *  The write buffer is located in the address range
 *  \ref ETXSTH : \ref ETXSTL .. \ref ETXNDH : \ref ETXNDL within the
 *  ENC28J60's 8K dual port RAM.
 *
 *  If the \ref AUTOINC bit is set in \ref ECON2 this is incremented after
 *  each read automatically wrapping around within range 0x0000 .. 0x1FFF
 *  i.e. the write wrapping is independent of this setting.
 *  \hideinitializer
 *  \sa ERXSTL
 */
#define ETXSTL    (0x04)
#define ETXSTH    (0x05) //!< the write buffer range  \sa ETXSTL
#define ETXNDL    (0x06) //!< the write buffer range  \sa ETXSTL
#define ETXNDH    (0x07) //!< the write buffer range  \sa ETXSTL

/** The read / receive buffer range
 *
 *  The read buffer is located in the address range
 *  \ref ERXSTH : \ref ERXSTL .. \ref ERXNDH : \ref ERXNDL within the
 *  ENC28J60's 8K dual port RAM.
 *
 *  If the \ref AUTOINC bit is set in \ref ECON2 this is incremented after
 *  each read automatically wrapping around within the receive buffer address
 *  range \ref ERXSTL "ERXST .. ERXND".
 *
 *  \hideinitializer
 *  \sa ETXSTL
 */
#define ERXSTL    (0x08)
#define ERXSTH    (0x09) //!< the write buffer range \sa ERXSTL
#define ERXNDL    (0x0A) //!< the write buffer range \sa ERXSTL
#define ERXNDH    (0x0B) //!< the write buffer range \sa ERXSTL


/** The receive  buffer  (already) read address register
 *
 *  The address \ref ERXRDPTH : \ref ERXRDPTL is the index in the ENC28J60's
 *  8K dual port RAM to where received bytes were processed by the software.
 *
 *  This location is never overwritten by the receiver.
 *  \sa  ERXWRPTL
  */
#define ERXRDPTL  (0x0C)
#define ERXRDPTH  (0x0D) //!< receive  buffer already read \sa ERXRDPTL


/** The receive  buffer write address register
 *
 *  The ENC28J60's 8K dual port RAM is written while receiving using the
 *  (indirect 13 bit) address in  \ref ERXWRPTH : \ref ERXWRPTL
 *  auto-incrementing it after each received byte.<br />
 *  The address \ref ERXRDPTH : \ref ERXRDPTL will never be overwritten
 */
#define ERXWRPTL  (0x0E)
#define ERXWRPTH  (0x0F) //!< receive  buffer write \sa ERXWRPTL

#define EDMASTL   (0x10) //!< DMA start low byte
#define EDMASTH   (0x11) //!< DMA start high byte
#define EDMANDL   (0x12) //!< DMA end low byte
#define EDMANDH   (0x13) //!< DMA end high byte
#define EDMADSTL  (0x14) //!< DMA destination low byte
#define EDMADSTH  (0x15) //!< DMA destination high byte
#define EDMACSL   (0x16) //!< DMA checksum low byte
#define EDMACSH   (0x17) //!< DMA checksum high byte

/* Bank 1 Register
 *
 *  \hideinitializer
 */
#define EHT0   (0x00 | 0x20) //!< hash table byte 0
#define EHT1   (0x01 | 0x20) //!< hash table byte 1
#define EHT2   (0x02 | 0x20) //!< hash table byte 2
#define EHT3   (0x03 | 0x20) //!< hash table byte 3
#define EHT4   (0x04 | 0x20) //!< hash table byte 4
#define EHT5   (0x05 | 0x20) //!< hash table byte 5
#define EHT6   (0x06 | 0x20) //!< hash table byte 6
#define EHT7   (0x07 | 0x20) //!< hash table byte 7
#define EPMM0  (0x08 | 0x20) //!< pattern match mask byte 0
#define EPMM1  (0x09 | 0x20) //!< pattern match mask byte 1
#define EPMM2  (0x0A | 0x20) //!< pattern match mask byte 2
#define EPMM3  (0x0B | 0x20) //!< pattern match mask byte 3
#define EPMM4  (0x0C | 0x20) //!< pattern match mask byte 4
#define EPMM5  (0x0D | 0x20) //!< pattern match mask byte 5
#define EPMM6  (0x0E | 0x20) //!< pattern match mask byte 6
#define EPMM7  (0x0F | 0x20) //!< pattern match mask byte 7
#define EPMCSL (0x10 | 0x20) //!< pattern match checksum low byte
#define EPMCSH (0x11 | 0x20) //!< pattern match checksum high byte
#define EPMOL  (0x14 | 0x20) //!< pattern match offset low byte
#define EPMOH  (0x15 | 0x20) //!< pattern match offset high byte
#define ERXFCON (0x18 | 0x20) //!< Ethernet receive filter control register
#define EPKTCNT (0x19 | 0x20) //!< Ethernet package count

/* Bank 2 Registers
 *
 *  \hideinitializer
 */
#define MACON1    (0x00 | 0x40)  //!< MAC control register 1
#define    TXPAUS  3 //!< MAC control register 1 (bit number)
#define    RXPAUS  2 //!< MAC control register 1 (bit number)
#define    PASSALL 1 //!< MAC control register 1 (bit number)
#define    MARXEN  0 //!< MAC control register 1 (bit number)
#define MACON2    (0x01 | 0x40)  //!< MAC control register 2 (inofficial)
#define MACON3    (0x02 | 0x40)  //!< MAC control register 3
#define    PADCFG2 7 //!< MAC control register 3 (bit number)
#define    PADCFG1 6 //!< MAC control register 3 (bit number)
#define    PADCFG0 5 //!< MAC control register 3 (bit number)

/** All short frames will be padded to 64 Bytes and have valid CRC appended
 *
 *  MAC control register 3 (bit pattern)
 */
#define        EXPSF64  (1<<PADCFG2) | (1<<PADCFG1) | (1<<PADCFG0)

/** All short frames will be padded to 60 Bytes and have valid CRC appended
 *
 *  MAC control register 3 (bit pattern)
 */
#define        EXPSF60  (1<<PADCFG0)

/** No handling short frames
 *
 *  MAC control register 3 (bit pattern)
 */
#define        NOSFPAD  0

/** Automatic handling short frames
 *
 *  The MAC will automatically detect VLAN Protocol frames (by 8100h type
 *  field). It pads short frames to 64 bytes if it's a valid VLAN frame or
 *  to 60 bytes otherwise. A valid CRC will then be appended.
 *
 *  MAC control register 3 (bit pattern)
 */
#define    DTCTVLAN (1<<PADCFG2) | (1<<PADCFG0)


#define    TXCRCEN 4 //!< MAC control register 3 (bit number)
#define    PHIDREN 3 //!< MAC control register 3 (bit number)
#define    HFRMEN  2 //!< MAC control register 3 (bit number)
#define    FRMLNEN 1 //!< MAC control register 3 (bit number)
#define    FULDPX  0 //!< MAC control register 3 (bit number)


#define MACON4       (0x03 | 0x40)  //!< MAC control register 4
#define    DEFER   6 //!< MAC control register 4 (bit number)
#define    BPEN    5 //!< MAC control register 4 (bit number)
#define    NOBKOFF 4 //!< MAC control register 4 (bit number)
#define MABBIPG   (0x04 | 0x40)  //!< back-to-back inter-packet gap
#define MAIPGL    (0x06 | 0x40)  //!< non back-to-back inter-packet gap  low
#define MAIPGH    (0x07 | 0x40)  //!< non back-to-back inter-packet gap  high
#define MACLCON1  (0x08 | 0x40)  //!< retransmission maximum (4 bit)
#define MACLCON2  (0x09 | 0x40)  //!< collision window (6 bit)
#define MAMXFLL      (0x0A | 0x40) //!<  maximum frame length low byte
#define MAMXFLH      (0x0B | 0x40) //!<  maximum frame length high byte
#define MICMD     (0x12 | 0x40)  //!<  MII command register
#define  MIISCAN  1  //!<  MII command register (bit number)
#define     MIIRD 0  //!<  MII command register (bit number)

/** MII register address register
 *
 *  This control register (5 bit) is used to address other
 *  register types. 16 bit registers are indirectly accessed via (this)
 *  address register and write respectively read registers.
 *  \sa MIWRL
 *  \sa MIWRH
 *  \sa MIRDL
 *  \sa MIRDH
 */
#define MIREGADR  (0x14 | 0x40)
#define MIWRL     (0x16 | 0x40) //!< MII write data low byte  \sa MIREGADR
#define MIWRH     (0x17 | 0x40) //!< MII write data high byte  \sa MIREGADR
#define MIRDL     (0x18 | 0x40) //!< MII read data low byte  \sa MIREGADR
#define MIRDH     (0x19 | 0x40) //!< MII read data high byte  \sa MIREGADR

// Bank 3 Registers
#define MAADR1    (0x04 | 0x60) //!< MAC address register (first)
#define MAADR2    (0x05 | 0x60) //!< MAC address register
#define MAADR3    (0x02 | 0x60) //!< MAC address register
#define MAADR4    (0x03 | 0x60) //!< MAC address register
#define MAADR5    (0x00 | 0x60) //!< MAC address register
#define MAADR6    (0x01 | 0x60) //!< MAC address register  (last)

#define EBSTSD    (0x06 | 0x60) //!< built-in self test fill seed
#define EBSTCON (0x07 | 0x60)   //!< Ethernet self test control register
#define EBSTCSL (0x08 | 0x60)   //!< built-in self test checksum low byte
#define EBSTCSH (0x09 | 0x60)   //!< built-in self test checksum high byte
#define MISTAT    (0x0A | 0x60)  //!< MII status register
#define     NVALID 2    //!< MII status register (bit number)
#define     SCAN   1    //!< MII status register (bit number)
#define     BUSY   0    //!< MII status register (bit number)
#define EREVID    (0x12 | 0x60)  //!< Ethernet revision ID (5 bit, R/O)
#define ECOCON (0x15 | 0x60)     //!< Clock output control
#define EFLOCON   (0x17 | 0x60)  //!< EFLOCON (Ethernet Flow Control
#define EPAUSL (0x18 | 0x60)     //!< Pause timer value low byte
#define EPAUSH (0x19 | 0x60)     //!< Pause timer value high byte


/** Mask for register number bits in symbolic register address
 *
 *  \showinitializer
 */
#define REG_MASK 0x1F

/** Mask for bank  bits in symbolic register address
 *
 *  \showinitializer
 */
#define REG_BANK_MASK 0x60

/** Mask for bank  bits in control register
 *
 *  The last 2 Bits of \ref ECON1 are the bank value (0..3).
 *
 *  \hideinitializer
 */
#define SPI_BANK_MASK 0x03

/** First number of common registers
 *
 *  Control registers live in 4 banks of each theoretically 32 registers.
 *
 *  The last "key" or "common" registers appear in all for banks.
 *  For them there's no need to even consider bank switching.
 *
 *  \showinitializer
 */
#define KEY_REGISTERS 0x1B

/**  @}  */ // name  Symbolic names for registers


/** \name  Per packet control bits for transmission settings
 * @{
 */
#define   PHUGEEN   3 //!< Huge frame enable (bit number)
#define   PPADEN    2 //!< per packet padding enable (bit number)
#define   PCRCEN    1 //!< per packet  CRC enable (bit number)
#define   POVERRIDE 0 //!< packet overrides (bit number)

/**  @}  */ // Per packet control bits for transmission settings



/** Receive status vector
 *
 *  Holds ENC's 4 byte receive status vector belonging to the last
 *  received package.
 */
extern uint8_t receiveStatVec[];

/** Transmission status vector
 *
 *  Used to see any failures during transmission. It is
 *  generated after the ESTAT.TXABRT flag is set by the ENC28J60.
 */
extern uint8_t transmitStatVec[];



//--- SPI commands --------------------

/** Read a control register
 *
 *  This function reads one of the ENC28J60's control registers.<br />
 *  The register's symbolic address consists of the register number (0..31
 *  within the register bank) + the bank number (0..3)*32. All symbolic
 *  control register names &mdash; i.e. macros &mdash; within ENC28J60.h
 *  are formed in this way.<br />
 *  The bank is switched only if necessary.
 *
 *  All control registers are 8 bit. They fall into one of three categories:
 *  ETH, MAC and MII. One step bit set and bit clear operations eliminate
 *  the need of going through read/modify/write &mdash; but only for the ETH
 *  type.
 *
 *  For reading the MAC or MII registers the answer will be prepended
 *  by an extra or dummy byte. In other words: The ENC28J60 needs 16 SPI
 *  clocks to read from an ETH register and  24 SPI clocks to read from a
 *  MAC/MII register. This (little) complication is, of course, handled by
 *  this function.
 *
 *  \param regAdd  the register's symbolic address
 *  \return         it's value
 *  \sa writeControlRegister
 *  \sa setBank
 *  \sa setBitfield
 */
uint8_t readControlRegister(uint8_t regAdd);

/** Write a control register
 *
 *  This function writes one of the ENC28J60's control registers.<br />
 *  The bank is switched if necessary.
 *
 *  \param regAdd   the register's symbolic address
 *  \param data     the new value
 *  \sa  readControlRegister
 */
void writeControlRegister(uint8_t regAdd, uint8_t data);


/** Set bits in a control register
 *
 *  This is an OR operation on the content of the register regAdd. For
 *  ETH type control registers it is done internally by the ENC28J60 saving
 *  a read modify write cycle.
 *
 *  For MII and MAC registers this internal OR is not available and hence the
 *  read and if necessary modify + write is done.
 *
 *  \param regAdd symbolic register address
 *  \param data the bits to be set
 *  \sa  readControlRegister
 *  \sa clearBitfield
 */
void setBitfield(uint8_t regAdd, uint8_t data);

/** Clear bits in a control register
 *
 *  This is an AND NOT on the content of the register regAdd. Conditions for
 *  a quick internal implementation by the ENC28J60 are the same as for
 *  \ref setBitfield.
 *
 *  \param regAdd symbolic address of the register to be modified
 *  \param data the bits to be reset
 *  \sa  readControlRegister
 *  \sa setBitfield
 */
void clearBitfield(uint8_t regAdd, uint8_t data);



/** Read one of ENC28J60's physical registers
 *
 *  This function handles the indirect access to those registers. All
 *  physical registers are 16 bit and can be read or written as whole words
 *  only.
 *
 *  The complication of being only indirectly accessible via \ref MIREGADR
 *  etc. is handled by this function.
 *
 *  \param regAdd  the register address
 *  \return        its value
 *  \sa writePhysicalRegister
 */
uint16_t readPhysicalRegister(uint8_t regAdd);

/** Write to one of ENC28J60's physical registers
 *
 * \param regAdd   the register address
 * \param data the (16 bit) value to be written
 * \sa readPhysicalRegister
 */
void writePhysicalRegister(uint8_t regAdd, uint16_t data);


/** Read n bytes from ENC28J60's memory buffer
 *
 *  This function reads a sequence of bytes from the dual port RAM
 *  indirectly addressed by the \ref ERXRDPTL "ERXRDPT" control registers.
 *  For n > 1 this usually makes sense only if auto-increment is on.
 *
 *  \param dest  pointer to buffer in RAM to write to
 *  \param n     number of bytes to read from ENC28J60's memory
 */
void readBufferMemory(uint8_t * dest, uint16_t n);


/** Write one byte to ENC28J60's memory buffer
 *
 *  This function writes to the dual port RAM byte indirectly addressed by
 *  the \ref EWRPTL "EWRPT" control registers.
 *
 *  \param data  to be written into the buffer memory
 */
void putBufferMemory(uint8_t data);

/** Write n bytes to ENC28J60's memory buffer
 *
 *  This function writes a sequence of bytes to the dual port RAM byte
 *  indirectly addressed by the \ref EWRPTL "EWRPT" control registers. This
 *  usually makes sense only if auto-increment is on.
 *
 *  \param source  pointer to data in RAM to be written to ENC28J60's memory
 *  \param n       number of bytes to be transferred from RAM to ENC
 */
void writeBufferMemory(uint8_t * source, uint16_t n);



//---  Special functions of the ENC28J60 driver

/**  Sets a register bank (in ECON1) if necessary for the given register
 *
 *  This function does nothing if no bank switch is needed to access the
 *  register regAdd
 *
 *  \param regAdd symbolic address of the register (bank bits relevant)
 *  \return the lower 5 bits (the "bankless" address) of regAdd
 */
uint8_t setBank(uint8_t regAdd);

/** Read the current MAC address
 *
 *  The MAC address will be read from the Ethernet driver ( ENC28J60) and
 *  put in the structure supplied.
 *
 *  \param mac the (6 byte) address structure
 */
void encGetMacAdd(eth_addr_t * mac);

/** Set the MAC address
 *
 *  The MAC address from the structure supplied will be written to the
 *  Ethernet driver ( ENC28J60).
 *
 *  This initialisation function is usually called as
 *  \code encSetMacAdd(& actMACadd); \endcode
 *  using the Ethernet stack's (uIP's) set MAC address for the driver.<br />
 *  Both the driver's and the stack's setting must be kept the same.<br />
 *
 *  The uIP's default / initialisation value for \ref  actMACadd is
 *  \code {0x40,0x1B,0x50,0xCA,0xFE,0x01} \endcode <br />
 *
 *  40 is &quot;locally assigned unicast 0&quot; <br />
 *  1B50 read "IBS 0" or Inbetriebsetzung (comissioning to use) 0
 *  is a fixed preset for first test or standalone use<br />
 *  CAFE01 in this context means "check all functional element of weAut01
 *
 *  As this is one of the most basic ENC28J60 initialisations this
 *  function checks the chips and the SPI's operation by re-reading
 *  the ENC's MAC address registers. 0 is returned only if this check
 *  passes.
 *  \note This test was implemented because the CLKRDY bit test, that would
 *  do the same basic hardware testing, simply does not work on most ENC
 *  cip releases and hence can as well be omitted.
 *
 *  \param mac the (6 byte) address structure
 *  \return 0 : OK; 1..6 the number of failed check reads
 */
uint8_t encSetMacAdd(eth_addr_t * mac);

/** Go to power safe mode
 *
 *  The ENC28J60 can be set to save power (if not used).
 */
void encEnablePowersave(void);

/** Leave the power safe mode
 *
 *  \sa encEnablePowersave
 */
void encDisablePowersave(void);


/** Chip select the ENC28J60  */
#define selectEnc() selectEtherChip()

/** De-select the ENC28J60
 *
 *  Does the de-select after (waiting for) send (on UART as SPI 2 to ) is
 *  completed.
 */
#define deSelectEnc()  deSelectSPI2()

/* * Force de-select the ENC28J60
 *
 *  De-selects ENC (and all devices on same SPI) anyhow. <br />
 *  This de-select does not regard any completion flags. It can (and should)
 *  be used when SPI completion is irrelevant or otherwise sure.
 * /
#define forceDeSelectEnc()  deSelectSPI2_impl()   */

/** Transmit and receive one byte to / from ENC28J60  */
#define tranceiveEnc(sendB)  spi2Tranceive(sendB)

/** Transmit two bytes and receive one byte to / from ENC28J60
 *
 *  \param sendB1 the byte to be sent first
 *  \param sendB2 the second byte to be sent
 *  \return the (second) byte read from SPI while transmitting \c sendB2
 */
#define tranceiveEnc2(sendB1, sendB2)  spi2Tranceive2(sendB1, sendB2)


/** @} */ // Doxygen addtogroup ethCom

//-- Initialisation functions ------------------------
/**
* \addtogroup  sysInit
* @{
*/


// Buffer limits

/** Receive buffer end
 *
 *  This should be an odd value due to ENC's bug B7 list point 11
 *  (called ERXND there). Due to a bug the receive buffer start will always
 *  be 0x0000 i.e. the start of ENC's dual port RAM.
 *
 *  This setting  (1A0D = 6669) leaves 5F2 = 1522 bytes for send buffer (and
 *  all else).
 *  This in-balance is due to not having multiple send packets in the ENC
 *  buffer which would hardly be organised with \ref uip "uIP" while having
 *  multiple incoming packets buffered is normal operation.
 *
 *  \showinitializer
 */
#define RXBUF_END  0x1A0D // 6669

/** Transmit buffer start
 *
 *  This is the minimal and the initial / default value for the start of a
 *  transmit package in the dual port RAM. It should be even (due to an ENC
 *  bug) and must be larger than \ref RXBUF_END as well as (considerably)
 *  smaller than the \ref  ENCBUF_END.
 *
 *  \showinitializer
 */
#define TXBUF_STRT 0x1A0E
#define ENCBUF_END 0x1FFF  //!< ENC's dual port RAM end  \showinitializer


/**  Set the sizes of ENC28J60's receive and transmit buffers
 *
 *   The ENC28J60's total buffer is 8191 bytes. This function divides this
 *   to a larger buffer for received packets at the begin and a smaller part
 *   as buffer for a transmit package (and maybe for all else).
 *
 *   Before changing the buffer limits this function disables and resets
 *   the receiver and transmitter and leaves it so (i.e. in reset state).
 *
 *   The receive packet index \ref actPackInd as well as read and write
 *   pointers are set accordingly.
 *
 *   /note This function uses macros (defines) to set the buffer sizes.
 */
void encSetBufferLimits(void);

/** Index of the actual receive packet in ENC's dual port memory
 *
 *  If at least one package has been received this will be the index to its
 *  begin. Most of the time this should be in accordance with
 *  \ref ERXRDPTL "ERXRDPT". Under most circumstances this must not be
 *  modified by user / application software.
 */
extern uint16_t actPackInd;

/** Reset the Ethernet controller ENC28J60
 *
 *  This function sends a (software) reset command to the ENC28J60. Due to
 *  an ENC28J60 bug (B7 errata point 1) this reset is quite hard as it
 *  scrambles the ENC's internal clocking. The only workaround is waiting
 *  1 ms before attempting further accesses.
 *
 *  This function does nothing else: a) no further ENC settings due to the
 *  bug and b) no (conditional) log message, as for usages early after
 *  reset, no logging resources (like UART e.g.) may be available.
 */
void encReset(void);

/** Initialise the Ethernet controller ENC28J60
 *
 *  This function initialises the Ethernet driver ENC28J60  but does
 *  not \ref encReset "reset" it. This function should be called not
 *  earlier than 1ms after resetting the ENC by hardware signal or software
 *  command.
 *
 *  This function checks ENC's (PHY) clock ready and does nothing but setting
 *  the bit \ref ENC_HWP in \ref  networkNotReady.<br />
 *  Otherwise the initialisation is done and all ENC related bits in
 *  \ref  networkNotReady will be cleared.
 */
void encInit(void);

/** @} */ // Doxygen addtogroup sysInit

#endif /* ENC_H */
/** @} */ // Doxygen addtogroup llSySer
/** @} */ // Doxygen addtogroup weAutSys
