// Copyright (c) 2012 Albrecht Weinert, Bochum
// modbus.h 2 2017-01-25 16:49:03Z albrecht

/**
 * \addtogroup weAutSys
 * @{
 * \addtogroup appLaySup
 * @{
 * \addtogroup modbusServer
 * @{
 *
 * \file modbus.h
 *
 * \ref intro_sec "weAutSys"' system calls, services and types for the Modbus server
 *
 * This file contains the definitions for \ref intro_sec "weAutSys"'
 * system calls, services and types for the build in Modbus server. It
 * implies an \ref uip "uIP" used as TCP/IP stack.
 *
 * This is system software and must not be modified for user or
 * application programs.
 *
 * This file is part of \ref intro_sec "weAutSys" &nbsp;
 *          <<a href="http://weinert-automation.de">weinert-automation.de</a>>
 *
 * &nbsp; \ref copyLicNote " Copyright &copy; " 2012  Albrecht Weinert, Bochum
 * \author  Albrecht Weinert  &nbsp;
 *                            <<a href="http://a-weinert.de">a-weinert.de</a>>
 * 2
 * 25.01.2017
 */
//  SVN last change by A. Weinert

/** @} */ // Doxygen addtogroup modbusServer
/** @} */ // Doxygen addtogroup appLaySup

#ifndef MODBUS_H
#define MODBUS_H

#include "we-aut_sys/network.h"
#include "we-aut_sys/timing.h"

/**
 * \addtogroup appLaySup
 * @{
 * \addtogroup modbusServer
 * @{
 */

/** @name Defines for Modbus function codes and the like
 *
 *  This is a subset of the function codes defined for the Modbus protocol.
 *  The others not put here are not implemented by this server.
 */
///@{

/** Modbus function code: write one bit output
 *
 *  This function is to write one single bit output. The lower 3 bits in the
 *  target address would be the bit number within a byte and the address / 8
 *  would be the byte number if mapped to bytes.
 *  \showinitializer
 */
#define WRITE_COIL 0x05

/** Modbus function code: read back bitwise output
 *
 *  The Modbus standard considers this function code's address as bit number
 *  and the quantity as bit count. This implementation will ignore the lower
 *  three bits of both values, thus accepting just whole (mapped) bytes as
 *  both source address and quantity. Requests with the lower three bits set
 *  will be rejected as erroneous.
 *
 *  Rationale: This implementations target processor has neither word
 *  operations nor multi-bit shifts. To implement requests not obeying the
 *  above restrictions would bring a resource burden not worth the wile.
 *
 *  The same restrictions aplly to \ref WRITE_COILS
 *  and \ref READ_DISCRETE_INPUTS. To set a single output bit use
 *  \ref WRITE_COIL and to manipulate a selection of bits within a word use
 *  \ref MASK_WRITE_REGISTER.
 *  \showinitializer
 */
#define READ_COILS 0x01

/** Modbus function code: write bitwise output
 *
 *  For the interpretation of target address and quantity and the respective
 *  restrictions see \ref READ_COILS.
 *  \showinitializer
 */
#define WRITE_COILS 0x0F

/** Modbus function code: read bitwise inputs
 *
 *  For the interpretation of source address and quantity and the respective
 *  restrictions see \ref READ_COILS.
 *  \showinitializer
 */
#define READ_DISCRETE_INPUTS 0x02

#define READ_INPUT_REGISTERS 0x04 //!< Modbus function code: read word (16 bit) inputs \showinitializer

#define READ_HOLDING_REGISTERS 0x03 //!< Modbus function code: read back word (16 bit) outputs \showinitializer
#define WRITE_HOLDING_REGISTERS 0x10 //!< Modbus function code: write word (16 bit) outputs \showinitializer
#define WRITE_HOLDING_REGISTER 0x06 //!< Modbus function code: write (one) word (16 bit) output \showinitializer

#define MASK_WRITE_REGISTER 0x16 //!< Modbus function code: and and or (one) word (16 bit) output \showinitializer

// No 0x14 (20) and no 0x15 (21) file record operations

/** Modbus function code: write than read word (16 bit) outputs
 *
 *  This is a time and communication saving combination of
 *  \ref WRITE_HOLDING_REGISTERS followed by \ref READ_HOLDING_REGISTERS.
 *  This combined output / input usually makes sense if the application maps
 *  all (four) Modbus data models to the same address space or uses Modbus'
 *  so called holding registers for all analogue, digital or other (process
 *  image) input and output.
 *  \showinitializer
 */
#define WRITE_READ_REGISTERS 0x17

// No 0x18 (24) read FiFo operations

#define MODB_EXC_FUNC 1 //!< Modbus exception: unimplemented function code \showinitializer
#define MODB_EXC_ADDR 2 //!< Modbus exception: illegal address\showinitializer
#define MODB_EXC_DATA 3 //!< Modbus exception: invalid data (or length)  \showinitializer
#define MODB_EXC_OPER 4 //!< Modbus exception: the (partly) performed operation failed \showinitializer


///@} // end member group Modbus command (state) and options (negotiation)


#define DISC_INP_BYMAPD 'i' //!< Modbus data model: Discrete inputs byte mapped 'i'
#define COILS_BYMAPD 'c'   //!< Modbus data model: Coils byte mapped 'c'
#define COIL_SINGLE 'C'   //!< Modbus data model: Single coil bit addressed 'C'
#define INP_REGISTERS 'r'  //!< Modbus data model: Input registers 'r'
#define HLD_REGISTERS 'h' //!< Modbus data model: Holding registers 'h'
#define HLD_MASK 'H'//!< Modbus data model: Single Holding register for mask 'H'


#define MODB_MBAB_LEN 7 //!< Modbus: length of the TCP/IP telegram start (MBAB header) \showinitializer
#define MODB_FCIND 7 //!< Modbus: index of function code in the TCP/IP telegram \showinitializer

/** Initialise the Modbus (server)
 *
 *  This function initialises the Modbus server. It is (indirectly) called
 *  whenever the Ethernet link goes up and hence usually after reset.
 *
 *  At present the implementation does nothing, but that may change in
 *  future releases.
 */
void modbusInit(void);

/** Register the application Modbus handler function
 *
 *  \param appModFun pointer to the function to register; NULL to de-register
 *  \sa modConfData_t
 */
void registerAppModFun(p2ptFunM appModFun);


/** Handle Modbus server events
 *
 *  This (system) function handles all Modbus events. And it is a as a
 *  protothread (part of) the Modbus server state machine. At present this
 *  feature is hardly used as this implementation does its work in one step
 *  as is recommended for a \ref registerAppModFun "registered" application
 *  / user software handler (see
 *  \ref modConfData_t "modConfData_t.appModFun").
 *
 *  This Modbus server implements the basic communication with the Modbus
 *  clients and calls the appropriate \ref registerAppModFun "registered"
 *  application software hooks.
 */
ptfnct_t modbusAppcall(void);


/** Handle Modbus server events
 *
 *  This application function handles the Modbus events prepared and
 *  forwarded by the \ref modbusAppcall "system handler". This function may
 *  or may be not implemented by user software. The name could be chosen
 *  freely; this declaration is for clarity's sake. If implemented it must
 *  be \ref registerAppModFun "registered" prior to incoming Modbus requests
 *  to be handled.
 *
 *  For this functions conditions see
 *  \ref modConfData_t "modConfData_t.appModFun",
 *  \ref modConfData_t "modConfData", \ref p2ptFunM and
 *  \ref funM_t "funM_t".
 *
 *  \param m  pointer to a Modbus handler (thread) state structure<br />
 *          All details of the request to handle are prepared there. The
 *          result has to be put there in the appropriate  fields respectively
 *          via the pointers provided.
 *
 *  \return \ref PT_WAITING, \ref PT_YIELDED, \ref PT_EXITED or \ref PT_ENDED
 *     <br /> If, as very strongly recommended, all
 *  work is done in one stint there's no need for thread organisation and the
 *  return value is one of the latter two.
 */
ptfnct_t appModFun(struct modThr_data_t * m);


/** @} */ // Doxygen addtogroup modbusServer
/** @} */ // Doxygen addtogroup appLaySup

#endif /* MODBUS_H */
/** @} */ // Doxygen addtogroup weAutSys
