// Copyright (c) 2012 Albrecht Weinert, Bochum
// smc2fs.h 2 2017-01-25 16:49:03Z albrecht

/**
 * \addtogroup weAutSys
 * @{
 * \addtogroup appLaySup
 * @{
 * \addtogroup filSyst
 * @{
 *
 * \file smc2fs.h
 *
 *  \ref intro_sec "weAutSys"' file system adaption to a small memory card
 *
 *  This file contains the definitions for \ref intro_sec "weAutSys"' adaption of ChaN's
 *  fatFS to the (low level) services to communicate with a
 *  \ref memCard_sec "small memory card" (SMC) via SPI.
 *
 *  This is system software and must not be modified for user or
 *  application programs.
 *
 * This file is part of \ref intro_sec "weAutSys" &nbsp;
 *          <<a href="http://weinert-automation.de">weinert-automation.de</a>>
 *
 * &nbsp; \ref copyLicNote " Copyright &copy; " 2012 Albrecht Weinert, Bochum
 * \author  Albrecht Weinert  &nbsp;
 *                            <<a href="http://a-weinert.de">a-weinert.de</a>>
 * 2
 * 25.01.2017
 */
//  SVN last change by A. Weinert

/** @} */ // Doxygen addtogroup filSyst
/** @} */ // Doxygen addtogroup appLaySup

#ifndef SMC2FS_H
#define SMC2FS_H

#include "fatFS/ff.h"

/**
 * \addtogroup appLaySup
 * @{
 * \addtogroup filSyst
 * @{
 */

/** File system object (the one for SMC)
 *
 *  This is the (one) file system state object used.
 *
 *  In  \ref intro_sec "weAutSys" / \ref intro_secH "weAut_01" and alike
 *  there can be just one or zero file-system for a small memory card (SMC
 *  card inserted.
 */
extern FATFS fileSystSMC;

/** Mount / initialise the SMC file system as drive 0
 *
 *  <b> &nbsp; See also: &nbsp; </b> \ref f_mount "fmount("0,
 *       \ref fileSystSMC "&fileSystSMC);"
 */
#define fMountSMC() f_mount(0, &fileSystSMC)

/** SMC file system (structure) locked for runtime / system use
 *
 *  \showinitializer
 *  \sa SMC_FS_CLIUSE
 */
#define SMC_FS_SYSTUSE 0xFB

/** SMC file system (structure) locked for (application) CLI
 *
 *  \showinitializer
 *  \sa SMC_FS_SYSTUSE
 */
#define SMC_FS_CLIUSE 0xCB

//-------------------------------------------------------------------------

/** Work space for file system operations (structure FS_WORK)
 *
 *  A structure of this type holds the state of the system or application
 *  software's ongoing file system operations.
 *
 *  \sa fsWork
 */
typedef struct {

/** Lock state of the file operation structure
 *
 *  A value 0 means that this structure is not locked or the implementation
 *  chooses not to use / implement that feature. <br />
 *  The (reserved) value \ref SMC_FS_SYSTUSE says that the
 *  \ref intro_sec "weAutSys" runtime itself is in the process of file system
 *  operations and uses this structure for state. This will only happen by
 *  respective (human) commands.
 *  \sa SMC_FS_CLIUSE
 *  \sa SMC_FS_SYSTUSE
 *  \sa  lockFsWorkFor
 *  \sa  unlockFsWorkFrom
 */
   uint8_t lock;

 /** A current file's or directory's name
  *
  *  This is a pointer to a null terminated string holding a path. Its
  *  content is read only for all file system functions.
  */
   char* path;

#if _USE_LFN
   char lfn[_MAX_LFN + 1]; //!<  the long file name
#endif

/** Result of file system operations
 *
 *  This variable is to hold the last main file system operation result.
 */
   FRESULT fRes;

/** Result of file system operations
 *
 *  This variable is to hold the last secondary file system operation result.
 */
   FRESULT fRes1;

/** Result of file system operations
 *
 *  This variable is to hold the last main tertiary file system operation
 *  result.
 */
   FRESULT fRes2;

/** Current file's or directory's information structure
 *
 *  This variable is to hold one file's state (over thread yields).
 */
   FILINFO filInf;

/** Current directory's status structure
 *
 *  This variable is to hold one directory's state (over thread yields).
 */
   DIR dir;

/** Current file's status structure
 *
 *  This variable is to hold one file's state (over thread yields).
 */
   FIL fil;

/** Current (32 bit) result or parameter value
 *
 */
   ucnt32_t szi32;

} FS_WORK;

//-------------------------------------------------------------------------


/** Work space for file system operations
 *
 *  This structure holds the state of the system or application software's
 *  ongoing file system operations. In a small realtime system file system
 *  operations on SMCs (interfaced via SPI) are relatively expensive in terms
 *  of both total processor time and size of indivisible (by yielding)
 *  sub-steps.
 *
 *  Hence file system operations should occur in one (well designed / planned)
 *  thread at a time only. This structure \ref fsWork of type \ref FS_WORK is
 *  to hold all file system related state and should remain singleton for
 *  said reasons.
 *
 *  As this variables singleton property is strongly recommended some
 *  functions and macros work directly on it. Their services are not available
 *  for further instances if application software chooses to have them.
 */
extern  FS_WORK fsWork;

/** Set the lock on \ref fsWork
 *
 *  \note This (macro) is an expression
 *
 *  \param ls lock signature (uint8_t); see \ref SMC_FS_SYSTUSE
 *  \return true if operation was feasible
 */
#define lockFsWorkFor(ls) ((fsWork.lock = fsWork.lock ? fsWork.lock : ls) == ls)

/** Unset the lock on \ref fsWork
 *
 *  \note This (macro) is an expression
 *
 *  \param ls lock signature (uint8_t) of current lock holder
 *  \return false if operation was feasible; true (holder) if other holds lock
 */
#define unlockFsWorkFrom(ls) (fsWork.lock = (fsWork.lock == ls) ? 0 : fsWork.lock)

/** Write file info data to standard output
 *
 *  This function outputs the information on the file \c filInf to the
 *  standard output stream.
 *
 *  The bits of the parameter \c inf control the information output: <br />
 *  Bit 7 (0x80) : output leading blank  <br />
 *  Bit 5 (0x20) : output last modified date / time <br />
 *  Bit 4 (0x10) : output attributes <br />
 *  Bit 3 (0x08) : output length <br />
 *
 *  Bit 0 (0x01) : output trailing new line
 *
 *  \param filInf pointer to the \ref FILINFO "file information" structure
 *  \param inf    the informations to output
 *  \return       0: nothing was output
 */
uint8_t stdPutFilInf(FILINFO * filInf, const uint8_t inf);

/** @} */ // Doxygen addtogroup filSyst
/** @} */ // Doxygen addtogroup appLaySup

/**
 * \addtogroup  utilHelp
 * @{
 * \addtogroup  utilForm
 * @{
 */

/** Format a FAT time
 *
 *  This function formats a 16 bit time delivered as (old fashined)
 *  \ref getFATtime "FAT time" structure. <br />
 *  The format will be \code "23:59:58" \endcode. <br />
 *  The length of the text delivered (to \c *s) is 9 characters including a
 *  trailing zero.
 *
 *  Note that the seconds in that (old DOS) FAT time format are 2 s
 *  increments so no odd value for seconds is possible.
 *
 *  \param s        start position within string resp. char array
 *  \param fatTime  the 16 bit FAT time structure
 *  \return pointer to the (0-) character following the formatted time
 *  \sa getFATtime()
 */
char * formFATtime(char * s, uint16_t fatTime);

/** Format a FAT date
 *
 *  This function formats a 16 bit delivered as (old fashined)
 *  \ref getFATtime "FAT date" structure (upper 16 bits 16..31). <br />
 *  The format (ISO 8601 / EN 28601) will be \code "2009-01-31" \endcode. <br />
 *  The length of the text delivered (to \c *s) is 11 characters including a
 *  trailing zero.
 *
 *  \param s        start position within string resp. char array
 *  \param fatDate  the 16 bit FAT date structure
 *  \return pointer to the (0-) character following the formatted date
 *  \sa getFATtime()
 */
char * formFATdate(char * s, uint16_t fatDate);

/** @} */ // Doxygen addtogroup utilForm
/** @} */ // Doxygen defgroup utilHelp

#endif /* SMC_H */
/** @} */ // Doxygen addtogroup weAutSys
