// Copyright (c) 2012 Albrecht Weinert, Bochum
// streams.h 2 2017-01-25 16:49:03Z albrecht

/**
 * \addtogroup weAutSys
 * @{
 * \addtogroup appLaySup
 * @{
 * \addtogroup stdStreams
 * @{
 */

/**
 *  \file streams.h
 *
 *  Streams (definitions)
 *
 *  This file contains the definitions for \ref intro_sec "weAutSys"' I/O via (standard i.e.
 *  stdio like) streams. It is system software and must not be modified for
 *  user or application programs.
 *
 *  This file is part of \ref intro_sec "weAutSys" &nbsp;
 *          <<a href="http://weinert-automation.de">weinert-automation.de</a>>
 *
 * &nbsp; \ref copyLicNote " Copyright &copy; " 2012 Albrecht Weinert, Bochum
 * \author  Albrecht Weinert  &nbsp;
 *                            <<a href="http://a-weinert.de">a-weinert.de</a>>
 * 2
 * 25.01.2017
 */
//  SVN last change by A. Weinert
/** @} */ // Doxygen addtogroup stdStreams
/** @} */ // Doxygen addtogroup appLaySup

#ifndef __STREAMS_H__
#define __STREAMS_H__

#include <avr/pgmspace.h>

/**
 * \addtogroup appLaySup
 * @{
 * \addtogroup stdStreams
 * @{
 */

// -- extra types and definitions for common stream handling

#define AVAILABLE_UNKNOWN 0xFFFF //!< Available input unknown \showinitializer


/** @name Common stream handling definitions
 * @{
 */

/** Function pointer / type: number of characters buffered for input
 *
 *  A function of this type returns the number of bytes that can be fetched
 *  from the stream's input without getting EOF.
 *
 *  For buffered devices with known state the answer will be correct.
 *
 *  In other cases the number of characters readable in one run without EOF
 *  is unknown. Then a large number with bit 15 set
 *  (0xFFFF \ref AVAILABLE_UNKNOWN) is returned.
 *  \sa inBuffered
 */
typedef uint16_t (*str2InBufferd)(const FILE * stream);

/** Number of characters buffered for input
 *
 *  This is the common function implementing the standard behavior. It
 *  delegates all else cases (late binding) to the stream type specific
 *  function if such is set in the stream (type)
 *  \ref streamClassDescript_t "specific data".
 *
 *  For this function's behavior see the
 *  \ref str2InBufferd "description" in its
 *  \ref str2InBufferd "pointer / type".
 */
uint16_t inBufferd(const FILE * stream);


/** Function pointer / type: buffer space available for output
 *
 *  A function of this type returns the number of bytes that can currently
 *  be output to the stream in one run without waiting or overflow.
 *
 *  For buffered devices with known state the answer will be correct.
 *
 *  In other cases a large number with bit 15 set
 *  (0xFFFF \ref AVAILABLE_UNKNOWN) is returned.
 *  \sa str2CheckOutSpace()
 *  \sa outSpace
 */
typedef uint16_t (*str2OutSpace)(const FILE * stream);

/** Buffer space available for output
 *
 *  This is the common function implementing the standard behavior. It
 *  delegates all else cases (late binding) to the stream type specific
 *  function if such is set in the stream (type)
 *  \ref streamClassDescript_t "specific data".
 *
 *  For this function's behavior see the
 *  \ref str2OutSpace() "description" in its
 *  \ref str2OutSpace()"pointer / type".
 */
uint16_t outSpace(const FILE * stream);


/** Function pointer / type: checking the space available for output
 *
 *  A function of this type checks the requested space against the buffer
 *  space \ref  outSpace "available".
 *
 *  If the stream is, for example the UART, its function checks the requested
 *  space against the UART \ref UART_OUT_BUF_CAP "buffer size" or available
 *  \ref uartOutSpace() "space". Return values are 0, 2 or 4.
 *
 *  If the stream is the LAN respectively the currently handled IP connection
 *  the respective function checks the requested space against the actual
 *  connection's \ref lanStreamsOutSpace() "remaining output space". Return
 *  values are 0, 3 or 4.
 *
 *  For other devices with known state the correct / appropriate answer is
 *  given. <br />
 *  For an unknown device 1 is returned and for no (NULL) device 4.
 *
 *  \param reqSpace the number of bytes requested for output in one complete
 *         step
 *  \param stream the stream <br /> In Implementations for one single
 *  (singleton) concrete stream, this parameter is usually ignored.
 *  \return <br />
 *       0: required space is available.
 *          I.e. go ahead with planned output.<br />
 *       1: may be available, but can't be determined for
 *          the given device / stream. I.e. blindly go ahead or
 *          give up;<br />
 *       2: is not available, but may become (automatically in the future,
 *          i.e. yielding / waiting is OK;<br />
 *       3: is not available, but may become by own activity (flushing,
 *          sending). I.e. just yielding / waiting would be endless.
 *       4: will never be available. I.e. give up or request less space.
 *  \sa PT_YIELD_OUT_SPACE
 *  \sa checkOutSpace
 */
typedef uint8_t (*str2CheckOutSpace)(uint16_t reqSpace, const FILE * stream);

/** Checking the space available for output
 *
 *  This is the common function implementing the standard behavior. It
 *  delegates all else cases (late binding) to the stream type specific
 *  function if such is set in the stream (type)
 *  \ref streamClassDescript_t "specific data".
 *
 *  For this function's behavior see the
 *  \ref str2CheckOutSpace() "description" in its
 *  \ref str2CheckOutSpace()"pointer / type".
 */
uint8_t checkOutSpace(uint16_t reqSpace, const FILE * stream);


/** Set standard I/O and exit or yield for requested output space
 *
 *  This is a macro to be used inside a protothread.
 *
 *  It sets a re-entry point for the next schedule of \c pt (after eventually
 *  yielding or waiting). Then it switches the standard output to \c streams
 *  and exits the calling thread \c pt (returning \ref PT_EXITED) if that is
 *  not possible (cause \c streams is \c NULL e.g.).
 *
 *  If the \ref switchStdStreams "switching to / opening of" \c streams was
 *  feasible this macro then <ul>
 *  <li> does nothing else (just yields once), if the requested space is
 *       immediately available, or if the availability can't be
 *       determined. Anyhow
 *       \ref PT_YIELDED is returned at first entry. </li></ul>
 *  Otherwise the \ref switchStdStreams "opening of" \c streams is
 *  \ref streamsClose "un-done" and <ul>
 *  <li> \ref PT_YIELDED "yields" the protothread \c pt, if the requested
 *       space may become available in the future by just waiting </li>
 *  <li> just \ref PT_WAITING "blocks" the protothread \c pt, if the requested
 *       space will not become available by yielding / waiting but may be by
 *       other's activities.</li>
 *  <li> \ref PT_EXITED "exits" and resets the protothread, if the requested
 *       space will never become  available for the current (set) standard
 *       output stream (connection) or if \c streams is not usable or
 *       \c NULL </li></ul>
 *
 *  \param pt         the pointer to the protothread control structure.
 *  \param requSpace  the requested output space
 *  \param streams    the stream to use and switch stdout to
 *  \hideinitializer
 *  \sa checkOutSpace()
 *  \sa PT_OR_YIELD_REENTER
 */
#define PT_YIELD_OUT_SPACE(pt, requSpace, streams) \
  do { \
     PT_YIELD_FLAG = 0; \
     LC_SET((pt)->lc); \
     if (!switchStdStreams(streams)) { PT_INIT(pt); return PT_EXITED; } \
     const uint8_t cS = checkOutSpace((uint16_t)(requSpace), streams); \
     if (cS >= 2) { \
        streamsClose(); \
       if (cS == 2) return PT_YIELDED; \
       if (cS == 4) { PT_INIT(pt); return PT_EXITED; }     \
       return PT_WAITING; \
     } \
     if (!PT_YIELD_FLAG) return PT_YIELDED; \
  } while(0)



/** Set standard I/O and exit or wait for requested output space
 *
 *  This is a macro to be used inside a protothread.
 *
 *  It sets a re-entry point for the next schedule of \c pt (after eventually
 *  yielding or waiting). Then it switches the standard output to \c streams
 *  and exits the calling thread \c pt (returning \ref PT_EXITED) if that is
 *  not possible (cause \c streams is \c NULL e.g.).
 *
 *  If the \ref switchStdStreams "switching to / opening of" \c streams was
 *  feasible this macro then <ul>
 *  <li> does nothing else (just goes ahead), if the requested space is
 *       immediately available, or if the availability can't be
 *       determined. </li></ul>
 *  Otherwise the \ref switchStdStreams "opening of" \c streams is
 *  \ref streamsClose "un-done" and <ul>
 *  <li> \ref PT_YIELDED "yields" the protothread \c pt, if the requested
 *       space may become available in the future by just waiting </li>
 *  <li> just \ref PT_WAITING "blocks" the protothread \c pt, if the requested
 *       space will not become available by yielding / waiting but may be by
 *       other's activities.</li>
 *  <li> \ref PT_EXITED "exits" and resets the protothread, if the requested
 *       space will never become  available for the current (set) standard
 *       output stream (connection) or if \c streams is not usable or
 *       \c NULL </li></ul>
 *
 *  \param pt         the pointer to the protothread control structure.
 *  \param requSpace  the requested output space
 *  \param streams    the stream to use and switch stdout to
 *  \hideinitializer
 *  \sa checkOutSpace()
 *  \sa PT_OR_YIELD_REENTER
 */
#define PT_WAIT_OUT_SPACE(pt, requSpace, streams) \
  do { \
     PT_YIELD_FLAG = 0; \
     LC_SET((pt)->lc); \
     if (!switchStdStreams(streams)) { PT_INIT(pt); return PT_EXITED; } \
     const uint8_t cS = checkOutSpace((uint16_t)(requSpace), streams); \
     if (cS >= 2) { \
       streamsClose(); \
       if (cS == 2) return PT_YIELDED; \
       if (cS == 4) { PT_INIT(pt); return PT_EXITED; }     \
       return PT_WAITING; \
     } \
  } while(0)


/** Function pointer / type:  writing a RAM string to output
 *
 *  A function of this type writes the content of the string \c src to the
 *  \c stream. It will never block.
 *
 *  It returns the number of characters transferred. If not enough space is
 *  available the returned value will be smaller than the length of the
 *  string \c src. It may even be 0. In the latter case nothing was output.
 *
 *  Hence this operation should be (indirectly) guarded by \ref outSpace() or
 *  \ref PT_YIELD_OUT_SPACE .
 *
 *  The output stops after at the first 0 in src. By that this is a string
 *  output function. To send a (binary) 0 use \ref stdPutC() or \c fputc().
 *
 *  \param src    the characters to be output (pointer to RAM not NULL !)
 *  \param stream the destination stream
 *  \return the number of characters output
 *  \sa putSt
 */
typedef int  (*str2PutS)(char * src, const FILE * stream);


/** Write a RAM string to output
 *
 *  This is the common function implementing the standard behavior. It
 *  delegates all normal cases (late binding) to the stream type specific
 *  function if such is set in the stream (type)
 *  \ref streamClassDescript_t "specific data"; if no such function is set
 *  it uses \c fputs().
 *
 *  For this function's behavior see the
 *  \ref str2PutS() "description" in its
 *  \ref str2PutS() "pointer / type".
 */
int  putSt(char * src, const FILE * stream);

/** Function pointer / type: writing a flash memory string to output
 *
 *  A function of this writes the content of the string \c src to the
 *  \c stream. It will never block.
 *
 *  It returns the number of characters transferred. If not enough space is
 *  available the returned value will be smaller than the length of the
 *  string \c src. It may even be 0. In the latter case nothing was output.
 *
 *  Hence this operation should be (indirectly) guarded by \ref outSpace() or
 *  \ref PT_YIELD_OUT_SPACE .
 *
 *  The output stops after at the first 0 in src. By that this is a string
 *  output function. To send a (binary) 0 use \ref stdPutC() or \c fputc().
 *
 *  \param src    the characters to be output
 *                (pointer to flash memory not NULL !)
 *  \param stream the destination stream
 *  \return the number of characters output
 *  \sa putSt_P
 */
typedef int (*str2PutS_P)(prog_char * src, const FILE * stream);

/** Write a flash memory string to output
 *
 *  This is the common function implementing the standard behavior. It
 *  delegates all else cases (late binding) to the stream type specific
 *  function if such is set in the stream (type)
 *  \ref streamClassDescript_t "specific data"; if no such function is
 *  set it uses \c fputs_P().
 *
 *  For this function's behavior see the
 *  \ref str2PutS_P() "description" in its
 *  \ref str2PutS_P() "pointer / type".
 */
int putSt_P(char const * src, const FILE * stream);


/** Extra stream (type) specific data
 *
 *  By extensions to the stdio type \c FILE  \ref intro_sec "weAutSys" will
 *  handle textual output and input in a common and extensible way as well as
 *  late binding of stream type specific functions.
 *
 *  For additional user / application streams do something like \code
  // have seven specialised functions defined for my extra stream  myStreams
FILE myStreams = FDEV_SETUP_STREAM(myPutChar, myGetChar, _FDEV_SETUP_RW );

struct streamClassDescript_t myStreamsData = {
      .inBufferdFunc = (str2InBufferd)myInBufferd,
      .outSpaceFunc =  (str2OutSpace)myStreamsOutSpace,
      .checkOutSpaceFunc = (str2CheckOutSpace)myCheckOutSpace,
      .putSfunc    = (str2PutS)myPutSt,
      .putS_P_func = (str2PutS_P)myPutSt_P()
};

   // in an initialisation block
   fdev_set_udata(&myStreams, &myStreamsData);
\endcode
 *  That's it. The basic I/O functions \c myPutChar and/or \c myGetChar are,
 *  of course, mandatory. All others could be omitted if the respective
 *  common / default behaviour is sufficient. For performance reasons
 *  those omissions are seldom recommended.
 *
 *  The common behaviour is defined by \ref inBufferd, \ref outSpace,
 *  \ref checkOutSpace, \ref putSt and \ref putSt_P. They can be called
 *  for the exemplary \c myStream after
 *  \ref switchStdStreams "switching" to it.
 */
struct streamClassDescript_t {

/** This stream's input buffer available function
 *
 *  Let or set to null if no number in a range 0..buffer size can be
 *  determined.
 *
 *  The common function \ref  inBufferd() will return
 *  \ref AVAILABLE_UNKNOWN then.
 */
   str2InBufferd inBufferdFunc;

/** This stream's output buffer available function
 *
 *  Let or set to null if no number in a range  0..buffer size can be
 *  determined.
 *
 *  The common function
 *  \ref  outSpace() will return \ref AVAILABLE_UNKNOWN then.
 */
   str2OutSpace  outSpaceFunc;


/** This stream's check enough output buffer available function
 *
 *  Let or set to null if this can't never be determined (i.e. the return
 *  value would always be 1).
 *
 *  The common function \ref  outSpace() will return 1 then.
 */
   str2CheckOutSpace checkOutSpaceFunc;

/** This stream's output a RAM string function
 *
 *  If let or set \c NULL the common function \ref putSt() will
 *  use \c fputs().
 */
   str2PutS putSfunc;


/** This stream's output a string in flash / program memory function
 *
 *  If let or set \c NULL the common function \ref putSt_P() will
 *  use \c fputs_P().
 */
   str2PutS_P putS_P_func;

}; // streamClassDescript_t

/** Extra stream (type) specific data (\ref streamClassDescript_t "structure")
 *
 * \internal this is only for a Doxygen bug that lets
 *  struct streamClassDescript_t fall out of the member group
 */
typedef struct streamClassDescript_t streamClassDescript_s;

///@} // end member group Common stream handling definitions


//======================  common streams (since June 12) end  ==============

// -- std I/O stream switching  --------------------------------------------

/** stdout is the serial output (UART)
 *
 *  This evaluates to true, if stdout goes to the UART.
 */
#define SER_IS_STDOUT (stdout == &serStreams)


/** stdout is the LAN stream output (Ethernet, uIP)
 *
 *  This evaluates to true, if stdout goes via uIP to the Ethernet.
 */
#define LAN_IS_STDOUT (stdout == &lanStreams)


/** Initialise the standard streams
 *
 *  This function is usually called once at system start up but might be
 *  called any time to change or assure the standard streams.
 *
 *  Anyway it will switch to the (then) actual standard streams acting
 *  hence effectively as \ref streamsClose(). (Additionally all extra data
 *  for pre-defined streams will be reset / refreshed.) Do not call when
 *  these (side-) effects may be disturbing.
 *
 *  If the parameter is NULL and standard streams were set already they are
 *  just kept. Otherwise \ref nulStreams are used as default.
 *
 *  \param initStream pointer to initial stream for standard streams
 *  \sa serStreams
 *  \sa lanStreams
 *  \sa nulStreams
 */
void initStdStreams(FILE * initStream);

/** Switch the standard streams to another stream
 *
 *  This function sets the standard streams stdin/out/err to the
 *  \c toStream. If \c toStream is NULL nothing is done.
 *
 *  \param toStream pointer streams to switch the standard streams to
 *  \return 0: toStream is NULL; 1: stdIO is or was set to toStream
 */
uint8_t switchStdStreams(FILE * toStream);

/** Close other standard streams
 *
 *  This function just reverses the (stream switching) effect of
 *  \ref switchStdStreams() by setting back all standard streams to their
 *  \ref initStdStreams "initial standard" source and destination.
 */
void streamsClose(void);


//----    serial (UART) stream  -------------------------------------------

/// \name The serial streams
/// @{

/** The stream "device" connected to the serial IO / UART(0)
 *
 *  On (default) \ref initStdStreams "initialisation" this stream will be
 *  used for the standard streams.
 */
extern FILE serStreams;

/** Open the serial (UART) streams
 *
 *  This function sets the standard streams stdin/out/err to the
 *  \ref serStreams if (and only if) not yet so set.<br />
 *  This "switching" will be rescinded by  \ref streamsClose() .
 */
void serStreamsOpen(void);

/** The buffer space available for serial output
 *
 *  Same as \ref uartOutSpace (other common signature).
 */
uint16_t serStreamsOutSpace(FILE * streams);

/** Check the space available for serial output
 *
 *  \sa str2CheckOutSpace
 */
uint8_t serCheckOutSpace(uint16_t reqSpace, FILE * streams);

/** Put one byte to serial output
 *
 *  This function never blocks. It returns the byte \c c output if the
 *  operation was feasible or -1 resp. EOF if not. In the latter case
 *  \c c is not output.
 *
 *  Hence this operation should be (indirectly) guarded by \ref uartOutSpace().
 *
 *  \param c the character to be output
 *  \param stream unused, irrelevant (signature required by stdio)
 *  \return c in case of success, EOF if output buffers full.
 */
int serPutChar(char c, FILE *stream); //  __attribute__ ((pure));


/// @}  // end member group serial streams


//--------------------  simple std out functions -------------------------

#define USE_AVR_GCC_FILE 1 //!< use AVR FILE type internals \showinitializer


/// \name Recommended standard I/O functions
/// @{

/** Write a character to \c stdout
 *
 *  If \ref USE_AVR_GCC_FILE is defined with a not 0 value this function
 *  calls the stdout stream's character output implementation directly.
 *
 *  Without the direct call optimisation the following holds:
 *
 *  If the stdout is the UART or the current LAN this function just calls
 *  \ref serPutChar  "serPutChar(c, NULL)"  resp. \ref lanPutC "lanPutC(c)"
 *  directly. <br />
 *  For all other streams \c fputc(c, stdout) is used.
 */
void  stdPutC(char c);

/** Write a RAM string to \c stdout
 *
 *  This is an abbreviation for \c putSt(src, stdout).
 */
#define stdPutS(src) putSt((src), stdout)

/** Write a flash memory string to \c stdout
 *
 *  This is an abbreviation for \c putSt_P(src, stdout).
 */
#define stdPutS_P(src) putSt_P((src), stdout)

/** Log the stack frame to stdout
 *
 *  \param maxBytes the max. number of bytes logged including the stack frame
 */
void logStackS(uint8_t maxBytes);

/// @}  // end member group stdio


//-------------------------------------  LAN streams  ------------------------

/// \name The Ethernet / LAN streams
/// @{

/** Initialise the Ethernet / LAN streams
 *
 *  This function initialises the \ref lanStreams for usage with
 *  \ref uip_appdata newly received or to be sent and additionally
 *  \ref lanStreamsOpen "opens" it.
 */
void lanStreamsInit(void);

/** Set the Ethernet / LAN streams to echo input
 *
 *  This function should be called after
 *  \ref lanStreamsInit() "inititialising" the \ref lanStreams and before
 *  the first output. <br />
 *  The received data will then be set as beginning of the data to be sent.
 *  This will be a 1:1 echo if nothing else is output.
 */
void lanStreamsEcho(void);

/** Open the Ethernet / LAN streams
 *
 *  This function sets the standard streams stdin/out/err to the
 *  \ref lanStreams.<br />
 *  This "switching" will be rescinded by \ref streamsClose() .
 *
 *  This function just (re-) opens leaving all else state (like buffer
 *  indices) un-changed. For a \ref lanStreamsInit() "first open" use
 *  \ref lanStreamsInit() .
 */
void lanStreamsOpen(void);

/** Sends the LAN streams data (over the Ethernet)
 *
 *  This function \ref uip_send() "sends" the LAN streams data written so far
 *  and \ref streamsClose "closes" the LAN streams.
 */
void lanStreamSend(void);

/** The buffer space available for Ethernet / LAN streams output */
uint16_t lanStreamsOutSpace(FILE * streams);

/** Check the space available Ethernet / LAN streams output
 *
 *  \sa str2CheckOutSpace
 */
uint8_t lanCheckOutSpace(uint16_t reqSpace, FILE * streams);

/** The number of characters buffered from LAN (stream) input
 *
 *  This function returns the number of bytes that can be fetched from
 *  LAN (stream) input by \ref lanGetChar(FILE*) respectively
 *  \ref lanGetC() without getting EOF.
 *
 *  The returned value is in the range 0 .. \ref uip_datalen(). If 0 is
 *  returned nothing should be read.
 */
uint16_t lanInBufferd(FILE * streams);

/** Get one byte from LAN stream input
 *
 *  This function never blocks. It returns a byte value read or EOF. <br />
 *  Hence this operation should be (indirectly) guarded by \ref lanInBufferd().
 *
 *  \return a character 0..255 or EOF
 */
int lanGetC(void);

/** Put one byte to LAN stream output
 *
 *  This function never blocks even if no output operation is feasible.
 *  Hence this operation should be (indirectly) guarded by
 *   \ref lanStreamsOutSpace().
 *
 *  This function will be used by stdio as output driver entry indirectly
 *  by \ref lanPutChar "lanPutChar(char c, FILE *stream)". It may as well be
 *  used directly for output to the current IP connection.
 *
 *  \param c the character to be output
 */
void lanPutC(char c);

//--------------  LAN   stream   minimal C-Lib definitions   ----------------

/** The stream "device" connected to the Ethernet (uIP)
 *
 *  This stream shall only be used while handling Ethernet messages, i.e.
 *  directly or indirectly in \ref uip_appcall(), \ref uipAppcall() or
 *  \ref  udp_appcall(). It will work on \ref uip_appdata, i.e. use it as
 *  buffer for both directions.
 *
 *  Output functions must not be used before \ref uip_appdata as input
 *  (message) is fully processed.
 *
 *  Every usage shall be embraced by \ref lanStreamsOpen() and
 *  \ref streamsClose() if used via standard streams.
 */
extern FILE lanStreams;

/** Get one byte from LAN stream input
 *
 *  This function never blocks. It returns a byte value read or EOF. <br />
 *  Hence this operation should be (indirectly) guarded by \ref lanInBufferd().
 *
 *  This function may be used directly for input but then
 *  \ref lanGetC(() is a better choice. Usually this function is used
 *  indirectly via stdin of stdio.h as this functions signature is what
 *  stdio expects of underlying input streams.
 *
 *  \param stream (passed by stdio input) totally ignored by this function
 *  \return a character 0..255 or EOF
 */
int lanGetChar(FILE *stream);

/** Put one byte to LAN stream output
 *
 *  This function does exactly the the same as the preferable
 *  \ref lanPutC() except for the extra parameter of type \c FILE * that is
 *  for sake of the standard I/O usage.
 *
 *  \param c the character to be output
 *  \param stream (passed by stdio output) totally ignored by this function
 *  \return c in case of success, EOF if output buffers full.
 */
int lanPutChar(char c, FILE *stream); //  __attribute__ ((pure));

/** Put some characters to LAN stream output
 *
 *  This function never blocks. It returns the number of characters
 *  transferred. If not enough space is available the returned value will
 *  be smaller than \c n or than the length of the string \c src. It may
 *  even be 0. In the latter case nothing was output.
 *
 *  Hence this operation should be (indirectly) guarded by
 *  \ref lanStreamsOutSpace().
 *
 *  The output stops after n (or space permitting) characters or at the first
 *  0 in src. By that this is a string output function. To send a (binary) 0
 *  use \ref lanPutC(char c).
 *
 *  \param src the characters to be output
 *  \param n   the maximum number of characters to be output
 *  \return the number of characters output
 */
int lanPutChars(char * src, uint16_t n);

/** Put a RAM string (some characters) to LAN stream output
 *
 *  This function never blocks. It returns the number of characters
 *  transferred. If not enough space is available the returned value will
 *  be smaller than the length of the string \c src. It may
 *  even be 0. In the latter case nothing was output.
 *
 *  Hence this operation should be (indirectly) guarded by
 *  \ref lanStreamsOutSpace().
 *
 *  The output stops after at the first 0 in src. By that this is a string
 *  output function. To send a (binary) 0 use  \ref lanPutC(char c).
 *
 *  \param src the characters to be output (0 terminated)
 *  \param stream the current Ethernet connection as stream (not used)
 *  \return the number of characters output
 */
int lanPutSt(char * src, const FILE * stream);

/** Put some characters from program space to LAN stream output
 *
 *  This function never blocks. It returns the number of characters
 *  transferred. If not enough space is available the returned value will
 *  be smaller than the length of the string \c src. It may
 *  even be 0. In the latter case nothing was output.
 *
 *  Hence this operation should be (indirectly) guarded by
 *  \ref lanStreamsOutSpace().
 *
 *  \param src the characters to be output in flash memory (0 terminated)
 *  \param stream the current Ethernet connection as stream (not used)
 *  \return the number of characters output
 */
int lanPutSt_P(char * src, const FILE * stream);


/// @}  // end member group lan device


//-------------------------------------  nul streams  ------------------------

/// \name The nul device
/// @{

/** The stream "device" connected to the nul device
 *
 *  This stream shall only be used as \ref initStdStreams "initial standard"
 *  streams if no other character / byte oriented I/O is permanently
 *  available (from reset on). That usually means the \ref serStreams "UART"
 *  is reserved for other purposes.
 *
 *  This streams shall be \ref nulStreamsOpen "set" respectively
 *  \ref nulStreamsOpen "opened" as standard I/O (stdio) whenever no other
 *  stream is available therefore. The rationale is any (user) thread might
 *  use standard I/O (stdio) any time.
 *
 *  The "device" features all in/output functions optimised for doing
 *  nothing. It is always available, i.e. from reset phase 1 on until
 *  shutdown.
 */
extern FILE nulStreams;

/** Open the nul streams
 *
 *  This function sets the standard streams stdin/out/err to the
 *  \ref nulStreams.<br />
 *  This "switching" will be rescinded by \ref streamsClose() .
 */
void nulStreamsOpen(void);

/** The buffer space available for nul output
 *
 *  This function always says 4095. It is provided to implement a stream
 *  that swallows and forgets all output.
 */
uint16_t nulStreamsOutSpace(FILE * streams);

/** Check the space available for nul streams output
 *
 *  This function always returns 0 i.e. available. It is provided to
 *  implement a stream that swallows unlimited amount of output (by just
 *  forgetting it).
 */
uint8_t nulCheckOutSpace(uint16_t reqSpace, FILE * streams);

/** The number of characters buffered from nul input
 *
 *  This function returns always 0 as there will be no
 *  input from nul. It is provided to implement an input stream that will
 *  never deliver anything respectively be always empty.
 */
uint16_t nulInBufferd(FILE * streams);

/** Get one byte from nul stream input
 *
 *  This function is provided to implement an input stream that will
 *  never deliver anything respectively be always empty.
 *
 *  \return EOF (always)
 */
int nulGetC(void);


/** Put one byte to nul stream output
 *
 *  This function never blocks and does nothing. It is provided to
 *  implement a stream that swallows unlimited amount of output (by just
 *  forgetting it).
 *
 *  All other output functions to nul do have no effect, also.
 *
 *  \param c the character to be output
 */
void nulPutC(char c);

/** Put one byte to nul stream output
 *
 *  This function never blocks. It returns the byte \c c provided as parameter
 *  value and does nothing else. It is provided to implement a stream
 *  that swallows and forgets all output.
 *
 *  \param c the character to be output
 *  \param stream unused, irrelevant (signature required by stdio)
 *  \return c
 */
int nulPutChar(char c, FILE *stream); //  __attribute__ ((pure));

/** Get one byte from nul stream input
 *
 *  This function returns \c EOF.  It is provided to implement an input stream
 *  that will never deliver anything respectively always be empty.
 *
 *  \param stream (passed by stdio input) totally ignored by this function
 *  \return \c EOF
 */
int nulGetChar(FILE *stream);

/** Put a string (some characters) to nul stream output
 *
 *  This function never blocks. It returns 1 no matter the parameters and does
 *  nothing else. It is provided to implement a stream that swallows unlimited
 *  amount of output (by just forgetting it).
 *
 *  As the pointer to the characters to output \c src is not de-referenced
 *  this function works for RAM and program (flash) memory strings as well.
 *
 *  \param src the characters to be output (not used)
 *  \param stream  (not used)
 *  \return 1 if src is not null or 0 else
 */
int nulPutSt(char * const src, const FILE * stream);

/// @}  // end member group nul device


/** @} */ // Doxygen addtogroup stdStreams
/** @} */ // Doxygen addtogroup appLaySup

#endif /* __STREAMS_H__ */

/** @} */ // Doxygen addtogroup weAutSys
