// Copyright (c) 2011, 2014  Albrecht Weinert, Bochum *
// $Id: syst_threads.h 2 2017-01-25 16:49:03Z albrecht $

/**  \addtogroup weAutSys
 * @{
 *    \addtogroup sysSer
 * @{
 *     \addtogroup sysThr
 * @{
 *
 * \file syst_threads.h
 *
 *  \ref intro_sec "weAutSys"' system threads
 *
 *  This file contains the definitions for \ref intro_sec "weAutSys"' system threads.
 *  Please find detailed description of the variables, functions etc.
 *  in the module \ref sysThr "system threads".
 *
 *  This is system software and must not be modified for user or
 *  application programs.
 *
 *  This file is part of \ref intro_sec "weAutSys" &nbsp;
 *          <<a href="http://weinert-automation.de">weinert-automation.de</a>>
 *
 * &nbsp; \ref copyLicNote " Copyright &copy; " 2014 Albrecht Weinert, Bochum
 * \author  Albrecht Weinert  &nbsp;
 *                            <<a href="http://a-weinert.de">a-weinert.de</a>>
 * $Revision: 2 $
 * $Date: 2017-01-25 16:49:03 +0000 (Wed, 25 Jan 2017) $
 */
//  SVN last change by $Author: albrecht $

/** @} */  // \addtogroup sysSer
/** @} */  // \addtogroup sysThr

#ifndef __SYST_THREADS_H__
#define __SYST_THREADS_H__

#include "pt/pt.h"
#include "we-aut_sys/common.h"

/**
 * \addtogroup sysSer
 * @{
 * \addtogroup sysThr
 * @{
 */

/** The milliseconds system function (or thread)
 *
 *  This thread is scheduled by the runtime very often (high priority).
 *  It does work on the condition that one or more ms-ticks occurred since
 *  its last schedule. Otherwise it returns / yields at once.
 *
 *  This thread function does<ul>
 *  <li> the basic time keeping for the timed scheduling
 *       of other system and user threads, </li>
 *  <li> the handling of timers with milliseconds resolution, </li>
 *  <li> the basic process digital / analougue input (DI / AI)
 *       handling and filtering, </li>
 *  <li> some time critical flow / overload control on communication
 *       links as well as </li>
 *  <li> the main \ref deviceDriverPoll() "polling" of the Ethernet driver
 *       ENC28J60 every 10 ms.</li></ul>
 *
 *  In normal operation this function should be called at least twice every
 *  ms. But if other processor tasks inhibit this it is quite capable to
 *  handle several events (= ms ticks) at once.
 *  Nevertheless, in case of more than 10 resp. 100 collected ticks the
 *  scheduling of user and system threads might come offbeat as well as the
 *  sequence of effects of lapsed timers.
 *
 *  In case of more than 250 collected ticks, all might go wrong.
 *
 *  Internal remark: This function is (still) called "...Thread" as it could
 *  as well organised as one (as it once was). On the other hand the system
 *  1ms actions have to be handled in one optimised compact action without
 *  yielding and blocking. So the (small) Protothreads overhead is skimped
 *  at the moment.
 */
void sys1msThread(void);


/** The 100 milliseconds system thread, the function
 *
 *  This thread is scheduled quite often (high priority). It yields
 *  internally until sys100msPeriodFlag is set.
 *
 *  This system thread handles the watchdog and does all periodic timing
 *  for the Ethernet stack
 *  \ref uip "uIP".
 *
 * \sa ptSys100msThread
 */
ptfnct_t sys100msThread(void);

/** The 100 milliseconds system thread, the (raw) Protothreads datastructure
 */
extern struct pt ptSys100msThread;

/** Flag for system 100ms thread
 *
 *  This flag will be set by (the time keeping part of) the 1ms thread
 *  sys1MsThread.
 */
extern uint8_t sys100msPeriodFlag;



/** Flag for system 1s thread
 *
 *  This flag will be set by (the time keeping part of) the 1ms thread
 *  sys1msThread. <br />
 *  \ref SINCsINCR "Bit   0": "run since" seconds were incremented <br />
 *  \ref ABSsecADJ "Bit   7": "absolute date time" seconds were incremented
 *           / adjusted<br />
 *  \ref ABSsecSET "Bit   6": "absolute date time" seconds were set
 *         (heavily changed)
 */
extern uint8_t sys1sPeriodFlag;

#define ABSsecSET 0x40 //!< absolute date time seconds were set \showinitializer
#define ABSsecADJ 0x80 //!< absolute date time seconds were adjusted \showinitializer
#define SINCsINCR 0x01 //!< run since seconds were incremented \showinitializer

/** The 1 second system thread, the function
 *
 *  This thread is scheduled repeatedly (medium priority).
 *  It yields internally or exits until \ref sys1sPeriodFlag is set.
 *
 *  It does all date / time and  timer with seconds resolution handling.
 *
 * \sa ptSys1sThread;
 */
ptfnct_t sys1sThread(void);

/** The 1 second system thread, the (raw) Protothreads datastructure */
extern struct pt ptSys1sThread;


/** Initialise the threads
 *
 *  Besides some other initialisations this will de-register all
 *  \ref appThr "user / application threads".
 *
 *  \sa pt.h
 */
void initThreads(void);


//------------------------------------------------------------------------


/** Output some system (version) info, the thread function
 *
 *  This thread will output some system info, like <br />
 *  name, version, copyright, build and so on to the standard output.
 *  This is meant for system start respectively welcome info. This thread
 *  will usually started as child thread (of \c pt) by the macro
 *  \c  #runOutSysInfoThread(parent, pt, toStream) .
 *
 *  As the multi-line output may be longer than \c toStream's output buffer
 *  this thread will repeatedly yield until buffer space is available. If
 *  started as child thread the parent thread will "inherit" these blocks or
 *  waits.
 *
 *  \param pt the thread state (pointer to raw structure; not NULL)
 *  \param toStream streams to switch the standard streams to (if not MULL)
 *  \return \ref PT_YIELDED if waiting for conditions that may become
 *           automatically true; \ref PT_WAITING if blocked by conditions
 *           that other's action may set true;  \ref PT_EXITED if stopped
 *           by conditions that never will become true (sorry, not all work
 *           done and never will; \ref PT_ENDED if ready with all work
 */
ptfnct_t outSysInfoThreadF(pt_t * pt, FILE * toStream);


/** Start (spawn) the #outSysInfoThreadF thread as child.
 *
 *  \param parent (type pt) the calling parent's thread raw (!) stucture
 *  \param pt    the thread's to be called raw (!) stucture
 *  \param toStream streams to switch the standard streams to (if not MULL)
 */
#define runOutSysInfoThread(parent, pt, toStream) do { \
   PT_INIT(&pt); \
   PT_WAIT_ASYIELD_WHILE((parent), (outSysInfoThreadF( &(pt), toStream)) < PT_EXITED);  \
 } while(0)


//-------------------------------------------------------------------------

/** Reset respectively initialise an user / application thread
 *
 *  This is a macro: Do set the parameter as name (i.e. \c the_thread) and
 *  not as a pointer to resp. address of (not \c &the_thread).
 * \hideinitializer
 * \param thread (type \ref mThr_data_t) the thread to reset.
 */
#define resetAppThread(thread) do { \
      (thread).fun = NULL; \
      (thread).retC = (thread).flag = 0; \
   } while(0)


/** Schedule a registered user / application thread
 *
 *  This macro will schedule the thread provided as parameter if it
 *  has a thread-function registered.
 *
 *  Besides checking this condition its like #scheduleAppThread(thread).
 *
 * \hideinitializer
 * \param thread (type \ref mThr_data_t) the thread to schedule;
 *        must not be NULL.
 */
#define scheduleRegdAppThread(thread) do { \
   if (thread.fun) thread.retC = thread.fun(&thread); } while(0)

/** Unconditionally schedule an user / application thread
 *
 *  This macro will schedule the thread provided as parameter. As this is
 *  done unconditionally the caller is responsible for two hard
 *  pre-conditions<ol>
 *  <li> thread must not be NULL </li>
 *  <li> thread.regd must be true </li></ol>
 *
 *  This is a macro, set the parameter as name (thread) and not as pointer
 *  to / address of. (That means not use &thread)
 *
 * \hideinitializer
 * \param thread the thread to reset; must not be NULL.
 */
#define scheduleAppThread(thread) do { \
   thread.retC = thread.fun(&thread); } while(0)



/** Schedule a flagged user / application thread
 *
 *  This macro will schedule the thread provided as parameter if its flag
 *  is set (and it has a thread-function registered).
 *
 *  Besides checking this condition its like #scheduleRegdAppThread(thread).
 *
 * \hideinitializer
 * \param thread the thread to schedule; must not be NULL.
 */
#define scheduleFlgdAppThread(thread) do { \
   if (thread.fun && thread.flag) \
      thread.retC = thread.fun((struct mThr_data_t *)&thread); \
   } while(0)


/** Schedule a flagged or yielding user / application thread
 *
 *  This macro will schedule the thread provided as parameter if its flag
 *  is set or it has ended by yielding or waiting on its last schedule.
 *
 *  Besides checking this ORed run conditions its
 *  like #scheduleFlgdAppThread(thread).
 *
 * \hideinitializer
 * \param thread (type \ref mThr_data_t) the thread to schedule;
 *        must not be NULL.
 */
#define scheduleYldAppThread(thread)  do { \
  if (thread.fun && (thread.flag || thread.retC == PT_YIELDED)) \
       thread.retC = thread.fun((struct mThr_data_t *)&thread); } while(0)

/** Exit a not flagged user / application thread
 *
 *  This macro will exit the thread provided as parameter if its flag
 *  is not set.
 *
 *  \sa PT_EXIT
 *  \sa PT_EXITED
 *  \hideinitializer
 *  \param thread (state type \ref mThr_data_t) the thread to
 *               conditionally exit; must not be NULL.
 */
#define exitNotFlaggedThread(thread)   do {   \
  if (!((thread).flag)) { PT_EXIT(&(thread).pt); }} while(0)

// macros above here avoids warning: documentation for unknown define... found
//------------------------------------------------------------------------


/*  Output some system runtime and state information
 *
 *  This thread will output some info on about how long the system is
 *  running, the reset cause and some others on standard out (which is
 *  initially / default the serial output).<br />
 *  This thread will usually be run as background task getting orders
 *  and clearing them by bits of this flag.
 *
 *  This thread will block until the  output (of \c toStream) buffer has
 *  enough space.<br />
 *
 *  This thread function acts on \ref hierThr_data_t "threadD"'s primary
 *  thread and primary flag.  \c flag's bits  signal which informations
 *  are to be output: <br />
 *
 *  Single Bits of the \c flag signal which informations are to be
 *  output: <br />
 *  Bit 0 : SMC (small memory card) basic driver state info <br />
 *  Bit 1 : SPI2 (UART1 as SPI) info <br />
 *  Bit 2 : UART (0) info <br />
 *  Bit 3 : NTP info <br />
 *  Bit 4 : running time, reset cause and RAM usage info <br />
 *  Bit 5 : current local date <br />
 *  Bit 6 : current local time <br />
 *  Bit 7 : zone offset and DST <br />
 *
 *  This thread function will delegate on \ref hierThr_data_t "threadD"'s
 *  secondary thread and flag to the separate thread function
 *  \ref outEncLanInfoThreadF .  See the  \c flag2's bits meaning
 *  \ref outEncLanInfoThreadF "there".
 *
 *  \param threadD the thread state data (pointer to; not NULL)
 *  \param toStream streams to switch the standard streams to (if not MULL)
 *  \return \ref PT_YIELDED if waiting for conditions that may become
 *           automatically true; \ref PT_WAITING if blocked by conditions
 *           that other's action may set true;  \ref PT_EXITED if stopped
 *           by conditions that never will become true (sorry, not all work
 *           done and never will; \ref PT_ENDED if ready with all work
 */
ptfnct_t outRunInfoThreadF(hierThr_data_t * threadD, FILE * toStream);


//------------------------------------------------------------------------


/**  Output a flash string array to standard output thread
 *
 *  This thread will output a number of strings in flash memory given as an
 *  array of pointers to them by initialising the thread with
 *  \ref initOutFlashTextThread.<br />
 *  The \ref initOutFlashTextThread "initialising" must be done before
 *  starting the thread.
 *
 *  This thread will block (multiple times) until the output buffer has
 *  enough space.<br />
 *  If started as child thread the parent thread will "inherit" these
 *  blocks or waits as yields.
 *
 *  \param outFlashTextThread the thread state data
 *  \param toStream streams to switch the standard streams to (if not MULL)
 *  \return \ref PT_YIELDED if waiting for conditions that may become
 *           automatically true; \ref PT_WAITING if blocked by conditions
 *           that other's action may set true;  \ref PT_EXITED if stopped
 *           by conditions that never will become true (sorry, not all work
 *           done and never will; \ref PT_ENDED if ready with all work
 */
ptfnct_t outFlashTextThreadF(outFlashTextThr_data_t * outFlashTextThread,
                                                            FILE * toStream);

/** Initialise the output a flash string array to standard output thread
 *
 *  After this call the thread \ref outFlashTextThreadF() will (block and)
 *  work until \c noOfFlashStrings2out strings were output or a NULL is
 *  found in  \c theFlashStrings2out which is considered as end of the array.
 *
 *  \param outFlashTextThread the thread state data
 *  \param theFlashStrings2out pointer to (flash) array
 *                             of (flash) strings to be output
 *  \param noOfFlashStrings2out  number of (flash) strings to be output
 *
 */
void initOutFlashTextThread(outFlashTextThr_data_t * outFlashTextThread,
      char const * const * theFlashStrings2out, uint8_t noOfFlashStrings2out);

/** Register an user / application thread
 *
 * \param thread the thread to register to; must not be NULL.
 * \param threadF the thread function to be registered;
 *         NULL means de-register the thread
 * \return the previously registered thread function; NULL if there was none
 * \sa initSetAppThread(struct mThr_data_t *, p2ptFun);
 */
p2ptFun registerAppThread(struct mThr_data_t * thread, p2ptFun threadF);

/** Register and (force) init an user / application thread.
 *
 *  This function will register \c threadF with \c thread unconditionally
 *  and init / reset all. It is meant for first initialisation.
 *
 *  \note Never call more than once with the same parameters and never
 *  call on a started thread. The results may be unwelcome to put it very
 *  mildly. If the situation is unclear in that respect, use
 *  \c registerAppThread(mThr_data_t, p2ptFun) instead.
 *
 * \param thread the thread to register to; must not be NULL.
 * \param threadF the thread function to be registered;
 *         NULL means de-register the thread
 * \sa  registerAppThread(struct mThr_data_t *, p2ptFun )
 */
void initSetAppThread(struct mThr_data_t * thread, p2ptFun threadF);

/** @} */ // Doxygen ..group sysThr

/**
 * \addtogroup appThr
 * @{
 */


/** The user / application 10ms thread
 *
 *  The flag value is the number of periods the scheduled thread should work
 *  on. It is normally 1 except for delays due to processor overload.
 *  The flag must be reset by the thread function
 */
extern  struct mThr_data_t app10msThread;

/** The user / application 1ms thread
 *
 *  The flag value is the number of periods the scheduled thread should work
 *  on. It might well be greater than 1 as processor load might very well
 *  inhibit this thread to be scheduled at every single milliseconds tick.
 *  The flag must be reset by the thread function.
 */
extern struct mThr_data_t app1msThread;

/** The user /  application 100ms thread
 *
 *  The flag value is the number of periods the scheduled thread should work
 *  on. It is normally 1 except for delays due to processor overload.
 *  This flag must be reset by the thread function.
 */
extern struct mThr_data_t app100msThread;

/** The user / application 100ms thread
 *
 *  The flag value is the number of periods the scheduled thread should work
 *  on. It is normally 1 except for delays due to processor overload.
 *  This flag must be reset by the thread function.
 */
extern struct mThr_data_t app1sThread;

/** The user / application serial input handling thread
 *
 *  This thread handles serial (UART0) input if the user software
 *  \ref registerAppSerInpThread "registers" an appropriate
 *  \ref appSerInpThreadF "function" therefore.
 *
 *  The flag will be set by system software, whenever <br />
 *  a) the UART receives a \ref LF "line feed" or \ref CR "return code",
 *     i.e. when \ref uartInRetBufferd() would return >= 1 or <br />
 *  b) the UART input buffer is filled to (near full) a level, i.e. when
 *     the the sender would be stopped by
 *     \ref setUARTflowcontrolByChar "flow control" (if
 *     \ref isCTSflowC "activated"). <br />
 *  appSerInpThread.flag will be cleared by system software if the UART
 *  input buffer gets \ref SER_RECVBUF_EMPTY "empty".
 *
 *  It is up to the  \ref registerAppSerInpThread "registered" user thread
 *  function to reset the flag on other conditions if appropriate.
 */
extern struct thr_data_t appSerInpThread;


/** Register a user / application serial input processing thread
 *
 *  The thread so registered will be scheduled if the serial input buffer
 *  contains a line feed (and hence a line to interpret as command) or is
 *  nearly full.
 *
 *  It is expected that the (user's / customer's) implementation reads
 *  one line (resp. all the input causing the near overflow) and interprets
 *  it according to its own specification. The function \ref setCliLine()
 *  can be used to hand an input line to system and application command line
 *  interpreters.
 *
 * \param threadF the thread function to register;
 * \return the previously registered thread function
 * \sa appSerInpThreadF
 * \sa appInitThreadF
 * \sa uartGetLine(char line[], int max)
 */
p2ptFunC registerAppSerInpThread(p2ptFunC  threadF);


/** The user / application serial input processing thread
 *
 *  If something is to be done in the user / application software
 *  for characters or lines received via serial input an input processing
 *  thread (i.e. a protothread function) must be provided therefore in
 *  an application specific source file and that thread function has
 *  to be registered according to this declaration.
 *
 *  For that function any name can be chosen. This declaration is provided
 *  just for convenience and harmonisation.
 *
 *  The parameter this function (if registered) would be called with will
 *  (at present) point to \ref appSerInpThread. Its flag will be set by
 *  system software as described there. Its up to the implementation to
 *  read (buffered) UART input and act on it accordingly. If the UART connects
 *  to a human operated terminal one would normally use the the variable part
 *  of the parameter thread structure as command line interpreter
 *  \ref cliThr_data_t "(CLI) structure" by help of the \ref setCliLine()
 *  function.
 *
 *  \param serInpThread
 *  \sa appSerInpThread
 *  \sa registerAppSerInpThread(p2ptFun)
 *  \sa appInitThreadF
 */
ptfnct_t appSerInpThreadF(struct thr_data_t * serInpThread);


/** Register a user / application background task thread
 *
 *  The thread so registered thread will be scheduled at lowest priority.
 *
 * \param threadF the thread to register;
 * \return the previously registered thread function
 * \sa appBgTskThreadF
 * \sa appInitThreadF
 */
p2ptFun registerAppBgTskThread(p2ptFun  threadF);

/** The user / application specific background task thread */
extern struct mThr_data_t appBgTskThread;

/** The user / application specific background task thread (the function)
 *
 *  If something can or has to be done with the lowest priority (and in
 *  CPU usage partitioning < 2ms (!)) it may be done the background
 *  thread.<br />
 *  In that case a protothread function must be provided therefore in
 *  an application specific source file and that thread function has
 *  to be \ref registerAppBgTskThread "registered" accordingly.
 *
 *  For that function any name can be chosen. This declaration is provided
 *  just for convenience and harmonisation.
 *
 *  \note It is really essential that the background thread function will
 *  yield, block or \ref PT_EXIT exit after less than 2 ms (better < 1ms)
 *  CPU usage. If no work is to be done it should do <br />
 *  return \ref PT_EXITED; <br />
 *  even before the <br />
 *  \ref PT_BEGIN "PT_BEGIN"(&appBgTskThread.pt); <br />
 *  <br /> Note also that the background thread is also scheduled in the
 *  state \ref PLCstop. If automation tasks are done in this thread the
 *  PLC state should be checked by \ref PLCinRUN (or  \ref PLCinSTOP).
 *
 *  \sa registerAppBgTskThread(p2ptFun)
 *  \sa appInitThreadF
 */
ptfnct_t appBgTskThreadF(void);


/** The user / application specific initialisation thread
 *
 *  This protothread function must be provided in an user / application
 *  specific source file. This thread is scheduled in one to three phases
 *  divided by yield points or until it exits.
 *
 *  Phase 1: pre system init. <br />
 *  In this phase user variables and data structures may be set as well
 *  as system variables controlling system initialisation.
 *
 *  Phase 2: post system init. <br />
 *  In this phase serial communication is available.
 *
 *  Phase 3: post persistence init. <br />
 *  In this phase persistent configuration data have been fetched and
 *  may be modified before being used in following phases.
 *
 *  Phase 4: post Ethernet init. <br />
 *  In this phase the driver and the stack are initialised. This does not
 *  necessarily mean the link is up or the valid IP address is set already.
 *
 *  Nevertheless communication might be prepared and
 *  \ref uip_listen "listen" calls might be made here.
 *
 *  Phase 5: post ? ? (future use) init. <br />
 *  In this phase is presently empty.
 *
 *  Phase 6: reschedule until exited. <br />
 *  From this phase on the initialisation thread will be rescheduled until
 *  it exits. This allows for any initialisation work whatever complicated
 *  or time consuming.
 *
 *  Once exited the initialisation thread will never be rescheduled. This
 *  is already true for all above phases.
 *
 *  It is essential that all user threads (cyclic, CLI etc.) be registered
 *  here.
 *
 *  <br />
 *  \section variant_sec Settings of variants
 *
 *  The weAut_01 board allows some hardware variants by <ul>
 *  <li> choosing other components or even just omitting some </li>
 *  <li> setting of jumpers </li>
 *  <li> connecting extensions to jumpers / connectors </li>
 *  <li> usage or non-usage of certain micro-controller features, like <br />
 *       AD, timer, counter, generator, 2wire, UART, flow control, <br />
 *       on the respective I/O. </li></ul>
 *
 *  The user's initialisation thread is the (main) place to adapt
 *  to such and other variants.<br />
 *  \param pt pointer to protothread structure
 */
ptfnct_t appInitThreadF(struct pt * pt);


/** The user / application specific one second thread
 *
 *  If something is to be done in the user / application one second thread
 *  a protothread function must be provided therefore in an application
 *  specific source file and that thread function has to be registered
 *  accordingly.
 *
 *  For that function any name can be chosen. This declaration is provided
 *  just for convenience and harmonisation.
 *
 *  \sa registerAppThread(struct mThr_data_t *, p2ptFun)
 *  \sa app1sThread
 *  \sa appInitThreadF
 *  \param uthr_data pointer to extended protothread structure
 */
ptfnct_t app1sThreadF(struct mThr_data_t * uthr_data);

/** The user / application specific 100 ms thread
 *
 *  If something is to be done in the user / application 100 milliseconds
 *  thread a protothread function must be provided therefore in
 *  an application specific source file and that thread function has
 *  to be registered accordingly.
 *
 *  For that function any name can be chosen. This declaration is provided
 *  just for convenience and harmonisation.
 *
 *  \sa registerAppThread(struct mThr_data_t *, p2ptFun)
 *  \sa app100msThread
 *  \sa appInitThreadF
 *  \param uthr_data pointer to extended protothread structure
 */
ptfnct_t app100msThreadF(struct mThr_data_t * uthr_data);

/** The user / application specific 10 ms thread
 *
 *  If something is to be done in the user / application 10 milliseconds
 *  thread a protothread function must be provided therefore in
 *  an application specific source file and that thread function has
 *  to be registered accordingly.
 *
 *  For that function any name can be chosen. This declaration is provided
 *  just for convenience and harmonisation.
 *
 *  \sa registerAppThread(struct mThr_data_t *, p2ptFun)
 *  \sa app10msThread
 *  \sa appInitThreadF
 *  \param uthr_data pointer to extended protothread structure
 */
ptfnct_t app10msThreadF(struct mThr_data_t * uthr_data);

/** The user / application specific 1 ms thread
 *
 *  If something is to be done in the user / application 1 milliseconds
 *  thread a protothread function must be provided therefore in
 *  an application specific source file and that thread function has
 *  to be registered accordingly. Best care should be taken to have the
 *  CPU load between two yield or block points well below 150s.
 *
 *  For that function any name can be chosen. This declaration is provided
 *  just for convenience and harmonisation.
 *
 *  \sa registerAppThread(struct mThr_data_t *, p2ptFun)
 *  \sa app10msThread
 *  \sa appInitThreadF
 *  \param uthr_data pointer to extended protothread structure
 */
ptfnct_t app1msThreadF(struct mThr_data_t * uthr_data);


/** @} */  // \addtogroup sysSer
/** @} */  // \addtogroup sysThr

#endif /* __SYST_THREADS_H__ */
/** @} */ // Doxygen addtogroup weAutSys
