/*
 * Copyright (c) 2004-2005, Swedish Institute of Computer Science.
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met: 
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright 
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution. 
 * 3. Neither the name of the Institute nor the names of its contributors 
 *    may be used to endorse or promote products derived from this software 
 *    without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE 
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE. 
 *
 * This file is part of the protothreads library.
 * 
 * Author: Adam Dunkels <adam@sics.se>
 */

/**
 * \addtogroup osSWinc
 * @{
 * \addtogroup pt
 * @{
 */

/**
 * \defgroup lc Local continuations
 * @{
 * \details
 * Local continuations form the basis for implementing protothreads. A
 * local continuation can be <i>set</i> in a specific function to
 * capture the state of the function. After a local continuation has
 * been set can be <i>resumed</i> in order to restore the state of the
 * function at the point where the local continuation was set.
 */

/**
 * \file lc.h
 *
 * Local continuations
 *
 * \author
 * Adam Dunkels <adam@sics.se>
 *
 *  Modification for \ref intro_sec "weAutSys" and GCC by A. Weinert <br />
 *  Copyright (c) 2011 Albrecht Weinert <br />
 *  weinert - automation, Bochum <br />
 *  weinert-automation.de  weAut.de  a-weinert.de
 *
 * $Revision: 2 $
 * $Date: 2017-01-25 16:49:03 +0000 (Wed, 25 Jan 2017) $
 */

#ifdef DOXYGEN

/** Initialise a local continuation
 *
 *   This operation initialises the local continuation, thereby
 *   unsetting any previously set continuation state.
 *
 * \hideinitializer
 */
#define LC_INIT(lc)

/** Set a local continuation.
 *
 *  The set operation saves the state of the function at the point where the
 *  operation is executed. As far as the set operation is concerned, the
 *  state of the function does <b>not</b> include the call-stack or local
 *  (automatic) variables, but only the program counter and such CPU
 *  registers that needs to be saved.
 *
 * \hideinitializer
 */
#define LC_SET(lc)

/** Resume a local continuation
 *
 *  The resume operation resumes a previously set local continuation, thus
 *  restoring the state in which the function was when the local continuation
 *  was set. If the local continuation has not been previously set, the
 *  resume operation does nothing.
 *
 * \hideinitializer
 */
#define LC_RESUME(lc)

/** Mark the end of local continuation usage
 *
 *  The end operation signifies that local continuations should not be used
 *  any more in the function. This operation is not needed for most
 *  implementations of local continuation, but is required by a few
 *  implementations.
 *
 * \hideinitializer 
 */
#define LC_END(lc)

/** The local continuation type
 *
 * \hideinitializer
 */
typedef void * lc_t;

#else

#ifndef __LC_H__
#define __LC_H__

/* At this point an implementation of local continuations has to be included.
 * Adam Dunkels gives us lc-switch.h and lc-addrlabels.h.
 *
 * lc-addrlabels.h only works with GCC C and is then the best solution for
 * devices with up to 128K flash. For larger ones the (void*) pointer size
 * gets to small. avr-gcc works around with a magical linker approach called
 * trampolin that alleges to cover "labels as values" too.
 */

#include "lc-addrlabels.h"
#ifndef __GNUC__
#  warning lc-addrlabels used hers run with GCC C only!
# endif /* __GNUC__  */

#endif /* __LC_H__ */

#endif /* DOXYGEN */
/** @} */ // group lc
/** @} */ // group pt
/** @} */ // osSWinc
