/*
 * Copyright (c) 2004-2005, Swedish Institute of Computer Science.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * This file was part of the Contiki operating system.
 *
 * Author: Adam Dunkels <adam@sics.se>
 */

/**
 * \addtogroup osSWinc
 * @{
 * \addtogroup pt
 * @{
 */

/**
 * \file pt.h
 *
 * Protothreads implementation
 * \author
 * Adam Dunkels <adam@sics.se>
 *
 * Modified by:     Albrecht Weinert <a-weinert.de>
 * Modifications'   Copyright (c) 2014 Albrecht Weinert, Bochum
 *
 * This file is part of the \ref intro_sec "weAutSys" runtime system.
 *
 * The changes Adam Dunkel's original are
 *
 * 1) changing the thread function's return type (centrally) to
 *    \c uint8_t for GCC C. (It was hard-coded \c char.)
 *
 * 2) unconditionally taking lc-adresslabes.h in lc.h for sake of
 *    optimised GCC C and tool usage. (It gives a warning on non GCC.
 *    In that case change the the implementation of local continuations.)
 *
 * 3) introducing an alias type for the Protothreads (raw) datastructure.
 *
 * 4) a handful of minor improvements and additions
 *
 * 5) extending the (doxygen) comments
 *
 * Remark: As the Protothreads return types and the (raw) datastructure are
 * intentionally more than lean. A.W. could also hardly resist the temptation
 * to make a pointer to the thread function, a run flag, the current state
 * and else minimal requirements part of struct pt. This would have brought
 * together what belongs together. <br />
 * But such (semantic) change would have spoiled the documented handling of
 * Protothreads (as e.g. in psocks). So this was not done here. The
 * necessities of \ref intro_sec "weAutSys" and akin were implemented in
 * \ref mThr_data_t (in \ref syst_threads.h).
 *
 * 2
 * 25.01.2017
 */

#ifndef __PT_H__
#define __PT_H__

#include "lc.h"

/** The return type of a protothread function
 *
 *  /note Within the \ref PT_BEGIN \ref PT_END braces a protothread function
 *  must not return directly (forgetting rare exceptions with lc-addrlabels).
 *  The return is done by Protothread operations.
 *
 *  Return values are #PT_WAITING, #PT_YIELDED, #PT_EXITED and #PT_ENDED.
 *
 *  The following holds: <br />
 *  a) #PT_WAITING < #PT_YIELDED < #PT_EXITED < #PT_ENDED <br />
 *  b) There is no real semantic difference between #PT_EXITED
 *     and #PT_ENDED <br />
 *
 *  The knowledge a) is useful in evaluating return values.
 *
 *  The difference of  #PT_EXITED and #PT_ENDED is almost
 *  insignificant:<br />
 *  The macro #PT_END which returns #PT_ENDED must appear exactly once at
 *  the end of a protothread function because it is the closing bracket
 *  to #PT_BEGIN.<br />
 *  The macro #PT_EXIT returning #PT_EXITED may on the other hand appear as
 *  often as useful.<br />
 *  The effect of #PT_EXIT and #PT_END is exactly the same: <br />
 *  The thread does end meaning its next schedule will "land" / start
 *  thread work after #PT_BEGIN. (Hence you will hardly find any code that
 *  won't OR #PT_EXITED and #PT_ENDED in concerning conditionals.<br />
 *  An extra macro \ref PT_LEAVE returns #PT_ENDED and can be used anywhere
 *  (saving jumps to \ref PT_END just for the intended return value).
 */
#ifdef __GNUC__
#  define ptfnct_t uint8_t
#else
#  define ptfnct_t char
#endif

/** A protothread's (raw) data structure  */
struct pt {
  lc_t lc; //!< The local continuation
}; // pt

/** A protothread's (raw) data structure as type */
typedef struct pt pt_t;


#define PT_WAITING 0  //!< thread paused due to a condition \showinitializer

/** Thread paused
 *
 *  Yielding is mostly done as courtesy to other threads or to obey CPU usage
 *  trime limits. As Protothreads are non-preemptive this is the
 *  responsibility of every single thread function!
 *
 *  Insofar the semantic difference to \ref PT_WAITING "waiting" is the
 *  (extra) reason to pause the thread, lest it runs too long. In that sense
 *  yielding with an condition should yield at least once no matter the
 *  condition's state &mdash; as does the original Protothread.
 *
 *  On the other hand \ref intro_sec "weAutSys" needed a to convey the
 *  (semantic) difference on the reason for a wait on a condition:<ul>
 *  <li> condition will fulfill automatically or </li>
 *  <li> the calling software must actively take action.</li></ul>
 *  Herefore we use \ref PT_YIELDED vs. \ref PT_WAITING to transport this
 *  information via the (child) thread functions return value.
 */
#define PT_YIELDED 1

#define PT_EXITED  2   //!< thread has finished \sa ptfnct_t
#define PT_ENDED   3   //!< thread has finished \sa ptfnct_t

/**
 * \name Initialization
 * @{
 */

/** Initialise a protothread
 *
 *  Initialisation must be done prior to starting to execute the protothread.
 *
 * \param pt A pointer to the protothread control structure.
 * \sa PT_SPAWN()
 * \hideinitializer
 */
#define PT_INIT(pt) LC_INIT((pt)->lc)

/** @} */
/**
 * \name Declaration and definition
 * @{
 */

/** Declare the start of a protothread inside the protothread function
 *
 *  This macro is used to declare the starting point of a protothread. It
 *  should be placed at or near the start of the function in which the
 *  protothread runs respectively is implemented. \ref PT_BEGIN() is the
 *  gridiron to the last set (left, yield, block) re-entry point.
 *
 *  All statements above the \ref PT_BEGIN()
 *  invocation will be executed each time the protothread is scheduled.
 *
 *  The statements below  \ref PT_BEGIN() will be executed at the first entry
 *  into an \ref PT_INIT "initialised" or \ref PT_RESTART "restarted"
 *  protothread.
 *
 *  \param pt the pointer to the protothread control structure.
 *  \hideinitializer
 */
#define PT_BEGIN(pt) { ptfnct_t PT_YIELD_FLAG = 1; LC_RESUME((pt)->lc)

/** Declare the end of a protothread
 *
 *  This macro is used for declaring that a protothread ends. It must
 *  always be used together with a matching \ref PT_BEGIN() macro. A statement
 *  below won't be reached (except when labeled and gone to).
 *
 *  \param pt the pointer to the protothread control structure.
 *  \hideinitializer
 *  \internal   PT_YIELD_FLAG = 0;  only to avoid GCC warning (not nec.)
 */
#define PT_END(pt) LC_END((pt)->lc);  PT_YIELD_FLAG = 1; \
                   PT_INIT(pt); return PT_ENDED; }

/** @} */
/**
 * \name Blocked wait
 * @{
 */

/** Block and wait until condition is true
 *
 *  This macro blocks the protothread until the specified condition is true.
 *
 *  \param pt        pointer to the protothread control structure
 *  \param condition the condition (a boolean expression)
 *  \hideinitializer
 */
#define PT_WAIT_UNTIL(pt, condition) \
  do { \
    PT_YIELD_FLAG = 0; \
    LC_SET((pt)->lc);  \
    if(!(condition)) {  \
      return PT_WAITING; \
    } \
  } while(0)

/** Block and wait while condition is true
 *
 *  This function blocks and waits while condition is true. See
 *  PT_WAIT_UNTIL().
 *
 *  \param pt   pointer to the protothread control structure
 *  \param cond the condition (a boolean expression)
 *
 *  \hideinitializer
 */
#define PT_WAIT_WHILE(pt, cond)  PT_WAIT_UNTIL((pt), !(cond))

/** @} */
/**
 * \name Hierarchical protothreads
 * @{
 */

/** Block and wait until a child protothread completes
 *
 *  This macro schedules a child protothread. The current protothread
 *  will \ref PT_WAITING "block" until the child protothread completes.
 *
 *  \note The child protothread must be manually initialised with the
 *        PT_INIT() function before this function is used.
 *
 *  \param pt     A pointer to the (calling) protothread control structure.
 *  \param thread The child protothread with arguments
 *
 *  \sa PT_SPAWN()
 *  \hideinitializer
 */
#define PT_WAIT_THREAD(pt, thread) PT_WAIT_WHILE((pt), ((thread) < PT_EXITED)))

/** Spawn a child protothread and wait until it exits
 *
 *  This macro spawns a child protothread and waits until it exits. The
 *  macro can only be used within a protothread.
 *
 *  \param pt A pointer to the (calling) protothread control structure
 *  \param child A pointer to the child protothread's control structure
 *  \param thread The child protothread's function call with arguments
 *  \hideinitializer
 */
#define PT_SPAWN(pt, child, thread)      \
  do {                  \
    PT_INIT((child));            \
    PT_WAIT_THREAD((pt), (thread));      \
  } while(0)

/** @} */
/**
 * \name Exiting and restarting
 * @{
 */

/** Restart the protothread
 *
 *  This macro will \ref PT_WAITING "block" once and cause the running
 *  protothread to restart its execution at the place after \ref PT_BEGIN().
 *  \param pt A pointer to the protothread control structure
 *  \hideinitializer
 */
#define PT_RESTART(pt) \
  do { \
    PT_INIT(pt);   \
    return PT_WAITING; \
  } while(0)

/** Exit the protothread
 *
 *  This macro causes the protothread to exit. If the protothread was
 *  spawned by another protothread, the parent protothread will become
 *  unblocked and can continue to run.
 *  \param pt A pointer to the (raw) protothread control structure
 *  \hideinitializer
 */
#define PT_EXIT(pt)   \
  do {  \
    PT_INIT(pt);      \
    return PT_EXITED; \
  } while(0)

/** End the protothread
 *
 *  This macro causes the protothread to end normally. It has the same effect
 *  as \ref PT_EXIT except returning \ref PT_ENDED. Insofar it is equivalent
 *  to falling through to the end of the \ref PT_BEGIN .. \ref PT_END brace.
 *  \param pt A pointer to the (raw) protothread control structure
 *  \hideinitializer
 */
#define PT_LEAVE(pt)  \
  do {  \
    PT_INIT(pt);     \
    return PT_ENDED; \
  } while(0)


/** @} */
/**
 * \name Yielding from a protothread
 * @{
 */

/** Yield from the current protothread
 *
 *  This function will yield the protothread, thereby allowing other
 *  processing to take place. As Protothreads are non-preemptive by nature
 *  unconditional yielding is one (and the only) way to dependably subdivide
 *  longer running tasks into junks of maximum allowed execution time.
 *
 *  \param pt A pointer to the protothread control structure
 *  \hideinitializer
 *  \sa PT_YIELD_UNTIL
 *  \sa PT_WAIT_ASYIELD_WHILE
 */
#define PT_YIELD(pt) \
  do { \
    PT_YIELD_FLAG = 0; \
    LC_SET((pt)->lc);   \
    if(!PT_YIELD_FLAG) return PT_YIELDED; \
  } while(0)

/** Yield from the protothread until a condition occurs
 *
 *  This function will yield the protothread, until the specified condition
 *  evaluates to true.
 *
 *  The thread yields at least once, even if the condition \c cond is
 *  true at first arrival. This is a bit counter-intuitive but may be
 *  wanted in some circumstances; i.e if a cut / subdivision of the task's
 *  runtime is due anyway.
 *
 *  If no yield should take place if the condition is yet true (behaviour
 *  like  \c PT_WAIT_UNTIL but with other returned state) use
 *  \c PT_WAIT_ASYIELD_WHILE with the inverse condition.
 *
 * \param pt   A pointer to the protothread control structure
 * \param cond The condition.
 * \hideinitializer
 * \sa PT_YIELD
 */
#define PT_YIELD_UNTIL(pt, cond)   \
  do { \
    PT_YIELD_FLAG = 0;   \
    LC_SET((pt)->lc);   \
    if (!(cond)) return PT_YIELDED;  \
    if (!PT_YIELD_FLAG) return PT_YIELDED; \
  } while(0)

/** Yield while a condition is true
 *
 *  This macro yields the protothread until and only if the specified
 *  condition is true.
 -  The difference to \ref PT_WAIT_WHILE is the returned state. <br />
    Semantically this is a wait returning \ref PT_YIELDED.
 -  The difference to \ref PT_YIELD_UNTIL is the more intuitive behaviour;
    i.e not yielding or just going ahead if \c condition is false from start.
 *
 *  \param pt        pointer to the protothread control structure
 *  \param condition the condition (a boolean expression).
 *  \sa PT_OR_YIELD_HERE
 *  \hideinitializer
 */
#define PT_YIELD_WHILE(pt, condition) PT_YIELD_UNTIL(pt, !(condidion))

/** Wait while a condition is true  mimicked as yield
 *
 *  This macro yields the protothread until and only if the specified
 *  condition is true.
 -  The difference to \ref PT_WAIT_WHILE is the returned state. <br />
    Semantically this is a wait returning \ref PT_YIELDED.
 -  The difference to \ref PT_YIELD_WHILE is the behaviour of
    not yielding or just going ahead if \c condition is false from start.
 *
 *  \param pt        pointer to the protothread control structure
 *  \param condition the condition (a boolean expression).
 *  \sa PT_OR_YIELD_HERE
 *  \hideinitializer
 */
#define PT_WAIT_ASYIELD_WHILE(pt, condition) \
  do { \
    PT_YIELD_FLAG = 0; \
    LC_SET((pt)->lc); \
    if(condition)  return PT_YIELDED; \
  } while(0)

/** Yield only if a previous conditional wait did not
 *
 *  This macro yields the protothread if and only if a previous conditional
 *  wait or wait_as_yield yield did not do so and just went ahead.
 *
 *  \param pt  pointer to the protothread control structure
 *  \sa PT_WAIT_ASYIELD_WHILE
 *  \sa PT_OR_YIELD_REENTER()
 *  \hideinitializer
 */
#define PT_OR_YIELD_HERE(pt) \
  do { \
    LC_SET((pt)->lc);  \
    if(!PT_YIELD_FLAG)  return PT_YIELDED; \
  } while(0)

/** Yield only if a previous conditional wait did not and re-enter that
 *
 *  This macro yields if and only if a previous conditional wait or
 *  wait_as_yield macro did not do so and just went ahead. In that case
 *  this \ref PT_YIELDED "yields" the protothread.
 *
 *  Contrary to the else similar \ref PT_OR_YIELD_HERE the resume point for
 *  the next schedule is <b>not</b> at this macro's (\ref PT_OR_YIELD_REENTER)
 *  call but (back) at the previous block or yield. The rationale lies in
 *  those cases where that previous block or yield macro's conditions must be
 *  rechecked or its other (side) effects have to be repeated.
 *
 *  Hint: As this behavior is not intelligibly to the reader of the source
 *  code a clarifying comment is strongly recommended: \code
      PT_YIELD_OUT_SPACE(&cliThread->pt, 31, cliThread->repStreams); // (A)
      // if and only if (A) didn't yield, do yield here and now and
      PT_OR_YIELD_REENTER(); // re-enter at (A) above on next schedule
 *  \endcode
 *
 *  \sa PT_WAIT_ASYIELD_WHILE
 *  \sa PT_OR_YIELD_HERE
 *  \hideinitializer
 */
#define  PT_OR_YIELD_REENTER() \
  do { \
    if(!PT_YIELD_FLAG)  return PT_YIELDED; \
  } while(0)


/** @} */
#endif /* __PT_H__ */
/** @} */
/** @} */ // osSWinc
