/**
 * \addtogroup uip
 * @{
 *
 * \addtogroup uiparp
 * @{
 */
 
/**
 * \file
 *
 * Implementation of the Address Resolution Protocol ARP
 *
 * \author Adam Dunkels <adam@dunkels.com>
 *
 * $Revision: 2 $
 * $Date: 2017-01-25 16:49:03 +0000 (Wed, 25 Jan 2017) $
 *
 */

/*
 * Copyright (c) 2001-2003, Adam Dunkels.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * This file is part of the uIP TCP/IP stack.
 */

#include "uip/uip_arp.h"

#include "we-aut_sys/common.h"
#include "we-aut_sys/network.h"
/* The ARP table */
struct arp_entry arp_table[UIP_ARPTAB_SIZE];

#include <string.h>


#if DEBUG_ARP
# include "we-aut_sys/log_streams.h"
// # include "we-aut_sys/uart0.h"
# include "we-aut_sys/utils.h"
//                             0123456 89.123456789.123
INFLASH(char const uipIPinMes[])  = "  uIP ARP IP in \n";
INFLASH(char const uipARPinMes[]) = "  uIP ARP pk in \n";
INFLASH(char const uipARPentry[]) = "  uIP ARP entry : ";
#endif


const struct uip_eth_addr broadcast_ethaddr =
  {{0xff,0xff,0xff,0xff,0xff,0xff}};
const uint16_t broadcast_ipaddr[2] = {0xffff,0xffff};



/* Change by Albrecht Weinert (A.W.):
 * We won't use arptime (the tick) == 0 any longer.
 * So it won't get used as ->time (the stamp).
 * As positive consequence ->time==0 means unused / cleared entry which
 * eliminates the 32bit all 0 / one bit non 0 hazzle.
 * The negative side is a one tick (10s default) inaccuracy of the (default)
 * 20 minutes.
 */
static uint8_t arptime;

#define BUF   ((struct arp_hdr *)uip_buf_alias)

/*-------------------------------------------------------------------------*/

/* Initialise the ARP module  */
void uip_arp_init(){
  arptime = 1;
  memset(arp_table, 0, sizeof(arp_table));
  arpTimeOut = arpTickPeriod = 10;
} //uip_arp_init()


/* Periodic ARP processing function
 *
 * This function performs periodic timer processing in the ARP module
 * and should be called at regular intervals. The recommended frequency is
 * every 10 seconds.
 */
void uip_arp_timer(void){
  struct arp_entry * tabptr;
  if (!(++arptime)) arptime = 1;
  for(uint8_t i = 0; i < UIP_ARPTAB_SIZE; ++i) {
    tabptr = &arp_table[i];
    if (tabptr->time) { // process used entry only
       const uint8_t actAge = arptime - tabptr->time;
       if(actAge >= UIP_ARP_MAXAGE) tabptr->time = 0;
    } // process used entry only
  } // for
} // uip_arp_timer(void)


static void uip_arp_update(uint16_t * ipaddr, struct uip_eth_addr * ethaddr){
   uint8_t i = 0;
   struct arp_entry *tabptr; // search pointer
   struct arp_entry *emptySlot = NULL; // found empty or old pointer

   // Walk through the ARP mapping table and try to find an entry to update.
   // If none is found, a new IP -> MAC address mapping is inserted
   for(i = 0; i < UIP_ARPTAB_SIZE; ++i) {
      tabptr = &arp_table[i];
      // Only check those entries that are actually in use
      if(tabptr->time) { // in use
         /* Check if the source IP address of the incoming packet matches
         the IP address in this ARP table entry. */
         if(ipaddr[0] == tabptr->ipaddr[0]
                  &&  ipaddr[1] == tabptr->ipaddr[1]) {
            /* An old entry found, update this and return. */
            memcpy(tabptr->ethaddr.addr, ethaddr->addr, 6);
            tabptr->time = arptime;
            return;
         }
      } else if (emptySlot == NULL) {
         emptySlot = tabptr;
      }
   }  // for search existing entry

   // If we get here, no used ARP table entry matching ipaddr was found.
   // So we create one in an unused slot (already found) or we overwrite
   // the oldest in use.
   if(emptySlot == NULL) {
      uint8_t tmpage = 0;
      for(i = 0; i < UIP_ARPTAB_SIZE; ++i) { // search oldest
         tabptr = &arp_table[i];
         const uint8_t actAge = arptime - tabptr->time;
         if(actAge > tmpage) {
            tmpage = actAge;
            emptySlot = tabptr;
         }
      }  // for search oldest
   }

   // Now, emptySlot is the ARP table entry  we will fill with the new info
#if DEBUG_ARP > 1
   bufLogTxt_P(uipARPentry);
   //            0123456789.123456789.123456789.12 3456789.
   char tmp[17];
   //            11:22:33:44:55:66
   //            123.123.123.123
   formMacAdd(tmp, ethaddr);
   //  while (uartOutSpace() < 38) ;
   bufLogTxt(tmp, 17);
   bufLogPutC(' ');
   formIpAdd(tmp, ipaddr);
   bufLogTxt(tmp, 15);
# if DEBUG_ARP > 2 // show pointer and tick
   bufLogPutC(' ');
   fourHexs(tmp, (uint16_t)emptySlot);
   tmp[4] = '-';
   twoHexs(&tmp[5], arptime);
   bufLogTxt(tmp, 7);
# endif
   bufLogPut2C(' ', '\n'); // 1 blank 1 LF
#endif

   memcpy(emptySlot->ipaddr, ipaddr, 4);
   memcpy(emptySlot->ethaddr.addr, ethaddr->addr, 6);
   emptySlot->time = arptime;
} // uip_arp_update()


/*  ARP processing for incoming IP packets
 *
 * This function should be called by the device driver when an IP
 * packet has been received. The function will check if the address is
 * in the ARP cache, and if so the ARP cache entry will be
 * refreshed. If no ARP cache entry was found, a new one is created.
 *
 * This function expects an IP packet with a prepended Ethernet header
 * in the uip_buf[] buffer, and the length of the packet in the global
 * variable uip_len.
 */
void uip_arp_ipin(void){

#if DEBUG_ARP > 2
   bufLogTxt_P(uipIPinMes); // uIP ARP IP in
#endif
	
  /* Only insert/update an entry if the source IP address of the
     incoming IP packet comes from a host on the local network. */
  if(((IPBUF->srcipaddr[0] ^ uip_hostaddr[0]) & uip_netmask[0])
    || ((IPBUF->srcipaddr[1] ^ uip_hostaddr[1]) & uip_netmask[1])) {
    return;
  }
  uip_arp_update(IPBUF->srcipaddr, &(IPBUF->ethhdr.src));
} // uip_arp_ipin()



/* ARP processing for incoming ARP packets
 *
 * This function should be called by the device driver when an ARP
 * packet has been received. The function will act differently
 * depending on the ARP packet type: if it is a reply for a request
 * that we previously sent out, the ARP cache will be filled in with
 * the values from the ARP reply. If the incoming ARP packet is an ARP
 * request for our IP address, an ARP reply packet is created and put
 * into the uip_buf[] buffer.
 *
 * When the function returns, the value of the global variable uip_len
 * indicates whether the device driver should send out a packet or
 * not. If uip_len is zero, no packet should be sent. If uip_len is
 * non-zero, it contains the length of the outbound packet that is
 * present in the uip_buf[] buffer.
 *
 * This function expects an ARP packet with a prepended Ethernet
 * header in the uip_buf[] buffer, and the length of the packet in the
 * global variable uip_len.
 */
void uip_arp_arpin(void){

#if DEBUG_ARP > 1
   bufLogTxt_P(uipARPinMes);
#endif
  
  if(uip_len < sizeof(struct arp_hdr)) {
    uip_len = 0;
    return;
  }
  uip_len = 0;
  
  switch(BUF->opcode) {
  case HTONS(ARP_REQUEST):
    /* ARP request. If it asked for our address, we send out a
       reply. */
    if(uip_ipaddr_cmp(BUF->dipaddr, uip_hostaddr)) {
      /* First, we register the one who made the request in our ARP
	 table, since it is likely that we will do more communication
	 with this host in the future. */
      uip_arp_update(BUF->sipaddr, &BUF->shwaddr);
      
      /* The reply opcode is 2. */
      BUF->opcode = HTONS(2);
      //      dest            , src              , len
      memcpy(BUF->dhwaddr.addr, BUF->shwaddr.addr, 6);
      memcpy(BUF->shwaddr.addr, actMACadd.addr, 6);
      memcpy(BUF->ethhdr.src.addr, actMACadd.addr, 6);
      memcpy(BUF->ethhdr.dest.addr, BUF->dhwaddr.addr, 6);
      
      BUF->dipaddr[0] = BUF->sipaddr[0];
      BUF->dipaddr[1] = BUF->sipaddr[1];
      BUF->sipaddr[0] = uip_hostaddr[0];
      BUF->sipaddr[1] = uip_hostaddr[1];

      BUF->ethhdr.type = HTONS(UIP_ETHTYPE_ARP);
      uip_len = sizeof(struct arp_hdr);
    }
    break;
  case HTONS(ARP_REPLY):
    /* ARP reply. We insert or update the ARP table if it was meant
       for us. */
    if(uip_ipaddr_cmp(BUF->dipaddr, uip_hostaddr)) {
      uip_arp_update(BUF->sipaddr, &BUF->shwaddr);
    }
    break;
  }
} // uip_arp_arpin(void)


/* Prepend Ethernet header to an out bound IP packet and see if we need
 * to send out an ARP request
 */
void uip_arp_out(void){

   struct arp_entry *tabptr = NULL;
  /* Find the destination IP address in the ARP table and construct
     the Ethernet header. If the destination IP address isn't on the
     local network, we use the default router's IP address instead.

     If not ARP table entry is found, we overwrite the original IP
     packet with an ARP request for the IP address. */

  /* First check if destination is a local broadcast. */
  if(uip_ipaddr_cmp(IPBUF->destipaddr, broadcast_ipaddr)) {
    memcpy(IPBUF->ethhdr.dest.addr, broadcast_ethaddr.addr, 6);
  } else {
    /* Check if the destination address is on the local network. */
     static uint16_t ipaddr[2];
     if(!uip_ipaddr_maskcmp(IPBUF->destipaddr, uip_hostaddr, uip_netmask)) {
      /* Destination address was not on the local network, so we need to
	 use the default router's IP address instead of the destination
	 address when determining the MAC address. */
      uip_ipaddr_copy(ipaddr, uip_draddr);
    } else {
      /* Else, we use the destination IP address. */
      uip_ipaddr_copy(ipaddr, IPBUF->destipaddr);
    }
      
    uint8_t notFound = 1;
    for(uint8_t i = 0; i < UIP_ARPTAB_SIZE; ++i) {
       tabptr = &arp_table[i];
       if(tabptr->time && uip_ipaddr_cmp(ipaddr, tabptr->ipaddr)) {
          notFound = 0;
          break;
       }
    }

    if(notFound) {
      /* The destination address was not in our ARP table, so we
	 overwrite the IP packet with an ARP request. */

      memset(BUF->ethhdr.dest.addr, 0xff, 6);
      memset(BUF->dhwaddr.addr, 0x00, 6);
      memcpy(BUF->ethhdr.src.addr, actMACadd.addr, 6);
      memcpy(BUF->shwaddr.addr, actMACadd.addr, 6);
    
      uip_ipaddr_copy(BUF->dipaddr, ipaddr);
      uip_ipaddr_copy(BUF->sipaddr, uip_hostaddr);
      BUF->opcode = HTONS(ARP_REQUEST); /* ARP request. */
      BUF->hwtype = HTONS(ARP_HWTYPE_ETH);
      BUF->protocol = HTONS(UIP_ETHTYPE_IP);
      BUF->hwlen = 6;
      BUF->protolen = 4;
      BUF->ethhdr.type = HTONS(UIP_ETHTYPE_ARP);

      uip_appdata = &uip_buf[UIP_TCPIP_HLEN + UIP_LLH_LEN];
    
      uip_len = sizeof(struct arp_hdr);
      return;
    }

    /* Build an Ethernet header. */
    memcpy(IPBUF->ethhdr.dest.addr, tabptr->ethaddr.addr, 6);
  }
  memcpy(IPBUF->ethhdr.src.addr, actMACadd.addr, 6);
  
  IPBUF->ethhdr.type = HTONS(UIP_ETHTYPE_IP);

  uip_len += sizeof(struct uip_eth_hdr);
} //  uip_arp_out(void)


/** @} */
/** @} */

