//  Copyright (c) 2014 Albrecht Weinert, Bochum
//  cli.c 2 2017-01-25 16:49:03Z albrecht

/**
 * \addtogroup weAutSys
 * @{
 * \addtogroup sysSer
 * @{
 *
 * \addtogroup sysThr
 * @{
 *
 * \file cli.c
 *
 * weAutSys implementation of the command line interpreter (CLI)
 *
 * This is system software and must not be modified for user or
 * application programs.
 *
 * Please find detailed description of the variables, functions etc.
 * in the module \ref sysThr "system threads" and in the include file
 * syst_threads.h.
 *
 * This file is part of weAutSys &nbsp;
 *          <<a href="http://weinert-automation.de">weinert-automation.de</a>>
 *
 * &nbsp; \ref copyLicNote " Copyright &copy; " 2014 Albrecht Weinert, Bochum
 \author  Albrecht Weinert  &nbsp; <<a href="http://a-weinert.de">a-weinert.de</a>>
 * 2
 * 25.01.2017
 */

/** @} */ // doxygen addtogroup sysThr
/** @} */ // doxygen addtogroup sysSer
/** @} */ // doxygen addtogroup weAutSys


#include <string.h>

#include "we-aut_sys/common.h"
#include "we-aut_sys/utils.h"
#include "we-aut_sys/ll_system.h"
#include "we-aut_sys/system.h"
#include "we-aut_sys/cli.h"
#include "we-aut_sys/smc2fs.h"
#include "we-aut_sys/log_streams.h" // for debug only
#include "we-aut_sys/streams.h"
#include "we-aut_sys/network.h"
#include "we-aut_sys/persist.h"
#include "we-aut_sys/enc28j60.h"
#include "we-aut_sys/smc.h"

#include "we-aut_sys/ntp.h"
#include "uip/apps/dhcpc.h"
#include "uip/apps/resolv.h"

/* The user / application command line interpreter (CLI) registration */
struct appCLIreg_t appCLIreg;

/* Register a user / application command line interpreter (CLI) */
void registerAppCli(p2ptFunA threadF, char const * const userCommands[]){
   if (!threadF || !userCommands) {
      appCLIreg.fun = NULL;
      appCLIreg.userCommands = NULL;
   } else {
      appCLIreg.fun = threadF;
      appCLIreg.userCommands = userCommands;
   }
} // registerAppCli(p2ptFun)

/* Register a  command line interpreter (CLI) thread  */
void initAsCLIthread(struct thr_data_t * thread, FILE * file){
   // init all to 0, NULL, false
   memset(&thread->cliThrData, 0, sizeof(struct cliThr_data_t));
   thread->cliThrData.repStreams = file;
   thread->cliThrData.commandStart = thread->cliThrData.commandEnd = 255;
} // registerAppCliThread(p2ptFun)


/* Set a command line in the command line interpreter (CLI) data
 *
 *  The given line if any will be set in the \ref appCliThread data, the
 *  command (first) token will be determined and the system and user CLI
 *  \ref appCliThread "thread state" will be forced to empty initial state.
 *
 *  Do not call this function while system or user CLI threads are active.
 *
 * \param  cliThread pointer to the CLI thread to interpret line
 * \param  line the command line (0-terminated RAM string)
 * \param  length line's length (line[length] should nevertheless better be 0)
 * \return the command number
 * see:  appCliThreadF  appInitThreadF
 */
uint8_t setCliLine(struct cliThr_data_t * const cliThread,
                                         char * const line, uint16_t length){
   cliThread->commandStart  =  cliThread->commandEnd
                = cliThread->paramStart  = cliThread->paramEnd
                                        = cliThread->optionNumb = 255;
   PT_INIT(&cliThread->pt);
   PT_INIT(&cliThread->systCLIpt);
   cliThread->commNumb = cliThread->length = 0;
   if (!line || length < 2) return 255;

   if (length > LEN_OF_CLITHR_LINE) length = LEN_OF_CLITHR_LINE;
   if (line != cliThread->line) {
      memcpy(cliThread->line, line, length);
   }
   cliThread->length = length;
   cliThread->line[length] = 0; // force terminating 0 (reserve)

   uint8_t tokStart = searchFirstToken(line, 0, length);
   if (tokStart == 255)  return 255; // no first token / no command
   uint8_t tokEnd   = searchTokenEnd(cliThread->line, tokStart, length);

   if ((tokEnd - tokStart) < 2) return 255; // no 1-letter command

   cliThread->commandStart = tokStart;
   cliThread->commandEnd = tokEnd;
   cliThread->commNumb = searchTokenIn(cliThread->line, tokStart, tokEnd,
                                    appCLIreg.userCommands,  systemCommands);
   tokStart = searchTokenStart(cliThread->line, tokEnd, length);
   if (tokStart != 255) { // one parameter found
      tokEnd = searchTokenEnd(cliThread->line, tokStart, length);
      cliThread->paramStart = tokStart;
      cliThread->paramEnd = tokEnd;
      if (tokEnd != tokStart && cliThread->line[tokStart] == '-'
            && (tokEnd-2) >= tokStart && cliThread->line[tokStart+1] >= '?') { // option
         cliThread->optionNumb =  searchTokenIn(cliThread->line,
               tokStart+1, tokEnd, systemOptions,  NULL);
         if (cliThread->optionNumb < optAmbigousNum ) { // option found
            tokStart = searchTokenStart(cliThread->line, tokEnd, length);
            if (tokStart != 255) {
               tokEnd = searchTokenEnd(cliThread->line, tokStart, length);
               cliThread->paramEnd = tokEnd; // will index end of param | opt
            }
            cliThread->paramStart = tokStart; // will index next param or 255
         } // option found
      } // may be option
   }  // one parameter found
   return cliThread->commNumb;
} // setCliLine(char *, uint8_t)


/*  Report the use of an un-implemented option for a command
 *
 *  This is a helper function for the implementation on CLIs (command line
 *  processors) only. This function requires stdout connected to CLI streams
 *  and enough space.
 *
 *  \param  cliThread pointer to the CLI (and hence the thread busy
 *          to execute actual "command -wrongOpt)
 */
void unimplOptionReport(struct cliThr_data_t * cliThread) {
   if (!cliThread->repStreams) return; // out
   stdPutS_P(optWrong);
   //  //           0123456789x1234 5
   char onimpl[] = "-1        ... \n";
   char * dest = onimpl + 1;
   char * src = (char *)pgm_read_word(&systemOptions[cliThread->optionNumb]);
   uint8_t i = 8;
   char c = pgm_read_byte(src);
   while (c > ' ' && --i) {
      *(dest++) = c;
      c = pgm_read_byte(++src);
   }
   stdPutS(onimpl);
}  // unimplOptionReport(struct cliThr_data_t *)

INFLASH(char const notAtUARTwarn[])
                = " *  Warning: You'r not connected by UART  \n";
INFLASH(char const notAtTelnetWarn[])
                = " *  Warning: This is'nt Ethernet / Telnet \n";

/* This is \ref PT_YIELD_OUT_SPACE copied and modified for use within
 * \ref sysCliThreadF only.
 * Modifications:
 * a) param pt replaced by fixed &cliThread->systCLIpt
 * b) param streams replaced by fixed cliThread->repStreams
 * d) when exiting added cliThread->commNumb = 0;
 * Effect:
 * PT_YIELD_OUT_SPACE(&cliThread->systCLIpt, 41, cliThread->repStreams);
 * PT_YIELD_SPACE4SYS_CLI(41);
 *
 */
#define PT_YIELD_SPACE4SYS_CLI(requSpace) \
  do { \
     PT_YIELD_FLAG = 0; \
     LC_SET(cliThread->systCLIpt.lc); \
     if (!switchStdStreams(cliThread->repStreams)) { \
        PT_INIT(&cliThread->systCLIpt); \
        cliThread->commNumb = 0; return PT_EXITED; } \
     const uint8_t cS = checkOutSpace((uint16_t)(requSpace), \
                  cliThread->repStreams); \
     if (cS >= 2) { \
       streamsClose(); \
       if (cS == 2) return PT_YIELDED; \
       if (cS == 4) { PT_INIT(&cliThread->systCLIpt); \
              cliThread->commNumb = 0; return PT_EXITED; }  \
       return PT_WAITING; \
     } \
  } while(0)


/* The system command line interpreter thread  */
ptfnct_t sysCliThreadF(struct cliThr_data_t * cliThread){
   if ( ! cliThread->commNumb) { // nothing to do
      PT_LEAVE(&cliThread->systCLIpt); // exit
   } // nothing to do ( commNumb is the flag)
   PT_BEGIN(&cliThread->systCLIpt); // system CLI protothread start
   SAMPLE_MIN_SP(2);  // sample min. SP once on fresh start (IE, one nest)

   if (switchStdStreams(cliThread->repStreams)) { // reports allowed
      PT_YIELD_SPACE4SYS_CLI(41);
   } // require at least 41 characters out space for all CLI inc. appl. CLI

   if (cliThread->commNumb  > 253  // no or ambiguous command
         || cliThread->optionNumb == 254) { //  //  ambiguous option
      if (cliThread->repStreams) { // reports allowed
         PT_YIELD_SPACE4SYS_CLI(41);
         if (cliThread->optionNumb == 254) {
            stdPutS_P(optAmbig);    // ambiguous option abbreviation
         } else if (cliThread->commNumb == 254) {
            stdPutS_P(commAmbig);  // ambiguous command abbreviation
         } else {
            stdPutS_P(commWrong); // non existing command
         }

         PT_YIELD_SPACE4SYS_CLI(cliThread->length);
         stdPutS(cliThread->line); // just echo for error report
         PT_YIELD_SPACE4SYS_CLI(5);
         stdPutS_P(bLF1); // 2 line feed
      } // reports allowed
      goto cliNormalEnd;
   } // no or ambiguous command or ambiguous option

   // connect log to current CLI device if UART and if not yet connected
   if (!bufLogCons) {
      if (cliThread->repStreams == &serStreams) { // we use the UART
         bufLogCons = (struct thr_data_t *)1;
      } //
   } // no log out yet

   if (cliThread->commNumb < INDEX_OFFSET_LIST2) { // user command
      if (cliThread->optionNumb <= optHelpNum) { // -?, -help
         initOutFlashTextThread(&cliThread->outFlashTextThread,
               &(appCLIreg.userCommands[cliThread->commNumb]), -1);

         PT_YIELD_SPACE4SYS_CLI(UART_OUT_BUF_CAP-10); // re-entry point <<<
         const uint8_t outFlashErg = outFlashTextThreadF(
               &cliThread->outFlashTextThread,  cliThread->repStreams);
         if (outFlashErg <= PT_YIELDED) return outFlashErg; // re-ent ^^^
         if (outFlashErg == PT_EXITED)  goto cliNormalEnd;  // can't go on

         PT_YIELD_SPACE4SYS_CLI(5);
         stdPutS_P(bLF1); // blank linefeed
         goto cliNormalEnd; // ready
      } // -?, -help

      if (!appCLIreg.fun) { //  no user CLI function registered
         stdPutS_P(noUserCLI); // Error user command w/o interpreter
         goto cliNormalEnd;
      } //  no user CLI function registered

      PT_INIT(&cliThread->pt);  // init user CLI thread

      while (cliThread->commNumb // as long as command not done
                        &&  cliThread->commNumb < INDEX_OFFSET_LIST2 ) {
         if (switchStdStreams(cliThread->repStreams)) { // reports allowed
            PT_YIELD_SPACE4SYS_CLI(43);
         } // guarantee at least 43 characters out space for appl. CLI
         const uint8_t retAppCLI = ((p2ptFunA)appCLIreg.fun)(cliThread);
         if (retAppCLI >= PT_EXITED) break; // while app cli done
         PT_YIELD(&cliThread->systCLIpt); // yield this if app yields | waits
      } // as long as (app) command not done: loop around app CLI schedules
      goto cliInfoEnd;
   } // user command    (end, below system command) -------------------------


   if (cliThread->optionNumb <= optHelpNum) { // -?, -help -----------  -help
      initOutFlashTextThread(&cliThread->outFlashTextThread,
             &(systemCommands[cliThread->commNumb - INDEX_OFFSET_LIST2]), -1);

      PT_YIELD_SPACE4SYS_CLI(UART_OUT_BUF_CAP - 10); // re-entry point <<<
      const uint8_t outFlashErg = outFlashTextThreadF(
            &cliThread->outFlashTextThread,  cliThread->repStreams);
      if (outFlashErg <= PT_YIELDED) return outFlashErg; // re-ent ^^^
      if (outFlashErg == PT_EXITED)  goto cliNormalEnd;  // can't go on

      PT_YIELD_SPACE4SYS_CLI(5);
      stdPutS_P(bLF1); // blank linefeed
      goto cliNormalEnd; // ready
   } // -?, -help

   /*  current system commands
     comHelp,  // 1
     comWatchD, comWDog,  bLF1,   // 2 3 (4 is gap)
     comBoot,  comPLContrl, comAutomation, // 5 6 7
     comVersion,    comRunInfo,    // 8 9
     comMacAddr, comEthAddr,     // 10 11
     comIPconfig,  comIPdefault,  // 12 13
     comUARTcnt, // 14
     comTime, ComDate, // 15, 16
     comZone, comNTP, // 17, 18
     comDHCP,   // 19
     comDNS,  // 20
    */
   switch (cliThread->commNumb - INDEX_OFFSET_LIST2) {
      case comHelpNum : // help optSystem  optApplic  optOption   ----- help -
         if (!cliThread->repStreams) break;
         stdPutS_P(bLF1);
         if (cliThread->optionNumb != optApplicNum
               && cliThread->optionNumb != optOptionNum) {
            initOutFlashTextThread(&cliThread->outFlashTextThread,
                  systemCommands, 127);

            PT_YIELD_SPACE4SYS_CLI(UART_OUT_BUF_CAP-10); // re-entry point <<<
            const uint8_t outFlashErg = outFlashTextThreadF(
                  &cliThread->outFlashTextThread,  cliThread->repStreams);
            if (outFlashErg <= PT_YIELDED) return outFlashErg; // re-ent ^^^
            if (outFlashErg == PT_EXITED)  goto cliNormalEnd;  // can't go on

            PT_YIELD_SPACE4SYS_CLI(5);
            stdPutS_P(bLF1); // blank linefeed
         }

         if (cliThread->optionNumb != optSystemNum
               && cliThread->optionNumb != optOptionNum) {
            initOutFlashTextThread(&cliThread->outFlashTextThread,
                  appCLIreg.userCommands, 127);

            PT_YIELD_SPACE4SYS_CLI(UART_OUT_BUF_CAP-10); // re-entry point <<<
            const uint8_t outFlashErg = outFlashTextThreadF(
                  &cliThread->outFlashTextThread,  cliThread->repStreams);
            if (outFlashErg <= PT_YIELDED) return outFlashErg; // re-ent ^^^
            if (outFlashErg == PT_EXITED)  goto cliNormalEnd;

            PT_YIELD_SPACE4SYS_CLI(5);
            stdPutS_P(bLF1); // blank linefeed
         }

         if (cliThread->optionNumb == optOptionNum
               || cliThread->optionNumb == optVerboseNum) {
            initOutFlashTextThread(&cliThread->outFlashTextThread,
                  systemOptions, 127);

            PT_YIELD_SPACE4SYS_CLI(UART_OUT_BUF_CAP-10); // re-entry point <<<
            const uint8_t outFlashErg = outFlashTextThreadF(
                  &cliThread->outFlashTextThread,  cliThread->repStreams);
            if (outFlashErg <= PT_YIELDED) return outFlashErg; // re-ent ^^^
            if (outFlashErg == PT_EXITED)  goto cliNormalEnd;

            PT_YIELD_SPACE4SYS_CLI(5);
            stdPutS_P(bLF1); // blank linefeed
         }  // -option or -verbose
         break;

      case comWatchDNum : // wd fall through
      case comWDogNum : // WDog optAbort optSharp             ----- watchdog -
         wdt_reset(); // re-trigger for fairness' sake
         char const * repWd = wdSetNormal;

         if (cliThread->optionNumb == optAbortNum) { // option abort
            repWd = wdSetShort;
            wdt_enable(WDTO_60MS);
         } else if (cliThread->optionNumb == optSharpNum) { // sharp
            repWd = wdSetSharp;
            wdt_enable(WDTO_120MS);
         } else if (cliThread->optionNumb == optLenientNum) { // lenient
            repWd = wdSetLeni;
            wdt_enable(WDTO_500MS);
         } else {   // all else sets normal
            wdt_enable(WDTO_250MS);
         }
         wdt_reset(); // re-trigger for fairness' sake
         if (cliThread->repStreams) {
            stdPutS_P(repWd);
            stdPutS_P(bLF1); // blank linefeed
         }
         break;

      case comBootNum :  // boot -load                            ----- boot -
         if (cliThread->optionNumb == optLoadNum ) {
            setAbortCommand(abortBoot);
         } else if (cliThread->optionNumb == optResetNum) {
            setAbortCommand(abortHMI);
         }
         break;

      case comPLContrlNum : // PLC  fall through
      case comAutomationNum : // Automation                        ----- PLC -
         if (cliThread->optionNumb >= optOffNum
               && cliThread->optionNumb <= optAbortNum) { // option abort
            setAbortCommand(PLCstop); // PLC stop
            break;
         } // option off stop reset abort
         if (cliThread->optionNumb >= optOnNum
               && cliThread->optionNumb <= optRestartNum) { // option on
            setAbortCommand(0);  // PLC start
         }
         break;

      case comVersionNum : // version greeting                 ----- version -
         if (!cliThread->repStreams) break;

         PT_INIT(&cliThread->outFlashTextThread.pt);
         PT_YIELD_SPACE4SYS_CLI(90); // re-entry point <<<
         const uint8_t outFlashErg =
               outSysInfoThreadF(&cliThread->outFlashTextThread.pt,
                     cliThread->repStreams);

         if (outFlashErg <= PT_YIELDED) return outFlashErg; // re-ent ^^^
         if (outFlashErg == PT_EXITED)  goto cliNormalEnd;
         break;

      case comRunInfoNum : //  runInfo                             ----- run -
         if (!cliThread->repStreams) break;
         cliThread->infoThread.flag |= 0x10; // Bit 4 run info
          break;

      case comIPconfigNum : // comIPconfig                         ----- ipc -
         if (!cliThread->repStreams) break;
         cliThread->infoThread.flag2 |= 3; // Bit 0 MAC, Bit 1 IPset
         goto makeLFnBreak;
         //  break; // comDHCP

      case comEthAddrNum : // comEthAddr                           ----- eth -
         if (!cliThread->repStreams) break;
         cliThread->infoThread.flag2  |= 9; // Bit 0 MAC 3 ARP
         goto makeLFnBreak;
         //  break; // comDHCP


      case comIPdefAdNum: // 13 IPDefault                          ----- ipd -
         if (cliThread->paramStart  != optNotGivenNum) {
            parseIpAdd(defaultConfData.ipConf.ipAddr, cliThread->line,
                  cliThread->paramStart); // default IP addr
         }
         if (!cliThread->repStreams) break;
         cliThread->infoThread.flag2  |= 4; // Bit 2 IPdef
         goto makeLFnBreak;
         //  break; // comDHCP

      case comMacAddrNum : //  MacAddr                            ----- mac -
         if (cliThread->paramStart  != optNotGivenNum) {
            parseMACadd(&actMACadd, cliThread->line,
                  cliThread->paramStart); // actual MAC addr
         }
         if (!cliThread->repStreams) break;
         cliThread->infoThread.flag2  |= 0x11; // Bit 0 MAC, Bit 4 Link state
         goto makeLFnBreak;
         //  break; // comDHCP

      case comTelnetNum:    //                                     ----- tel -
         // todo add some options (listen unlisten etc. here)
         if (!cliThread->repStreams) break;
         if (cliThread->repStreams != &lanStreams) { // we're not at Ethernet
            stdPutS_P(notAtTelnetWarn);
         } //  this can't be a Telnet connection
         break; //comTelnetNum


      case comUARTNum : // comUART                                 ----- uar -
         if (cliThread->optionNumb == optFlowCnNum) { // flowControl
            if (cliThread->paramStart  != optNotGivenNum) {
               setUARTflowcontrolByChar(cliThread->line[cliThread->paramStart]);
            }
         } else if (cliThread->optionNumb == optNoLogsNum) { // noLogs on UART
            if (bufLogConsIsUART) bufLogCons = NULL;
         } else if (cliThread->optionNumb == optLogHereNum) { // log on UART
            bufLogSetConsumer(1);
         } // log on UART
         if (!cliThread->repStreams) break;
         if (cliThread->repStreams != &serStreams) { // we use not the UART
            stdPutS_P(notAtUARTwarn);
         } // we're not at the UART here

         cliThread->infoThread.flag |= 0x04; //  *  Bit  2 : UART info
         break;

      case comTimeNum : // comTime                                 ----- tim -
         if (cliThread->paramStart  != optNotGivenNum) { // have a parameter
            datdur_t tim;
            if (parseTim(&tim, cliThread->line, cliThread->paramStart)) {
               // bufLogDecHex_P(testMeld,11, 1, secDatTimers.end.v32);
               tim.d = secDatTimers.end.v32 / SECONDS_IN_DAY;
               uint32_t newSet = datdur2sec(&tim);
               if (secDatTimers.end.v32 != newSet) {
                  secDatTimers.end.v32 = newSet;
                  msAbsClockCount = 0;
                  if (! (++adjustUTCcount)) adjustUTCcount = 1;
                  sys1sPeriodFlag |= ABSsecSET;
               }
               //  bufLogDecHex_P(testMeld,11, 4, secDatTimers.end.v32);
            }
            //  //          0123456789x123456789v12345 6
            //char s[] = "test parse time hh:mm:ss \n";
            //  formTimOfD(&s[16], &tim);
            //bufLogTxt(s, 26);
         } // have a parameter
         if (!cliThread->repStreams) break;
         cliThread->infoThread.flag |= 0x20; //   Bit 5 : current local date
         break;  // comTime
      case comDateNum: //  comDate                                 ----- dat -
         if (cliThread->paramStart  != optNotGivenNum) { // have a parameter
            date_t dat;
            if (parseDate(&dat, cliThread->line, cliThread->paramStart)) {
               // bufLogDecHex_P(testMeld,11, 1, secDatTimers.end.v32);
               secDatTimers.end.v32 = secDatTimers.end.v32 % SECONDS_IN_DAY;
               // bufLogDecHex_P(testMeld,11, 2, secDatTimers.end.v32);
               secDatTimers.end.v32 +=
                     mul16with17(getDaysByDat(&dat), (uint16_t)SECONDS_IN_DAY);
               sys1sPeriodFlag |= ABSsecSET;
               // bufLogDecHex_P(testMeld,11, 3, secDatTimers.end.v32);
            }
#if 0
            //          0123456789x123456789v123456789d1  2
            char s[] = "test parse date DoW 2009-01-31  \n";
            formWdShort(&s[16], dat.wd);
            formDateIso(&s[20], &dat);
            bufLogTxt(s, 33);
#endif
         } // have a parameter
         if (!cliThread->repStreams) break;
         cliThread->infoThread.flag |= 0x40; // Bit 6 : current local time
         break; //  comDate

      case comZoneNum:  //  comZone
         if (!cliThread->repStreams) break;
         cliThread->infoThread.flag |= 0x80; // Bit 7 : zone info
         break; //  comZone

      case comNTPNum :  //  comNTP  ---------------------------------  NTP --
#if HAVE_ETHERNET
         if (cliThread->optionNumb >= optSilentNum &&
               cliThread->optionNumb <= optDebugNum) {
            ntpState.logLev = cliThread->optionNumb;  //  3..7
         } else  if (cliThread->optionNumb >= optOffNum
               && cliThread->optionNumb <= optAbortNum) { // option abort
            ntpReset(); // initial state for NTP       // NTP stop
         } else if (cliThread->optionNumb >= optOnNum
               && cliThread->optionNumb <= optRestartNum) { // option start
            if(ntpState.state == NTP_STATE_RESET) {
               ntpInit(0); // NTP (re-)start
            }
         } else if (cliThread->optionNumb == optAlternNum) {
            ntpReset();  ntpInit(2);
         } else if (cliThread->optionNumb == optPrimarNum) {
            ntpReset(); ntpInit(0);
         } else if (cliThread->optionNumb != optNotGivenNum) {
            if (!cliThread->repStreams) goto cliNormalEnd; // out
                          unimplOptionReport(cliThread); // helper
         }
         if (!cliThread->repStreams) break;
         cliThread->infoThread.flag |= 0x08; // Bit 3 : NTP info
#endif // have Ethernet
         break; //  comNTP                                      -----  NTP --

      case comDHCPNum :  // comDHCP
#if HAVE_ETHERNET
         if (cliThread->optionNumb == optRestartNum) {
            if (curIpConf.useFlags & DHCP_MSK) { // use DHCP
               dhcpInit(&actMACadd);
            }
         }
         if (!cliThread->repStreams) break;
         // outEncLanInfoFlag
         cliThread->infoThread.flag2 |= 0x20; // Bit 5 DHCP state
         goto makeLFnBreak;
#else // have Ethernet
         break; // comDHCP
#endif

      case comDNSNum :  // comDNS                                -----  DNS --
#if HAVE_ETHERNET
         // the current experimental state of this command just starts the
         // resolvers (and triggers a debug message as response)
         if (cliThread->paramStart  != optNotGivenNum) { // have a parameter
            cliThread->line[cliThread->paramEnd] = 0;
            PT_YIELD_SPACE4SYS_CLI(75);
            char * const nameH = &cliThread->line[cliThread->paramStart];
            uint16_t * ipAdP = resolv_lookup(nameH);
            if (ipAdP) {
               stdPutS(nameH);
               //            0123456789.123456789z1 2
               char tmp[] = " : .......           \n";
               * formIpAdd(&tmp[3], ipAdP) = ' ';
               stdPutS(tmp);
            } else {
               resolv_query(nameH);
            }
         } //have parameter
         break; // comDNS

         // case comTelnetNum :
         // for LAN connections we will not get here as
         // command will be handled in telnet poll

         // todo handle other cases (which?) here
         // break;
#else // have Ethernet
         break; // comDHCP
#endif
      case comNICNum : // command: NICont [-off | -on | -reset | -[restart]
      case comENCNum : // command: ENC28J60   (same as NICont)
#if HAVE_NIC
         switch (cliThread->optionNumb) {

            case optSlowNum:  //  slow     slow(er) speed 1 MHz
               spi2EtherChipBaud = SPI1MHz;
               break;
            case optNormalNum: // normal   standard  speed 2 MHz
               spi2EtherChipBaud = SPI2MHz;
               break;
            case optFastNum:  // fast speed 5 MHz
               spi2EtherChipBaud = SPI5MHz;
               break;

            case optRestartNum:
               networkInit();
               lanComStdInit();
               // todo extra command
               uip_listen(HTONS(TELNET_PORT)); // start listen on ..  server
               uip_listen(HTONS(MODBUS_PORT)); // start listen on ..  server
               break;

            case optResetNum:
               encReset();
               break;

            case optOffNum:
               encEnablePowersave();
               break;

            case optOnNum:
               encDisablePowersave();
               break;

            default: // unimplemented option
               if (!cliThread->repStreams) goto cliNormalEnd; // out
               unimplOptionReport(cliThread); // helper
               break; // no option or not used here

            case optNotGivenNum:
               break;
         } // switch optionNumb

         if (!cliThread->repStreams) break;
         cliThread->infoThread.flag2  |= 0x10; // Bit 4 Link state
         if (spi2Errors) cliThread->infoThread.flag |= 0x02; // Bit 1  SPI 2 state
         goto makeLFnBreak;
#endif // have NIC
         break;

      case comSMCNum: //!<  command: SMCard  [ -on | -start | reset | -rest...
#if HAVE_SMC
         if (cliThread->repStreams) {
            PT_YIELD_SPACE4SYS_CLI(35);
         }
         switch (cliThread->optionNumb) {

            case optStartNum:  // send 80 clocks with standard speed
               clk80();          // 80 dummy clocks with CS and DI (MoSi) high
               break;
            case optResetNum: // send 80 clocks with 400 kHz and \CS peak
               smcSetIdle(5);
               break;

            case optRestartNum: // go idle with 80 clocks and 400 kHz
               smcSetIdle(1);
               break;

            case optOffNum: // send go idle command
               smcSetIdle(2);
               break;

            case optOnNum: // start card determination in thread
               getSMCtype();
               break;

            case optSlowNum:  //  slow     slow(er) speed 400 kHz
               spi2MemCardBaud = SPI400kHz;
               break;
            case optNormalNum: // normal   standard  speed 5 MHz
               spi2MemCardBaud = SPI5MHz;
               break;
            case optFastNum:  // fast speed 10 MHz
               spi2MemCardBaud = SPI10MHz;
               break;

            case optReadNum:
               if (cliThread->paramStart  != optNotGivenNum) { // have a parameter
                  uint32_t sector;
                  parseDWordNum(&sector, cliThread->line,
                        cliThread->paramStart);
                  readDataBlock(sector, smcState.sectorBuff);

               } // have parameter

               PT_YIELD_SPACE4SYS_CLI(35);

               stdPutS_P(sysRunSectorN);
               char tmp[58]; // must be 16 for smcReadCID  and 55 for form16BytSeq
               eightHexs(tmp, smcState.rwSector); tmp[8] = 0;
               stdPutS(tmp);
               stdPutS_P(bLF1);
               for(cliThread->paramEnd  = 0; cliThread->paramEnd  < 32;
                                                 ++cliThread->paramEnd ) {
                  PT_YIELD_SPACE4SYS_CLI(75); // re-entry  <<<
                  threeDigs(tmp, ((uint16_t)cliThread->paramEnd * 16));
                  tmp[3] = ':';   tmp[4] = ' '; tmp[5] = 0;
                  stdPutS(tmp);
                  uint32_t * begin = ((uint32_t *)smcState.sectorBuff)
                                                       + (4 * cliThread->paramEnd);
                  char * s = form16BytSeq(tmp, (uint8_t *)begin);
                  *s++ = ' '; *s++ = '\n';  *s = 0;
                  stdPutS(tmp);
               } // for
               break;

            case optApplicNum: // mis-used for test
               smcReadCID((uint8_t *)tmp);
               break;

            case optWriteNum: // mis-used
               sendCmd0arg(SEND_STATUS);
               if(!smcState.lastCmdResp[0]) smcState.lastCmdResp[1] = spi2Receive();
               deSelectSMC();
               // smcThreadF();
               break;

            case optNotGivenNum:
               break;
            default:
               if (!cliThread->repStreams) goto cliNormalEnd; // out
               unimplOptionReport(cliThread); // helper
               break; // no option or not used here
         } // option switch

         if (!cliThread->repStreams) break;
         cliThread->infoThread.flag |=
                       spi2Errors ? 0x03 : 0x01; // Bit 1  SPI 2 state 0: SMC
#endif // have SMC
         makeLFnBreak:  // makes start LF if bits in infoThread.flag2  are set
         PT_YIELD_SPACE4SYS_CLI(29); // re-entry point <<<
         stdPutS_P(bLF1);
         break;

      case comDirlistNum : // dirList [-option] [dirPath] memory card directory
#if HAVE_SMC
         if (!cliThread->repStreams) break; // no report streams no command
         if (!lockFsWorkFor(SMC_FS_SYSTUSE)) break; // others use FS
         fsWork.fRes = checkValidFS(0);
         if (fsWork.fRes == FR_NO_FILESYSTEM) {
            fsWork.fRes = lookForFS(0);
            PT_YIELD(&cliThread->systCLIpt); // yield anyhow after long op.
         }
         if (fsWork.fRes) goto weGoReportingDir;

         fsWork.path = "/";
         if (cliThread->paramStart != optNotGivenNum) { // we have a parameter
            fsWork.path = &cliThread->line[cliThread->paramStart];
            cliThread->line[cliThread->paramEnd] = 0; // make it 0-terminated
          } // we have a parameter


     #if _USE_LFN
         fsWork.filInf.lfname = fsWork.lfn;
         fsWork.filInf.lfsize = sizeof fsWork.lfn;
     #endif

         fsWork.fRes = f_opendir(&fsWork.dir, fsWork.path);
         PT_YIELD(&cliThread->systCLIpt); // yield anyhow after long op.
         if (fsWork.fRes == FR_OK) {
            stdPutS("directory : "); stdPutS(fsWork.path); stdPutC('\n');
            PT_YIELD(&cliThread->systCLIpt); // yield anyhow after long op.
            for (;;) {
                fsWork.fRes = f_readdir(&fsWork.dir, &fsWork.filInf); // item
                PT_YIELD(&cliThread->systCLIpt); // yield anyhow after long op.
                if (fsWork.fRes != FR_OK || fsWork.filInf.fname[0] == 0) break; // on error or end of dir
                if (fsWork.filInf.fname[0] == '.') continue;   // Ignore dot entry
                PT_YIELD_SPACE4SYS_CLI(41);
                stdPutFilInf(&fsWork.filInf,
                      cliThread->optionNumb == optQuietNum ? 0x80 : 0xB9);
            }  // for over dir
         }

         weGoReportingDir:
         PT_YIELD_SPACE4SYS_CLI(31);
         if (fsWork.fRes) {
            //              0123456789.123456789.123 4
            char outT[] =  " * dir.list err: OK  \n";
            twoHexs(&outT[17], fsWork.fRes);
            stdPutS(outT);
         } else stdPutC('\n');
         unlockFsWorkFrom(SMC_FS_SYSTUSE);
#endif // have SMC
         break;

      case comTypeFileNum : // type file
#if HAVE_SMC
         if (!cliThread->repStreams) break; // no report streams no command
         if (cliThread->paramStart == optNotGivenNum) break; // need parameter
         if (!lockFsWorkFor(SMC_FS_SYSTUSE)) break; // others use FS
         fsWork.fRes = checkValidFS(0);
         if (fsWork.fRes == FR_NO_FILESYSTEM) {
            fsWork.fRes = lookForFS(0); // try detect FS now
            PT_YIELD(&cliThread->systCLIpt); // yield anyhow after long op.
         }
         if (fsWork.fRes) goto weGoReportingType;

         fsWork.path = &cliThread->line[cliThread->paramStart];
         cliThread->line[cliThread->paramEnd] = 0; // make it 0-terminated
         PT_YIELD_SPACE4SYS_CLI(56);
         stdPutS(fsWork.path);
         stdPutS_P(bLF1);
         fsWork.fRes = f_open(&fsWork.fil, fsWork.path,
                          FA_OPEN_EXISTING | FA_READ);
         PT_YIELD_SPACE4SYS_CLI(70);
         PT_OR_YIELD_REENTER(); // yield anyhow after long fatFS op.
         if (fsWork.fRes) goto weGoReportingType;

         for (;;) {
            uint16_t ii;
            char buffer[20];
            // memset(buffer, 0, 16);  // only to fill last short with 0
            fsWork.fRes1 = f_read(&fsWork.fil, buffer, 20, &ii);
            if (fsWork.fRes1 || !ii) break;
            char s[70];
            char * beg = formN20BytSeq(s, (uint8_t *)buffer, ii);
            *beg++ = ' '; *beg++ = '\n'; *beg = 0;
            stdPutS(s);
            if (ii < 20) break;
            PT_YIELD_SPACE4SYS_CLI(70);
            PT_OR_YIELD_REENTER(); // yield anyhow after long fatFS op.
         } // for til break by EOF or fault
         fsWork.fRes2 = f_close(&fsWork.fil);

         weGoReportingType:
         PT_YIELD_SPACE4SYS_CLI(31);
         PT_OR_YIELD_REENTER(); // yield anyhow after long fatFS op.
         if (fsWork.fRes || fsWork.fRes1 || fsWork.fRes2) {
            //              0123456789.123456789.123456789t123456789q
            char outR[] =  " * fil.op.cl: op rd cl  \n";
            if (fsWork.fRes) twoHexs(&outR[14], fsWork.fRes);
            if (fsWork.fRes1) twoHexs(&outR[17], fsWork.fRes1);
            if (fsWork.fRes2) twoHexs(&outR[17], fsWork.fRes2);
            stdPutS(outR);
         } else stdPutC('\n');
         unlockFsWorkFrom(SMC_FS_SYSTUSE);
#endif // have SMC
         break;

      default:
         break;
   } // switch

   cliInfoEnd:   // all done but check / execute info flags
   if (cliThread->infoThread.flag2 || cliThread->infoThread.flag) {
      PT_YIELD_SPACE4SYS_CLI(69); // re-entry point <<<
      const uint8_t outFlashErg = outRunInfoThreadF(&cliThread->infoThread,
                           cliThread->repStreams);
      if (outFlashErg <= PT_YIELDED) return outFlashErg; // re-ent ^^^
      //  if (outFlashErg == PT_EXITED)  PT_EXIT(&cliThread->systCLIpt);
   }

   cliNormalEnd:   // all done
   cliThread->commNumb = 0;   // the flag
   PT_END(&cliThread->systCLIpt)
} // sysCliThreadF()

