/**
 *  \file network.c
 *
 *  Implementation of weAutSys' network system services
 *
 *  This is system software and must not be modified for user or
 *  application programs.
 *
 *  This file is part of weAutSys on WeAut_01 &nbsp;
 *           <<a href="http://weinert-automation.de">weinert-automation.de</a>>
 *
 * &nbsp; \ref copyLicNote " Copyright &copy; " 2011 Albrecht Weinert, Bochum
 * \author  Albrecht Weinert  &nbsp; <<a href="http://a-weinert.de">a-weinert.de</a>>
 * 2
 * 25.01.2017
 */
//  SVN last change by A. Weinert

#include "we-aut_sys/network.h"
#include "we-aut_sys/common.h"
#include "we-aut_sys/ll_system.h"
#include "we-aut_sys/utils.h"
#include "we-aut_sys/system.h"
#include "we-aut_sys/persist.h"
#include "we-aut_sys/enc28j60.h"
#include "we-aut_sys/ntp.h"
#include "we-aut_sys/streams.h"
#include "we-aut_sys/timing.h"
#include "we-aut_sys/telnet.h"
#include "we-aut_sys/modbus.h"
#include "uip/apps/resolv.h"
#include "uip/apps/dhcpc.h"
#include "pt/pt.h"  // only to please Eclipse

#if DEBUG_ETH > 1
# include "we-aut_sys/log_streams.h"
#endif

#include <string.h>


ipConf_t curIpConf;
uint8_t  networkNotReady; // Network not ready

INFLASH(char const upReady[])  = "up and ready"; //!< report text: up and ready
INFLASH(char const intfDown[]) = "interface (ENC) down"; //!< report: ENC down
INFLASH(char const linkDown[]) = "link down"; //!< report text: link down
INFLASH(char const linkWsDn[]) = "link run-up"; //!< report text: link up
INFLASH(char const encEthRev[])= ", ENC revID: ";
INFLASH(char const encSPIspeed[])= ", spiBRR: ";

/*  Format a MAC address
 *
 *  The MAC address from the structure supplied will be formatted as
 *  11:22:33:44:55:66 delivering 17 characters + a trailing zero.
 *
 *  \param s start position within string resp. char array
 *  \param mac the (6 byte) address structure
 *  \return pointer to the (0-) character following the formatted MAC-address
 */
char * formMacAdd(char * s, eth_addr_t * mac){
   twoHexs(s, mac->addr[0]);
   twoHexs(&s[3], mac->addr[1]);
   twoHexs(&s[6], mac->addr[2]);
   twoHexs(&s[9], mac->addr[3]);
   twoHexs(&s[12], mac->addr[4]);
   twoHexs(&s[15], mac->addr[5]);
   s[2] = s[5] = s[8] = s[11] = s[14] = ':';
   s += 17;
   *s = 0;
   return s;
} // formMacAdd(char *, eth_addr_t *)


/* Format an IP configuration
 *
 *  This function formats the IP configuration to a multi-line text.<br />
 *  The length of the text delivered (to  *s) is up to 165 characters
 *  including a trailing zero.
 *
 *  The output does not contain the DHCP information.
 *
 *  \param s start position within string resp. char array
 *  \param ipConf  pointer to the IP configuration (addresses, masks, flags)
 *  \return pointer to the character following the formatted configuration (set 0)
 *  \sa formMacAdd
 *  \sa formIPAdd
 *  \sa formIpConfDHCP
 */
char * formIpConf(char * s, ipConf_t * ipConf){
   memcpy_P(s, ipConIpAd, 16);
   s = formIpAdd(s +16, ipConf->ipAddr); // 7..15
   *s++ = ' '; *s++ = '\n';

   memcpy_P(s, ipConNMsk, 16);
   s = formIpAdd(s + 16, ipConf->netMask);
   *s++ = ' '; *s++ = '\n';

   memcpy_P(s, ipConDefR, 16);
   s = formIpAdd(s + 16, ipConf->defaultRouter);
   *s++ = ' '; *s++ = '\n';

   if (ipConf->setFlags & DNS1_MSK ) { // DNS server set
      memcpy_P(s, ipConDNSa, 16);
      s = formIpAdd(s + 16, ipConf->dnsAddr);
      if (ipConf->setFlags & DNS2_MSK ) { // DNS server 2 set
         *s++ = ' '; *s++ = ' ';
         s = formIpAdd(s, ipConf->dns2Addr);
      } // DNS server 2 set
      *s++ = ' '; *s++ = '\n';
   } // DNS server set

   if (ipConf->setFlags & NTP1_MSK ) { // NTP server set
      memcpy_P(s, ipConNTPa, 16);
      s = formIpAdd(s + 16, ipConf->ntpAddr);
      if (ipConf->setFlags & NTP2_MSK ) { // NTP server 2 set
         *s++ = ' '; *s++ = ' ';
         s = formIpAdd(s, ipConf->ntp2Addr);
      } // NTP server 2 set
      *s++ = ' '; *s++ = '\n';
   } // NTP server set

   *s++ = ' '; *s++ = '\n';
   *s = 0;
   return s;
} // formIpConf

/*  Format an IP configuration, the DHCP part
 *
 *  This function formats the IP configuration to a multi-line text.<br />
 *  The length of the text delivered (to  *s) is up to 65 characters
 *  including a trailing zero.
 *
 *  The output does only contain the DHCP information.
 *
 *  \param s start position within string resp. char array
 *  \param ipConf  pointer to the IP configuration (addresses, masks, flags)
 *  \return pointer to the character following the formatted configuration (set 0)
 *  \sa formMacAdd
 *  \sa formIPAdd
 *  \sa formIpConf
 */
char * formIpConfDHCP(char * s, ipConf_t * ipConf){
   if (ipConf->setFlags & DHCP_MSK ) { // set by DHCP
      memcpy_P(s, ipConDHCs, 11);
      s += 11;
      // format  the since set time here  as duration
      uint32_t dur =  secClock() - ipConf->setTime;
      datdur_t  setTP;
      sec2datdur(&setTP, dur);
      s = formTimDur(s, &setTP);
      *s++ = ' '; *s++ = '/';
      sec2datdur(&setTP, ipConf->leaseTime);
      s = formTimDur(s, &setTP);
   } else if (ipConf->useFlags &  DHCP_MSK ) { // use DHCP
      memcpy_P(s, ipConDHCu, 11);
      s += 11;
   } // use DHCP
   *s++ = ' '; *s++ = '\n';
   *s = 0;
   return s;
} // formIpConfDHCP()


/* Format an ipV4 Ethernet address
 *
 *  The IP address from the structure supplied will be formatted as
 *  112.12.0.1 and delivering 7..15 characters.
 *
 *  \param s start position within string resp. char array (length 16)
 *  \param ipAddr the (uint16-t[2] big endian) address structure
 *  \return pointer to the character following the formatted address (set 0)
 *  \sa formMacAdd
 *  \sa formIpConf
 */
char * formIpAdd(char * s, uip_ipaddr_t ipAddr){
   for(uint8_t vN = 4; vN; --vN) {
      uint8_t val;
      switch (vN) {
         case 4: val = LBYTE(ipAddr[0]); break; // L is H as is big endian
         case 3: val = HBYTE(ipAddr[0]); break;
         case 2: val = LBYTE(ipAddr[1]); break;
         default: val = HBYTE(ipAddr[1]); break;
      }
      if (val < 10) {
         *s++ = '0' + val;
         *s++ = '.';
         continue; // for
      }
      if (val >= 100) {
         if (val >= 200) {
            *s++ = '2';
            val -= 200;
         } else {
            *s++ = '1';
            val -= 100;
         }
      } // 100..255
      u8div_t q_r = divByVal10(val); // q:t r:o
      *s++ = '0' + q_r.quot;
      *s++ = '0' + q_r.rem;
      *s++ = '.';
   } // for
   *--s = '\0'; //
   return s;
} // formIpAdd(char *, ipaddr_t)

      //                           0123456789.123456
INFLASH(char const ipConIpAd[]) = "  IP4 address : ";
INFLASH(char const ipConNMsk[]) = "  IP4 netmask : ";
INFLASH(char const ipConDefR[]) = "  def. router : ";
INFLASH(char const ipConDNSa[]) = "  DNS address : ";
INFLASH(char const ipConNTPa[]) = "  NTP address : ";
INFLASH(char const ipConDHCu[]) = "  use DHCP ";
INFLASH(char const ipConDHCs[]) = "  DHCP set ";
INFLASH(char const ipConDefS[]) = "  def. set ";
       //                          0123456789.123456

//----------------------------------------------------------------------


/* Output Ethernet and driver (ENC28J60) state info, the thread function
 *
 *  \return \ref PT_YIELDED if waiting for conditions that may become
 *           automatically true; \ref PT_WAITING if blocked by conditions
 *           that other's action may set true;  \ref PT_EXITED if stopped
 *           by conditions that never will become true (sorry, not all work
 *           done and never will; \ref PT_ENDED if ready with all work
 */
ptfnct_t outEncLanInfoThreadF(hierThr_data_t * threadD, FILE * toStream){
   if (! (threadD->flag2 & 0x3F)) { // Bits 6 & 7 unused
      threadD->flag2 = 0;
      PT_LEAVE(&threadD->pt2); // quick out nope to do
   }

   char tmp[320];   // was 220
   // uint8_t cs;
   PT_BEGIN(&threadD->pt2);
   SAMPLE_MIN_SP(2);

   if (threadD->flag2 & 1) { // bit 0 MAC
      PT_WAIT_OUT_SPACE(&threadD->pt2, 38, toStream);
      stdPutS_P(macAdd_is);
      formMacAdd(tmp, &actMACadd);
      tmp[17] = 0;
      stdPutS(tmp);
      stdPutS_P(bLF1); // 1 blank 1 LF
      if (!(threadD->flag2 &= ~1)) PT_LEAVE(&threadD->pt2);
   }  // bit 0 MAC

   if (threadD->flag2 & 0x10) { // bit 4 Link state
      PT_YIELD_OUT_SPACE(&threadD->pt2, 64, toStream);
      stdPutS_P(encEthSta);
      tmp[2] = ' ';
      tmp[4] = tmp[3] = tmp[5] =0;
      if (networkNotReady) {
         twoHexs(tmp, networkNotReady);
         stdPutS(tmp);
         if (networkNotReady & 0xF0) {
            stdPutS_P(intfDown);
         } else if (networkNotReady & LNC_IDN) {
            stdPutS_P(linkDown);
         } else if (networkNotReady & LNC_WDN) {
            stdPutS_P(linkWsDn);
         }
      } else {
         stdPutS_P(upReady);
      }
      stdPutS_P(encEthRev);
      twoHexs(tmp, readControlRegister(EREVID)); // chip rev.
      stdPutS(tmp);
      stdPutS_P(encSPIspeed);
      threeDigsB(tmp, spi2EtherChipBaud);
      tmp[3] = ' ';
      tmp[4] = '\n';
      stdPutS(tmp);
      if (!(threadD->flag2 &= ~0x10)) PT_LEAVE(&threadD->pt2);
   }  // bit 4 Link state


   if (threadD->flag2 & 2) { // bit 1 IP setting
      PT_YIELD_OUT_SPACE(&threadD->pt2, (UART_OUT_BUF_CAP - 3), toStream);
      char * begin = formIpConf(tmp, &curIpConf);
      if (!(threadD->flag2 & 0x20)) { // minimal DHCP here
          //  begin =
                formIpConfDHCP(begin, &curIpConf);
      } // minimal DHCP here
      //  tmp[cs] = 0;
      stdPutS(tmp);
      stdPutS_P(bLF2); // 1 blank 2 LF
      if (!(threadD->flag2 &= ~2)) PT_LEAVE(&threadD->pt2);
    } // bit 1 IP setting

   if (threadD->flag2 & 0x20) { // bit 5 DHCP state
      PT_YIELD_OUT_SPACE(&threadD->pt2, 110, toStream);
      formIpConfDHCP(tmp, &curIpConf);
      // tmp[cs] = 0;
      stdPutS(tmp);
      stdPutS_P(bLF2); // 1 blank 2 LF
      if (!(threadD->flag2 &= ~0x20)) PT_LEAVE(&threadD->pt2);
    } // bit 5 DHCP state

   if (threadD->flag2 & 4) { // bit 2 IP default
      PT_YIELD_OUT_SPACE(&threadD->pt2, (UART_OUT_BUF_CAP - 3), toStream);
      stdPutS_P(ipDefault);
      stdPutS_P(bLF1); // 1 blank 1 LF
      char * begin = formIpConf(tmp, &(defaultConfData.ipConf));
      formIpConfDHCP(begin, &(defaultConfData.ipConf));
      // tmp[cs] = 0;
      stdPutS(tmp);
      stdPutS_P(bLF2); // 1 blank 2 LF
      if (!(threadD->flag2 &= ~4)) PT_LEAVE(&threadD->pt2);
   } // // bit 2 IP default

   if (threadD->flag2 & 8) { // bit 3 ARP
      PT_WAIT_OUT_SPACE(&threadD->pt2, (UART_OUT_BUF_CAP - 3), toStream);
      stdPutS_P(arpEntries);
      stdPutS_P(bLF1); // 1 blank 1 LF
      threadD->flag3 = threadD->flag4 = 0;
      struct arp_entry *tabptr;
      for(; threadD->flag3 < UIP_ARPTAB_SIZE; ++threadD->flag3) { // over ARP table
         PT_YIELD_OUT_SPACE(&threadD->pt2, 45, toStream);
         tabptr = &arp_table[threadD->flag3];
         if(tabptr->time) { // used entry
           char * begin = formMacAdd(tmp, &tabptr->ethaddr);
           *begin++ = ' ';
           begin = formIpAdd(begin, tabptr->ipaddr);
           *begin++ = ' '; *begin++ = '\n';
           *begin = 0;
           /// dangerous out : while (outSpace(stdout) < 18 + 17 + 30) ;
           stdPutS(tmp);
           threadD->flag4 = 1;  // signal an entry
        } // used entry
      }  // over ARP table
      if (threadD->flag4) {
         stdPutS_P(bLF2); // 1 blank 2 LF
      } else {
         stdPutS_P(arpEmpty); // no entries + 1 blank 2 LF
      }

      if (!(threadD->flag2 &= ~8)) PT_LEAVE(&threadD->pt2);
   } // bit 3 ARP

   PT_END(&threadD->pt2)
} // outEncLanInfoThreadF()


/* Set the (actual stack's) MAC address from flash structure */
void setMACaddP(const eth_addr_t * mac){
   memcpy_P(&actMACadd, mac, 6);
} // setMACaddP(eth_addr_t *)


/* Set the addresses in the uIP stack
 *
 *  This function takes the fields ipAddr, netMask, defaultRouter and
 *  dnsAddr and makes the appropriate settings in the \ref uip "uIP" stack.
 */
void uipSetAddresses(ipConf_t * ipConf){
   uip_sethostaddr(ipConf->ipAddr);
   uip_setnetmask(ipConf->netMask);
   uip_setdraddr(ipConf->defaultRouter);
   // we can use only one DNS server at present
   if (ipConf->setFlags &  ipConf->useFlags & DNS1_MSK) { // has & use DNS
      resolv_conf((uint16_t *)(ipConf->dnsAddr));
   }
   networkNotReady &= ~STC_NST;
} // uipSetAddresses(ipConf_t *)

/* Get the addresses from the uIP stack
 *
 *  \param ipConf  pointer to the destination of the addresses and masks
 */
void uipGetAddresses(ipConf_t * ipConf){
   memcpy(&(ipConf->ipAddr), uip_hostaddr, 4);
   memcpy(&(ipConf->netMask), uip_netmask, 4);
   memcpy(&(ipConf->defaultRouter), uip_draddr, 4);
   memcpy(&(ipConf->dnsAddr), resolv_getserver(), 4);
} //  uipGetAddresses(ipConf_t *)


/* Initialise the network stack  */
void networkInit() {
  encInit();   // HW driver
  uip_init(); // uiP TCP IP stack
  uip_arp_init(); // Start uip's Address Resolution Protocol
  dhcpReset();   // initial state of DCP
  ntpReset();   // initial state for NTP
  resolv_init(); // initial state for DNS (resolver)
  telnetInit(); // initial state for Telnet (server)
  modbusInit(); // initial state for Modbus (server)
  linkState();  // record link state as not to re-trigger
} // networkInit()

/* Initialise the LAN communication (standard way) */
void lanComStdInit() {
  if (curIpConf.useFlags & DHCP_MSK) { // use DHCP
     dhcpInit(&actMACadd);
      // not done here (yet)
  } else { // no DHCP : use default
     memcpy(&curIpConf, &(defaultConfData.ipConf), sizeof(ipConf_t));
     curIpConf.setFlags &= ~DHCP_MSK;  // omit
     curIpConf.setTime = secClock();
     curIpConf.leaseTime = 0;
     uipSetAddresses(&curIpConf);
     ntpInit(0); // checks (curIpConf.useFlags & NTP_CLIENT_MSK)
  } // no DHCP : use default and try to start NTP client
} // lanComStdInit()

#if DEBUG_ETH > 1
// # include "we-aut_sys/log_streams.h"
INFLASH(char ethHWsendMes[] ) = "  uIP send  _len ";
//                              0123456 89.12345678 9
INFLASH(char encBusyAbMes[] ) = "  ENC busy abort \n";
INFLASH(char encTxStalled[] ) = "  ENC tx stalled \n";
INFLASH(char encPacksRecv[] ) = "  ENC pack recvd ";
INFLASH(char encPackSize[]  ) = "  ENC rx pcksize ";
#endif


/* Send data which was written therefore to uip_buf  */
void deviceDriverSend(){
#if DEBUG_ETH > 2
   bufLogDec_P(ethHWsendMes, uip_len);
#endif
   // check if transmit logic is idle (must be !)
   // should be checked if unclear before calling this function
   if ( readControlRegister(ECON1) & (1<<TXRTS)) { //  transmit logic busy

      // transmit logic problem. See ENC Errata B7 #10 (B4 #12).
      if ( readControlRegister(EIR) &  (1<<TXERIF)) {
         setBitfield(ECON1, (1<<TXRST));    // set and reset the send
         clearBitfield(ECON1, (1<<TXRST)); // request works around enc's bug
         clearBitfield(EIR, (1<<TXERIF)|(1<<TXIF)); // clear ready and error, too
#if DEBUG_ETH > 2
         bufLogTxt_P(encTxStalled);
#endif
      } else {
#if DEBUG_ETH > 2
         bufLogTxt_P(encBusyAbMes);
#endif
         return; // should not happen
      }
   } //  transmit logic busy

   // set send write pointer to start of transmit buffer
   writeControlRegister(EWRPTL, LBYTE(TXBUF_STRT));
   writeControlRegister(EWRPTH, HBYTE(TXBUF_STRT));

   // per packet control byte
   putBufferMemory( 0 );  // use MACON3 settings

   /* Fill Data into the Buffer memory
    * UIP Doc P.50  (else branch is app data not in uip_buf)  */
   if( uip_len <= UIP_LLH_LEN + UIP_TCPIP_HLEN ) {
      writeBufferMemory(uip_buf, uip_len);
   } else {
      writeBufferMemory(uip_buf, UIP_LLH_LEN + UIP_TCPIP_HLEN);
      writeBufferMemory(uip_appdata, uip_len - UIP_TCPIP_HLEN - UIP_LLH_LEN);
   }
   // calculate the end point of the Buffer
   writeControlRegister(ETXNDL, LBYTE(TXBUF_STRT + uip_len));
   writeControlRegister(ETXNDH, HBYTE(TXBUF_STRT + uip_len));

   clearBitfield(EIR, (1<<TXERIF)|(1<<TXIF)); // clear ready and error, too
   setBitfield(ECON1, (1<<TXRTS));
} // deviceDriverSend()


/* Receiving data from ENC28J60 and filling it into the uip_buf correctly */
void networkPolling(void){     // package was received
   uip_len = deviceDriverPoll();
   if(uip_len) { // received package in uip_buf
      // Check if there ...
      if(BUF2EH->type == HTONS(UIP_ETHTYPE_IP)) {
         uip_arp_ipin();  // may be used as new ARP entry
         uip_input();
         if(uip_len) {
            uip_arp_out();
            deviceDriverSend(); // if there is a package to send send it now
         } // if
      } else if(BUF2EH->type == HTONS(UIP_ETHTYPE_ARP)) {
         uip_arp_arpin();
         if(uip_len) {
            deviceDriverSend(); // if there is a package to send send it now
         }
      }
   } // if something received
} // networkPolling()


/* Receive status vector */
uint8_t receiveStatVec[4];


/*  Polls package out of ENC28J60 and move it to the uip_buf  */
uint16_t deviceDriverPoll(void) {
   const uint8_t noOfPacksRec = readControlRegister(EPKTCNT);
   if (! noOfPacksRec ) return 0; // pack cnt 0

#if DEBUG_ETH > 2
   {
   bufLogTxt_P(encPacksRecv);
   char tmp[] = ".. [....] \n";
   //            0123456789. 1
   twoDigs(tmp, noOfPacksRec);
   fourHexs(&tmp[4], actPackInd);
   //  while (uartOutSpace() < 11);
   bufLogTxt(tmp, 11);
   }
#endif

   // read pointer to the start of the received & unprocessed packet
   writeControlRegister(ERDPTL, LBYTE(actPackInd));
   writeControlRegister(ERDPTH, HBYTE(actPackInd));

   // get next package pointer from buffer (low byte first)/
   readBufferMemory((uint8_t*)&actPackInd, 2);

   readBufferMemory(receiveStatVec, 4); // read status vector

   /* first 2 Bytes of status vector indicates the package_size in Bytes
     remove the 4 Bytes of the CRC Checksum at the end of the ENC buffer */
   ucnt16_t package_size;
   package_size.v8[0] = receiveStatVec[0];
   package_size.v8[1] = receiveStatVec[1];
   package_size.v16 -= 4; // ignore the last 4 CRC bytes (checked by enc)

#if DEBUG_ETH > 2
   {
   bufLogTxt_P(encPackSize);
   char tmp[] = ".. [....] = .... s:.... \n";
   //            0123456789.123456789.1234 5
   fourHexs(&tmp[4], actPackInd);
   fourDigs(&tmp[12], package_size.v16);
   twoHexs(&tmp[19], receiveStatVec[2]);
   twoHexs(&tmp[21], receiveStatVec[3]);
   // while (uartOutSpace() < 25);
   bufLogTxt(tmp, 25);
   }
#endif

   if (package_size.v16 > UIP_BUFSIZE + 1) package_size.v16 = UIP_BUFSIZE + 1;

   // optionally check CRC here (receiveStatVec[2] & 0x80)
   // not necessary as we let wrong CRC reject by ERXFCON setting
   readBufferMemory(uip_buf, package_size.v16);

   // go to next package (this frees memory just read out)
  /// writeControlRegister(ERXRDPTL, LBYTE(actPackInd));
  ///  writeControlRegister(ERXRDPTH, HBYTE(actPackInd));
   // as we point to next package pointer just read out no calculation

   // ENC Bug 11 requires that only odd values are written to ERXRDPT
   // actPackInd is said to be even and RXBUF_END is odd
   if ((actPackInd == 0) || (actPackInd -1 > RXBUF_END)) {
      writeControlRegister(ERXRDPTL, LBYTE(RXBUF_END));
      writeControlRegister(ERXRDPTH, HBYTE(RXBUF_END));
   } else {  // go to next package (this frees memory just read out)
      writeControlRegister(ERXRDPTL, LBYTE(actPackInd-1));
      writeControlRegister(ERXRDPTH, HBYTE(actPackInd-1));
   }

   // Decrement Package counter to indicate done with this
   setBitfield(ECON2, (1<<PKTDEC));

#if DEBUG_ETH > 2
   bufLogDec_P(encPackSize, package_size.v16);
#endif

   return package_size.v16;
} // deviceDriverPoll()

/* Output some Ethernet and driver ( ENC28J60) state info to serial output */
// mThr_data_t outEncLanInfoThread;

/* Output some Ethernet and driver (ENC28J60) state info to std output
 *  Bit 0 : MAC <br />
 *  Bit 1 : current IP setting <br />
 *  Bit 2 : default IP setting <br />
 *  Bit 3 : ARP <br />
 *  Bit 4 : Link state <br />
 *  Bit 5 : DHCP state
 */
//uint8_t outEncLanInfoFlag;

//-----------------------------------------------------------------------

/* Get the MAC address for an IP address
 *
 *  This function checks the ARP table for an entry about ipAddr and returns
 *  the (pointer to its) MAC address.
 *
 *  The caller must not manipulate the returned MAC.
 *  \param  ipAddr  IP address the MAC is to be determined for
 *  \return pointer to MAC if determinable or NULL else
 */
const struct uip_eth_addr * getMACforIPaddr(uip_ipaddr_t ipAddr){

   // First check if ipAddr is a local broadcast
   if(uip_ipaddr_cmp(ipAddr, broadcast_ipaddr)) { // broadcast IP
      return &broadcast_ethaddr; // gets broadcast MAC
   }
   // Check if ipAddr is this device's own address
   if(uip_ipaddr_cmp(ipAddr,  uip_hostaddr)) { // own IP
      return &actMACadd; // gets own MAC
   }

   uip_ipaddr_t ipaddr = {ipAddr[0],  ipAddr[1]};
   // Check if the ipAddris on the local network
   if(!uip_ipaddr_maskcmp(ipAddr, uip_hostaddr, uip_netmask)) {
     /* For a destination not on the local network we look up the
      * the default router's IP address instead */
        uip_ipaddr_copy(ipaddr, uip_draddr);
   }
   struct arp_entry *tabptr;
   uint8_t i;
     for(i = 0; i < UIP_ARPTAB_SIZE; ++i) {
        tabptr = &arp_table[i];
        if(tabptr->time && uip_ipaddr_cmp(ipaddr, tabptr->ipaddr)) {
           return &(tabptr->ethaddr);
        }
     }
   return NULL;
} // getMACforIPaddr   xxx untested yet

//---------------------------------------------------------------------------

#if DEBUG_ApL
INFLASH(char const uipAppCalledMes[]  ) = "uIP app_call : ... \n";
//                                   0123456 89.123456789 .
INFLASH(char const udpAppCalledMes[]  ) = "udp app_call : ... \n";
#endif


/* the uIP event function */
void uip_appcall(void){
#if DEBUG_ApL
   bufLogTxt_P(uipAppCalledMes);
   char tmp[17];
   twoHexs(tmp, uip_flags);
   tmp[2] = 'p';
   fourDigs(&tmp[3],  convert16endian(uip_conn->lport));
   bufLogTxt(tmp, 7);
   bufLogPut2C(' ', '\n'); // 1 blank 1 LF
#endif
   if(uip_conn->lport == HTONS(TELNET_PORT)) {  // Telnet protocol
      telnetAppcall();
      return;
   } // Telnet protocol
   if(uip_conn->lport == HTONS(MODBUS_PORT)) {  // Telnet protocol
      modbusAppcall();
      return;
   } // Telnet protocol
   uipAppcall(); // all other protocols by application software
} // uip_appcall(void)

void udp_appcall(void){

   if(uip_udp_conn->rport == HTONS(NTP_PORT)) {  // NTP protocol
     ntpAppcall();
     return;
   }  // NTP protocol (from NTP server)

   if(uip_udp_conn->rport == HTONS(DHCPC_SERVER_PORT)) {  // DHCP protocol
      dhcpc_appcall();
      return;
   } // DHCP protocol

   if(uip_udp_conn->rport == HTONS(53)) { // DNS port
      resolv_appcall();
      return;
   } // DNS port

   // further UDP protocols here
   // only unimplemented protocols get debug
 #if DEBUG_ApL
    bufLogTxt_P(udpAppCalledMes);
    char tmp[12];
    twoHexs(tmp, uip_flags);

    fourDigs(&tmp[2],  convert16endian(uip_udp_conn->lport));
    tmp[6] = 'L';

    fourDigs(&tmp[7], convert16endian(uip_udp_conn->rport));
    tmp[11] = 'R';

    bufLogTxt(tmp, 12);
    bufLogPut2C(' ', '\n'); // 1 blank 1 LF
 #endif
    udpAppcall(); // all other protocols by application software
} // udp_appcall(void)



