/**
 * \file spi.c
 *
 * weAutSys'/weAut_01' system calls and services for the SPI interfaces
 *
 *  This file contains the implementations of \ref intro_sec "weAutSys" /
 *  weAut_01  (low level) system calls and services for the serial peripheral
 *  interfaces (SPI).
 *  This is system software and must not be modified for user or
 *  application programs.
 *
 * This file is part of weAutSys &nbsp;
 *      <<a href="http://weinert-automation.de">weinert-automation.de</a>>
 *      <br />
 * This source is for weAut_00 and weAut_01 only. Other platforms
 * (arduMega2560, arduinoUno) must not include/translate it respectively have
 * a specific variant.
 *
 * &nbsp; \ref copyLicNote " Copyright &copy; " 2011 Albrecht Weinert, Bochum
 * \author  Albrecht Weinert  &nbsp; <<a href="http://a-weinert.de">a-weinert.de</a>>
 * 2
 * 25.01.2017
 */
//  SVN last change by A. Weinert

#include "we-aut_sys/spi.h"
#include "we-aut_sys/ll_system.h"
#include "we-aut_sys/system.h"

void deSelectSPI1_impl(){
    // Port C C2 /CS DIleds;   C3 /CS DO;
    orCP( (1<<3) | (1<<2) );
} // deSelectSPI1_impl()


/* De-select all SPI 1 devices  */
uint8_t deSelectSPI1(){
   uint8_t retV = SPSR;
   if (!(retV & (1<<SPIF))) {
      uint8_t mxW = 4; // assumes max. SPI1 clock rate (16 CPU clocks / byte
      // Hint less than 4 is optimized loop flat and hence to short
      while (!(SPSR & (1<<SPIF)) && --mxW); // wait with timeout 6 clock loop
   }
   deSelectSPI1_impl(); // is always inline
   return retV;
} // deSelectSPI1()

//------------------------------------------------------------------------

/* The SPI (2) clock frequency used for the Ethernet driver chip */
uint8_t spi2EtherChipBaud;


/* The SPI (2) clock frequency used for the memory card (holder) */
uint8_t spi2MemCardBaud;


/* Chip-select a SPI 1 device (the digital input display LEDS / HMI) */
void  selectDIdisplayLEDs(){
   // Port C C2 /CS DIleds; C3 /CS DO
   andOrCP(~ (1<<2),   (1<<3) );
} // selectDIdisplayLEDs()

/* Chip-select a SPI 1 device (the digital output driver) */
void selectDOdriver(){
   // Port C C2 /CS DIleds;   C3 /CS DO;
   andOrCP(~ (1<<3),   (1<<2) );
} // selectDOdriver()

/* Chip-select two SPI 1 devices (DO driver and DI LEDs) */
void selectDOdrivDIleds(){
   // Port C C2 /CS DIleds;   C3 /CS DO
   andCP( ~((1<<3) | (1<<2)) );
} //  selectDOdrivDIleds()  {


//--------------------   SPI 1 / SPI 2  ----------------------------------

/* De-select all SPI 2 devices
 *
 * \internal investigate if a dummy read (UART as SPI 2) is recommendable!
 *           seems not so
 */
void deSelectSPI2(){
   const uint8_t memsel =  memCardIsSelected();
   SPI2COND_ERR_RET(UCSR1A & (1<<TXC1), 0x84, )  // wait   for TX complete
#if defined(weAut_00)
    // Port C  C6 /CS memory card  Port D  D5 Ethernet CS
   orCP( (1<<6) );
   orDP( (1<<5) );
#elif defined(weAut_01)
# if USE_TESTPINS  == 3
   // Port C  C6 /CS memory card   C7 Ethernet CS; C0, C1 testpins
  orCP( 0xC3 );
# else
   // Port C  C6 /CS memory card  C7 Ethernet CS
  orCP( (1<<6) | (1<<7) );
# endif
#endif
   if (memsel) {
      ////  32.08.12 dummy send
      UCSR1A = (1<<TXC1); //  reset tx complete here necessary ??
      UDR1 = 0xFF;    ////  32.08.12 dummy send
      spi2Receive(); ////  32.08.12 dummy receive
   }
} // deSelectSPI2()


/* Chip-select a SPI 2 device: the Ethernet driver chip  */
void selectEtherChip(){
   // spi2Receive(); ////  32.08.12 dummy receive
   // JJF xx UBRR1 = (uint16_t)spi2EtherChipBaud;
   if ( UBRR1L != spi2EtherChipBaud) { // wrong SClk
      UCSR1B = 0; // disable
      UBRR1 = 0;  // 16bit register; Baud 0 for init
      //  UCSR1A |= (1<<U2X1);  // Set U2X mode (the manual is unclear here)
      /* Set MSPI mode of operation and SPI data mode 0. */
      UCSR1C = (1<<UMSEL11) | (1<<UMSEL10) | (0<<UCPHA1) | (0<<UCPOL1);
      /* Enable receiver and transmitter. */
      UCSR1B = (1<<RXEN1)|(1<<TXEN1);
      /* Set baud rate  IMPORTANT: has to be set after above enable  */

      UBRR1H = 0;
      UBRR1L = spi2EtherChipBaud;
   } // wrong SClk
#if defined(weAut_00)
   // Port C  C6 /CS memory card  Port D  D5 Ethernet CS
   orCP( (1<<6) );
   andDP( ~(1<<5) );
#elif defined(weAut_01)
# if USE_TESTPINS  == 3
   // Port C  C6 /CS memory card  C7 Ethernet CS  C1 C0 testpins
   andOrCP( 0x3C, 0x41); //    ~(1<<7), (1<<6) );
# else
   // Port C  C6 /CS memory card  C7 Ethernet CS
   andOrCP( 0x3F, (1<<6) ); //    ~(1<<7), (1<<6) );
# endif
#endif
} //  selectEtherChip()

/* Chip-select a SPI 2 device: the memory card (holder) */
void selectMemCard() {
   // JJF xx  UBRR1 = (uint16_t)spi2MemCardBaud;
   if ( UBRR1L != spi2MemCardBaud) { // wrong SClk
     UCSR1B = 0; // disable
     UBRR1 = 0;  // 16bit register; Baud 0 for init
      //  UCSR1A |= (1<<U2X1);  // Set U2X mode (the manual is unclear here)
      /* Set MSPI mode of operation and SPI data mode 0. */
      UCSR1C = (1<<UMSEL11) | (1<<UMSEL10) | (0<<UCPHA1) | (0<<UCPOL1);
      /* Enable receiver and transmitter. */
      UCSR1B = (1<<RXEN1)|(1<<TXEN1);
      /* Set baud rate  IMPORTANT: has to be set after above enable  */

      UBRR1H = 0;
      UBRR1L = spi2MemCardBaud;
   } // wrong SClk
#if defined(weAut_00)
    // Port C  C6 /CS memory card Port D  D5 Ethernet CS
   orCP( (1<<7) );
   andCP( ~(1<<6) );
#elif defined(weAut_01)
# if USE_TESTPINS  == 3
   // Port C  C6 /CS memory card  C7 Ethernet CS  C1 C0 testpins
   andOrCP( 0x3C, 0x82); //    ~(1<<7), (1<<6) );
# else
   // Port C  C6 /CS memory card  C7 Ethernet CS
  andOrCP( ~(1<<6), (1<<7) );
# endif
#endif
} // selectMemCard()

//----------------------------------------------------------------------

/* Send one byte over SPI 2  */
void spi2Transmit(uint8_t sendB){
   SPI2COND_ERR_RET(UCSR1A & (1<<UDRE1), 0x82, )
   //  while ( !( UCSR1A & (1<<UDRE1))); // spin wait empty transmit buffer

   //   UDR1 = sendB; was here / put after 13.08.12; see below
   UCSR1A = (1<<TXC1); //  reset tx complete  todo TEST if before UDR1 works
   UDR1 = sendB;

   // Data sheet: The TXCn Flag bit is automatically cleared when a transmit
   // complete interrupt is executed, or it can be cleared by writing a one
   // [sic!] to its bit location.
   // The TXCn Flag can be used to check that the Transmitter has completed
   // all transfers, and the RXCn Flag can be used to check that there are no
   // unread data in the receive buffer. Note that the TXCn Flag must be
   // cleared before each transmission (before UDRn is written) if it is
   // used for this purpose.
} // spi2Transmit(uint8_t)


/*  Accumulated errors at SPI 2
 *
 *  Driver function of SPI 2 (UART1 as SPI) record seen errors in this byte.
 *  It may be reset by user or system software (especially after logging or
 *  displaying the error(s)).software.
 *
 *  Bit 7 (8): any error. This bit will be set on every recognised SPI 2
 *         error. It may be reset alone leaving all others set.  <br/>
 *  Bit 6 (4): any memory card error. This bit will be set on every
 *         recognised SPI 2 error if at that time the
 *         \ref memCard_sec "small memory" card holder
 *         is \ref memCardIsSelected() "selected".
 *  Bit 5 (2): any network interface / ENC28J60 error. This bit will be set on
 *         every recognised SPI 2 error if at that time the Ethernet chip
 *         \ref nicIsSelected() "selected".
 *
 *  Bit 3 (8): network interface controller (NIC, ENC28J60) stays busy. This bit
 *         will be set if waiting for the \ref MISTAT "MISTAT.BUSY" bit to go
 *         away timed out. <br />
 *  Bit 2 (4): transmit complete failed. This bit will be set if waiting
 *         for transmission complete (TXC) timed out.  <br />
 *  Bit 1 (2): transmitter buffer empty failed. This bit will be set if waiting
 *         for empty transmitter buffer (UDRE) timed out.  <br />
 *  Bit 0 (1): receive complete failed. This bit will be set if waiting for
 *         receive complete (RXC) timed out.  <br />
 */
uint8_t spi2Errors;

/* Receive one byte from SPI 2 */
uint8_t spi2Receive(){
   SPI2COND_ERR_RET(UCSR1A & (1<<RXC1), 0x81, 0)
   return UDR1;
} // spi2Receive()

/* Send and Receive one byte via SPI 2 */
uint8_t spi2Tranceive(uint8_t sendB){
   // :: inlined spi2Transmit(sendB);
   SPI2COND_ERR_RET(UCSR1A & (1<<UDRE1), 0x82, 0)
   UCSR1A = (1<<TXC1); //  reset tx complete
   UDR1 = sendB;

   // :: inlined return spi2Receive();
   SPI2COND_ERR_RET(UCSR1A & (1<<RXC1), 0x81, 0)
   return UDR1;
} // spi2Tranceive(uint8_t)

/* Send two bytes and receive one byte via SPI 2
 *
 *  \param sendB1 the byte to be sent first
 *  \param sendB2 the second byte to be sent
 *  \return the byte read from SPI while transmitting \c sendB2 (0 may but
 *          must not indicate an \ref spi2Errors "error")
 */
uint8_t spi2Tranceive2(uint8_t sendB1, uint8_t sendB2){
   spi2Transmit(sendB1);
   spi2Transmit(sendB2);
   spi2Receive();
   return spi2Receive();
} // spi2Tranceive2(uint8_t, uint8_t)



/*  Send a byte n times and receive one byte via SPI 2
 *
 *  This function sends the byte \c sendB one or more times via SPI 2
 *  discarding all receptions except for the last one.
 *  Compared to doing this by \ref spi2Tranceive "spi2Tranceive(sendB)" n
 *  times this function is at least 20 % faster. On the other hand no useful
 *  work or yielding can be done while the SPI (2) interface is busy with
 *  the action. Insofar \n should be reasonably small.
 *
 *  \param sendB the byte to be sent \c n times
 *  \param n number to repeat \c sendB
 *  \return the byte read from SPI while transmitting \c sendB2
 */
uint8_t spi2TranceiveN(uint8_t sendB, uint8_t n){
   spi2Transmit(sendB);
   if (!n || !(--n)) goto lastRec;
   do {
     spi2Transmit(sendB);
     spi2Receive();  // should not wait long
   } while (--n);
   lastRec:
   return spi2Receive();
} // spi2TranceiveN(uint8_t sendB, uint8_t n)

/*  Receive n bytes to a buffer via SPI 2 with optional skip (before)
 *
 *  This function does \c skip + \c n receptions via SPI2 (putting all ones
 *  on MoSi). The (last) \c n bytes received are put into the buffer
 *  \c receiveB.
 *
 *  Compared to doing this by \ref spi2Tranceive "spi2Tranceive(0xFF)"
 *  \c skip + \c n times this function is at least 20 % faster. On the other
 *  hand no useful work or yielding can be done while the SPI (2) interface is
 *  busy with the action. Insofar \c skip + \c n should be reasonably small.
 *
 *  \param receiveB pointer to the buffer to receive \c n bytes to
 *                  (must not be NULL if \c n != 0)
 *  \param skip     if > 0 \c skip bytes are received and forgotten
 *                  before \c receiveB will be filled
 *  \param n        number of bytes to be received and filled into \c receiveB
 */
void spi2ReceiveSN(uint8_t * receiveB, uint16_t skip, uint16_t n){
   if (!(n | skip)) return;
   spi2Transmit(0xFF);
   while (skip) {
      spi2Transmit(0xFF);
      --skip;
      spi2Receive();  // should not wait long
   }
   if (!n || n==1) goto lastRec;
   const uint8_t * lastWr = receiveB + (n-1);
   do {
     spi2Transmit(0xFF);
     *receiveB = spi2Receive();  // should not wait long
   } while (++receiveB <  lastWr);

   lastRec: ;
   const uint8_t lRec = spi2Receive(); // may wait a bit (~1.5 s)
   if (n) *receiveB = lRec;
} // spi2ReceiveSN(uint8_t *, 2 * uint16_t)

/** Receive n bytes to a buffer via SPI 2 with optional skip (afterwards)
 *
 *  This function does \c skip + \c n receptions via SPI2 (putting all ones
 *  on MoSi). The (first) \c n bytes received are put into the buffer
 *  \c receiveB.
 *
 *  Compared to doing this by \ref spi2Tranceive "spi2Tranceive(0xFF)"
 *  \c n  + \c skip times this function is at least 20 % faster. On the other
 *  hand no useful work or yielding can be done while the SPI (2) interface is
 *  busy with the action. Insofar \c n + \c skip should be reasonably small.
 *
 *  \param receiveB pointer to the buffer to receive \c n bytes to
 *                  (must not be NULL if \c n != 0)
 *  \param n        number of bytes to be received and filled into \c receiveB
 *  \param skip     if > 0 \c skip bytes are received and forgotten
 *                  after \c receiveB was filled
 */
void spi2ReceiveNS(uint8_t * receiveB, uint16_t n, uint8_t skip){
   if (!(n | skip)) return;
   spi2Transmit(0xFF);

   if (!n) goto doSkip;
   if (n==1) goto lastRec;

   const uint8_t * lastWr = receiveB + (n-1);
   do {
     spi2Transmit(0xFF);
     *receiveB = spi2Receive();  // should not wait long
   } while (++receiveB < lastWr);

   lastRec: ;
   if (skip) spi2Transmit(0xFF);
   *receiveB = spi2Receive();

   doSkip: while (skip) {
      if (skip > 1) spi2Transmit(0xFF);
      --skip;
      spi2Receive();  // should not wait long
   }
} // spi2ReceiveNS(uint8_t *, uint16_t, uint8_t)

//----------------------------------------------------------------------


/* The SPI (1) serial transfer complete interrupt
 *
 * Hint :  SPI_STC_vect is 19 in iom1284p.h; but one manual says 20.
 *
 * Status: simple start version for weAut_01: just de-select all /CSs.
 *
 * Hint2:  Will / should not be used, at least with higher transfer rates
 *         like  ( clock / 4 )* 2.  Those high (10 MHz / 100 ns / 750ns)
 *         rates are no problem with shift registers and none with even the
 *         slowest memory cards.
 *         In high transfer rates a polling mode (w/o interrupt) is better,
 *         as interrupt overhead would consume considerable CPU time and
 *         would effectively slow down the transfers.
 * /    NOT USED
ISR( SPI_STC_vect) {
   // comment in for interrupt frequency measurement
   // static uint8_t int_toggle;
   //  setTestPin0(int_toggle = ~int_toggle); // measure send interrupt
   deSelectSPI1_impl();
   //  setTestPin0(OFF); // measure end interrupt
} //  SPI (1) serial transfer complete interrupt
*/


